(*
   File: elliptic2D.ml
   Author: Shivkumar Chandrasekaran
   Email: shiv@ece.ucsb.edu
   Date: 9.13.2004
   Purpose: Elliptic 2D solvers
   
   $Id$
 *)

open Bigarray
open Utils
open Nla
open Hss
open Chub


module Eller
    (Blop : BLOP)

    (Hss : HSS with type t         = Blop.t
                and type matrix    = Blop.matrix
                and type transform = Blop.transform)
    
    (Chub : CHUB with type t         = Blop.t
                  and type matrix    = Blop.matrix
                  and type transform = Blop.transform
		  and type hss       = Hss.hss) =
  struct

    open Blop

    type t      = Blop.t
    type matrix = Blop.matrix
    type hss    = Hss.hss

    type point = {x : float; y : float}

    (* Three counter-clockwise points *)
    type rectangle = {a : point; b : point; c : point}
    (* The line [a,b] is considered horizontal *)
    type cut = Horizontal | Vertical
    type grid = (rectangle, rectangle * cut) btree


    let size rectangle = (abs_float (rectangle.a.x -. rectangle.b.x),
			  abs_float (rectangle.c.y -. rectangle.b.y))


    let midPoint a b =
      {x = (a.x +. b.x) /. 2.0; y = (a.y +. b.y) /. 2.0}

    (* b - a *)
    let translator a b =
      {x = b.x -. a.x; y = b.y -. ay.y}

    (* p + a *)
    let translate p a =
      {x = a.x +. p.x; y = a.y +. p.y}


    let rec makeGrid stop rectangle =
      if stop rectangle then
	Leaf rectangle
      else begin
	let (width, height) = size rectangle in
	if width >= height then begin
	  let midH = midPoint rectangle.a rectangle.b in
	  let p = translator b midH in
	  let left  = {rectangle with a = midH}
	  and right = {rectangle with b = midH; c = translate p c} in
	  Node ((rectangle, Vertical),
		makeGrid stop left, makeGrid stop right)
	end else begin
	  let midV = midPpoint rectangle.b rectangle.c in
	  let p = translator b midV in
	  let top    = {rectangle with a = translate p a; b = midV}
	  and bottom = {rectangle with c = midV} in
	  Node ((rectangle, Horizontal),
		makeGrid stop top, makeGrid stop bottom)
	end
      end


    type boundaryCondition = Dirichlet | Neumann

  end (* module Eller (Blop) (Hss) (Chub) *)
 
