(*
   File: fmmOld.ml
   Author: Shivkumar Chandrasekaran
   Date: 3.31.2006
   Purpose: Fast algorithms for FMM matrices

   $Id: fmmOld.ml,v 1.21 2006/06/13 00:38:15 shiv Exp $
*)

open Bigarray
open Utils
open Nla
open Sparse
open Hss

module type FMM =
  sig
    type t
    type matrix
    type transform
    type sparse

    type sMatrix = Reg of matrix | Sup of superMatrix
    and  superMatrix = {oo : sMatrix; oi : sMatrix; io : sMatrix; ii : sMatrix}
    and  fmm_node = {bu : sMatrix; bl : sMatrix;
		     nl : int; nr : int; wl : matrix; wr : matrix;
		     ml : int; mr : int; rl : matrix; rr : matrix}
    and  fmm_leaf = {u : matrix; v : matrix; d : matrix}
    and  fmm = (fmm_leaf, fmm_node) btree
    val no_of_rows : fmm -> int
    val no_of_cols : fmm -> int
    val fastMul : fmm -> matrix -> matrix -> unit
    type odiag_split =
      | NoSplit
      | Split of odiag_split * odiag_split * odiag_split * odiag_split
    type leafSpine = { m : int; n : int }
    type nodeSpine = { mL : int; mR : int; nL : int; nR : int;
		       ou : odiag_split; ol : odiag_split }
    type fmmSpine = (leafSpine, nodeSpine) btree
    val fmmSize : fmmSpine -> int * int
    val fromFn : float -> fmmSpine ->
      (int -> int -> int -> int -> matrix -> unit) -> fmm
    type point = float array
    type nestedDissectionTree = (point * point * point list, point * point) btree
    val nestedDissection : int -> point list -> nestedDissectionTree
    val fmmSpineFromND : nestedDissectionTree -> fmmSpine
    val pointsFromNDtree : nestedDissectionTree -> point list
    val ppFmmRanks : fmm -> unit
    val ppBs : odiag_split -> unit
    val fastSolve' : fmm -> matrix -> matrix -> unit

  end (* module type FMM *)


module Fmmer (Blop : BLOP)
    (Sparse : SPARSE with type t = Blop.t and type matrix = Blop.matrix)
    (Hss : HSS with type t = Blop.t and type matrix = Blop.matrix) =
  struct

    open Blop
    type t = Blop.t
    type matrix = Blop.matrix
    type transform = Blop.transform
    type sparse = Sparse.sparse

    type sMatrix = Reg of matrix | Sup of superMatrix
    and  superMatrix = {oo : sMatrix; oi : sMatrix; io : sMatrix; ii : sMatrix}
    and  fmm_node = {bu : sMatrix; bl : sMatrix;
		     nl : int; nr : int; wl : matrix; wr : matrix;
		     ml : int; mr : int; rl : matrix; rr : matrix}
    and  fmm_leaf = {u : matrix; v : matrix; d : matrix}
    and  fmm = (fmm_leaf, fmm_node) btree

    let no_of_rows fmm =
      match fmm with
      | Leaf a -> noOfRows a.d
      | Node (a, _, _) -> a.ml + a.mr

    let no_of_cols fmm =
      match fmm with
      | Leaf a -> noOfCols a.d
      | Node (a, _, _) -> a.nl + a.nr


    let add_opt a b =
      match a, b with
      | None, None -> None
      | Some _, None -> a
      | None, Some _ -> b
      | Some a, Some b -> Some (a +$ b)

    let rec add_f fu fl =
      match fu, fl with
      | Leaf fu, Leaf fl ->
	  Leaf (fu +$ fl)
      | Node (fu, ful, fur), Node (fl, fll, flr) ->
	  Node (add_opt fu fl, add_f ful fll, add_f fur flr)
      | Leaf fu, Node (fl, fll, flr) ->
	  Node (add_opt (Some fu) fl, fll, flr)
      | Node (fu, ful, fur), Leaf fl ->
	  Node (add_opt fu (Some fl), ful, fur)

    let mulOpt a b nrhs =
      match b with
      | None -> zeros (noOfRows a) nrhs
      | Some b -> a *$ b

    let rec b_x_g b g =
      match b with
      | Reg b -> Leaf (b *$ (getNodeData g))
      | Sup b -> begin
	  match g with
	  | Leaf _ -> failwith "Bad fmm g tree"
	  | Node (_, gl, gr) ->
	      let fu = add_f (b_x_g b.oo gl) (b_x_g b.oi gr)
	      and fl = add_f (b_x_g b.io gl) (b_x_g b.ii gr) in
	      Node (None, fu, fl)
      end

    (* The actual FMM multiply. *)
    let fastMul fmm x b =
      let nrhs = noOfCols b in
      let rec upSweep fmm x =
	match fmm with
	| Leaf a -> Leaf (a.v *~$ x)
	| Node (a, lFmm, rFmm) ->
	    let (xl, xr) = partition2x1 a.nl a.nr x in
	    let gl = upSweep lFmm xl
	    and gr = upSweep rFmm xr in
	    Node (a.wl *~$ (getNodeData gl) +$ a.wr *~$ (getNodeData gr),
		  gl, gr)
      in
      let mulOpt1 a b = mulOpt a b nrhs in
      let rec downSweep fmm f gs x b =
	match fmm with
	| Leaf a ->
	    begin match f with
	    | Leaf f -> (a.d |*| x) b; (a.u |*++| f) b
	    | Node _ -> failwith "Bad fmm tree"
	    end
	| Node (a, lFmm, rFmm) ->
	    match gs with
	    | Leaf _ -> failwith "Bad fmm g tree"
	    | Node (_, gl, gr) -> begin
		let fu = b_x_g a.bu gr
		and fl = b_x_g a.bl gl
		and (bl, br) = partition2x1 a.ml a.mr b
		and (xl, xr) = partition2x1 a.nl a.nr x in
		let (foL, foR) =
		  match f with
		  | Leaf f -> (Leaf (a.rl *$ f), Leaf (a.rr *$ f))
		  | Node (f, fL, fR) -> (add_f (Leaf (mulOpt1 a.rl f)) fL,
					 add_f (Leaf (mulOpt1 a.rr f)) fR)
		in
		downSweep lFmm (add_f foL fu) gl xl bl;
		downSweep rFmm (add_f foR fl) gr xr br
	    end
      in (* let rec downSweep *)
      assert (noOfCols x = nrhs);
      downSweep fmm (Leaf (zeros 0 nrhs)) (upSweep fmm x) x b


    type odiag_split =
      | NoSplit
      | Split of odiag_split * odiag_split * odiag_split * odiag_split
    and leafSpine = { m : int; n : int }
    and nodeSpine = { mL : int; mR : int; nL : int; nR : int;
		      ou : odiag_split; ol : odiag_split }
    and fmmSpine = (leafSpine, nodeSpine) btree

    let rec fmmSize fmmSpine =
      match fmmSpine with
      | Leaf a -> (a.m, a.n)
      | Node (a, lSpine, rSpine) ->
	  let l = fmmSize lSpine
	  and r = fmmSize rSpine in
	  (fst l + fst r, snd l + snd r)

    let rec fillFmmSpine fmmSpine =
      let rec loop fmmSpine =
	match fmmSpine with
	| Leaf a -> (fmmSpine, a.m, a.n)
	| Node (a, lSpine, rSpine) ->
	    let (lSpine', mL, nL) = loop lSpine
	    and (rSpine', mR, nR) = loop rSpine in
	    (Node ({a with mL = mL; mR = mR; nL = nL; nR = nR},
		   lSpine', rSpine'),
	     mL+mR, nL+nR)
      in
      let (fmmSpine', _, _) = loop fmmSpine in
      fmmSpine'

    (* numerical-rank SVD *)
    let svdE' ?(ooc=false) tol a =
      let k = min (noOfRows a) (noOfCols a) in
      let (u, s, vt) = svd' ~ooc a in
      let r = ref 0 in
      while !r < k && s.{!r+1} > tol do
	r := !r + 1
      done;
      assert (!r <= k && (!r = 0 || s.{!r} > tol));
      (u $ (All, 1 <-> !r), Array1.sub s 1 !r, vt $ (1 <-> !r, All))


    (* acc is a disjoint union of intervals U_i (a_i, b_i) where a_i
	is the first integer of interval i, and b_i is length of interval. *)
    let mergeIntervals bs acc =
      let rec mergeInterval ((kf, k) as k0) acc =
	match acc with
	| [] -> [k0]
	| (nf, n) :: ns ->
	    let lk = kf + k - 1
	    and ln = nf + n - 1 in
	    if lk < nf then
	      (kf, k) :: acc
	    else if kf <= nf then
	      if lk <= ln then
		(kf, ln - kf + 1) :: ns
	      else
		mergeInterval (kf, k) ns
	    else if kf <= ln then
	      if lk <= ln then
		acc
	      else
		mergeInterval (nf, lk - nf + 1) ns
	    else
	      (nf, n) :: mergeInterval (kf, k) ns
      in
      List.fold_right mergeInterval bs acc

    (* We need a formal mechanism to hold the G and H matrices for the
       construction algorithm. These matrices are block rows/columns of
       the original matrix from which some sub-blocks have been stripped
       out. We also need to be able to merge two such rows/columns
       together. Compact storage is a plus as SVD's need to be
       computed. *)
    type strippedBlock =
      | Row of (int * int) list * matrix
      | Col of (int * int) list * matrix

    let checkStrippedBlock strpBlk =
      let rec loop ok nl n intervals =
	if ok then begin
	  match intervals with
	  | [] -> (ok, n)
	  | (kf, k) :: ks when k > 0 ->
	      if kf > nl then
		loop ok (kf + k - 1) (n + k) ks
	      else
		(false, n)
	  | _ -> (false, n)
	end else
	  (false, n)
      in
      match strpBlk with
      | Row (intervals, block) ->
	  let (ok, n) = loop true 0 0 intervals in
	  ok && n = noOfCols block
      | Col (intervals, block) ->
	  let (ok, m) = loop true 0 0 intervals in
	  ok && m = noOfRows block

    (* Strip the given intervals. Use mergeIntervals before stripping
       to reduce number of unwarranted copies. This can be improved
       dramatically if necessary. *)
    let stripIntervals strpBlk intervals =
      let stripRowInterval strpBlk (kf, k) =
	let kl = kf + k - 1 in
	let rec loop pl g blkIntervals acc =
	  match blkIntervals with
	  | [] -> failwith "Interval outside range"
	  | (nf, n) :: ns ->
	      let nl = nf + n - 1 in
	      assert (kf >= nf);
	      if kf > nl then
		loop (pl + n) g ns ((nf, n) :: acc)
	      else begin
		assert (kl <= nl);
		let ng = noOfCols g in
		let gnew = rawMatrix (noOfRows g) (ng - k) in
		matrix1x2 (pl+kf-nf) (ng-k-pl-kf+nf)
		  (iD (g $ (All, 1 <-> pl+kf-nf)))
		  (iD (g $ (All, pl+kf-nf+k+1 <-> ng))) gnew;
		if kf > nf then
		  if kl < nl then
		    (gnew, List.rev_append acc
		       ((nf, kf-nf) :: (kf+k, nl-kl) :: ns))
		  else
		    (gnew, List.rev_append acc ((nf, kf - nf) :: ns))
		else
		  if kl < nl then
		    (gnew, List.rev_append acc ((kl+1, nl-kl) :: ns))
		  else
		    (gnew, List.rev_append acc ns)
	      end
	in
	let (gnew, newBlkIntervals) =
	  match strpBlk with
	  | Row (blkIntervals, block) ->
	      if k > 0 then
		loop 0 block blkIntervals []
	      else
		(block, blkIntervals)
	  | Col _ -> failwith "Wrong stripping"
	in
	Row (List.filter (fun ints -> snd ints > 0) newBlkIntervals, gnew)
      in
      let stripColInterval strpBlk (kf, k) =
	let kl = kf + k - 1 in
	let rec loop pl g blkIntervals acc =
	  match blkIntervals with
	  | [] -> failwith "Interval outside range"
	  | (nf, n) :: ns ->
	      let nl = nf + n - 1 in
	      assert (kf >= nf);
	      if kf > nl then
		loop (pl + n) g ns ((nf, n) :: acc)
	      else begin
		assert (kl <= nl);
		let mg = noOfRows g in
		let gnew = rawMatrix (mg - k) (noOfCols g) in
		matrix2x1
		       (pl+kf-nf) (iD (g $ (1 <-> pl+kf-nf, All)))
		  (mg-k-pl-kf+nf) (iD (g $ (pl+kf-nf+k+1 <-> mg, All))) gnew;
		if kf > nf then
		  if kl < nl then
		    (gnew, List.rev_append acc
		       ((nf, kf-nf) :: (kf+k, nl-kl) :: ns))
		  else
		    (gnew, List.rev_append acc ((nf, kf - nf) :: ns))
		else
		  if kl < nl then
		    (gnew, List.rev_append acc ((kl+1, nl-kl) :: ns))
		  else
		    (gnew, List.rev_append acc ns)
	      end
	in
	let (gnew, newBlkIntervals) =
	  match strpBlk with
	  | Col (blkIntervals, block) ->
	      if k > 0 then
		loop 0 block blkIntervals []
	      else
		(block, blkIntervals)
	  | Row _ -> failwith "Wrong stripping"
	in
	Col (List.filter (fun ints -> snd ints > 0) newBlkIntervals, gnew)
      in
      match strpBlk with
      | Row _ ->
	  let newStrpBlk = List.fold_left stripRowInterval strpBlk intervals in
	  assert (checkStrippedBlock newStrpBlk);
	  newStrpBlk
      | Col _ ->
	  let newStrpBlk = List.fold_left stripColInterval strpBlk intervals in
	  assert (checkStrippedBlock newStrpBlk);
	  newStrpBlk

    (* From the given row strip block get the desired column blocks. *)
    let getRowIntervalBlks strpBlk intervals =
      let getInterval strpBlk (kf, k) =
	assert (k >= 0);
	let kl = kf + k - 1 in
	(* pl is the last column in the storage of strpBlk *)
	let rec loop pl g blkIntervals =
	  match blkIntervals with
	  | [] -> failwith
		(Printf.sprintf "(%i, %i) Interval not available" kf k)
	  | (nf, n) :: ns ->
	      let nl = nf + n - 1 in
	      assert (kf >= nf);
	      if kf > nl then
		loop (pl + n) g ns
	      else begin
		assert (kl <= nl);
		let bnew = rawMatrix (noOfRows g) k in
		bnew <-$ (g $ (All, pl+1+kf-nf <-> pl+kf-nf+k));
		bnew
	      end
	in
	match strpBlk with
	| Row (blkIntervals, block) ->
	    if k = 0 then
	      rawMatrix (noOfRows block) 0
	    else
	      loop 0 block blkIntervals
	| Col _ -> failwith "Wrong stripping"
      in
      List.map (getInterval strpBlk) intervals


    (* Merge two stripped blocks. Columns must line up! *)
    let mergeStrips strpBlk1 strpBlk2 =
      match strpBlk1, strpBlk2 with
      | Row (intervals1, block1), Row (intervals2, block2) ->
	  assert (intervals1 = intervals2);
	  let m1 = noOfRows block1
	  and m2 = noOfRows block2 in
	  let nBlk = rawMatrix (m1+m2) (noOfCols block1) in
	  matrix2x1 m1 (iD block1)
	            m2 (iD block2) nBlk;
	  Row (intervals1, nBlk)
      | Col (intervals1, block1), Col (intervals2, block2) ->
	  assert (intervals1 = intervals2);
	  let n1 = noOfCols block1
	  and n2 = noOfCols block2 in
	  let nBlk = rawMatrix (noOfRows block1) (n1+n2) in
	  matrix1x2  n1         n2
	        (iD block1) (iD block2) nBlk;
	  Col (intervals1, nBlk)
      | _ -> failwith "Incompatible row/column blocks"


    (* Pass down a list of row-blocks and column-blocks to each node,
       for computing U and V at that level. For each such block B is
       computed. The R and W follow from the U and V computation. Here
       fn fstRow noOfRows fstCol noOfCols is what is assumed. *)
    let fromFn tol fmmSpine fn =
      let (m, n) = fmmSize fmmSpine in
      let getNodeData tree =
	match tree with
	| Leaf a -> a
	| Node (a, _, _) -> a
      in
      let rec addBlks lBlks rBlks =
	match lBlks, rBlks with
	| Leaf bs, Leaf cs -> Leaf (bs @ cs)
	| Leaf bs, Node (cs, rL, rR) -> Node (bs @ cs, rL, rR)
	| Node (bs, lL, lR), Leaf cs -> Node (bs @ cs, lL, lR)
	| Node (bs, lL, lR), Node (cs, rL, rR) ->
	    Node (bs @ cs, addBlks lL rL, addBlks lR rR)
      in
      let rec stripeRowBlks nl od rSpine =
	match od, rSpine with
	| NoSplit, Leaf a -> Leaf [(nl+1, a.n)]
	| NoSplit, Node (a, _, _) -> Leaf [(nl+1, a.nL+a.nR)]
	| Split (od00, od01, od10, od11), Node (a, lSpine, rSpine) ->
	    Node ([],
		  addBlks (stripeRowBlks nl od00 lSpine)
		    (stripeRowBlks (nl+a.nL) od01 rSpine),
		  addBlks (stripeRowBlks nl od10 lSpine)
		    (stripeRowBlks (nl+a.nL) od11 rSpine))
	| _ -> failwith "Bad fmm/B"
      in
      let rec unStripeRowBlks bvs od =
	match bvs, od with
	| Leaf bs, NoSplit ->
	    (Reg (List.hd bs), Leaf (List.tl bs))
	| Node (bs, lBvs, rBvs), NoSplit ->
	    (Reg (List.hd bs), Node (List.tl bs, lBvs, rBvs))
	| Node (bs, lBvs, rBvs), Split (od00, od01, od10, od11) ->
	    let (b00, lBvs1) = unStripeRowBlks lBvs  od00 in
	    let (b01, lBvs2) = unStripeRowBlks lBvs1 od01 in
	    let (b10, rBvs1) = unStripeRowBlks rBvs  od10 in
	    let (b11, rBvs2) = unStripeRowBlks rBvs1 od11 in
	    (Sup {oo = b00; oi = b01; io = b10; ii = b11},
	     Node (bs, lBvs2, rBvs2))
	| _ -> failwith "Bad bvs/od"
      in
      let rec getBs bvs vs =
	match bvs, vs with
	| Reg bv, Leaf v -> Reg (bv *$ v)
	| Reg bv, Node (v, _, _) -> Reg (bv *$ v)
	| Sup bv, Node (_, vsL, vsR) ->
	    Sup { oo = getBs bv.oo vsL; oi = getBs bv.oi vsR;
		  io = getBs bv.io vsL; ii = getBs bv.ii vsR }
	| _ -> failwith "Sup/Leaf in getBs"
      in
      let rec stripeColBlks ml od lSpine =
	match od, lSpine with
	| NoSplit, Leaf a -> Leaf [(ml+1, a.m)]
	| NoSplit, Node (a, _, _) -> Leaf [(ml+1, a.mL+a.mR)]
	| Split (od00, od01, od10, od11), Node (a, lSpine, rSpine) ->
	    Node ([],
		  addBlks (stripeColBlks ml od00 lSpine)
		    (stripeColBlks (ml+a.mL) od10 rSpine),
		  addBlks (stripeColBlks ml od01 lSpine)
		    (stripeColBlks (ml+a.mL) od11 rSpine))
	| _ -> failwith "Bad fmm/B"
      in
      (* ml, nl, are last previous row and column *)
      let rec split fmmSpine ml nl rBlks cBlks =
	match fmmSpine with
	| Leaf a ->
	    let d = rawMatrix a.m a.n in
	    d <-| fn (ml+1) a.m (nl+1) a.n;
	    let (u, h, bvs) =
	      let h = rawMatrix a.m (n - a.n) in
	      matrix1x2 nl (n - a.n - nl)
		(fn (ml+1) a.m 1 nl) (fn (ml+1) a.m (nl+a.n+1) (n-a.n-nl)) h;
	      let (uh, sh, vht) = svdE' tol h in
	      rowScale_real' sh vht;
	      let h = Row ([(1, nl); (nl+a.n+1, n-a.n-nl)], vht) in
	      let (bvs, exclR) =
		match rBlks with
		| Leaf bs -> (Leaf (getRowIntervalBlks h bs),
			      mergeIntervals bs [])
		| _    -> failwith "Incorrect level in rBlks"
	      in
	      let hR = stripIntervals h exclR in
	      (uh, hR, bvs)
	    in
	    let (g, v) =
	      let g = rawMatrix (m - a.m) a.n in
	      matrix2x1    ml  (fn 1 ml (nl+1) a.n)
		(m - a.m - ml) (fn (ml+a.m+1) (m-a.m-ml) (nl+1) a.n) g;
	      let (ug, sg, vgt) = svdE' tol g in
	      colScale_real' ug sg;
	      let g = Col ([(1, ml); (ml+a.m+1, m-a.m-ml)], ug) in
	      let exclC =
		match cBlks with
		| Leaf bs -> mergeIntervals bs []
		| Node _  -> failwith "Incorrect level in cBlks"
	      in
	      let gR = stripIntervals g exclC in
	      (gR, transp vgt)
	    in
	    (Leaf {d = d; u = u; v = v}, h, g, bvs, Leaf v)    
	| Node (a, lSpine, rSpine) ->
	    let (bBlksT, bBlksD) =
	      let bBlksT = stripeRowBlks (nl+a.nL) a.ou rSpine
	      and bBlksD = stripeRowBlks nl a.ol lSpine in
	      match rBlks with
	      | Leaf _ -> (bBlksT, bBlksD)
	      | Node (_, rBlksT, rBlksD) ->
		  (addBlks bBlksT rBlksT, addBlks bBlksD rBlksD)
	    in
	    let (bBlksL, bBlksR) =
	      let bBlksL = stripeColBlks (ml+a.mL) a.ol rSpine
	      and bBlksR = stripeColBlks ml a.ou lSpine in
	      match cBlks with
	      | Leaf _ -> (bBlksL, bBlksR)
	      | Node (_, cBlksL, cBlksR) ->
		  (addBlks bBlksL cBlksL, addBlks bBlksR cBlksR)
	    in
	    let (fmmL, svhtL, usgL, bvT, vL) =
	      split lSpine ml nl bBlksT bBlksL in
	    let (fmmR, svhtR, usgR, bvD, vR) =
	      split rSpine (ml + a.mL) (nl + a.nL) bBlksD bBlksR in
	    let (buvs, buvs1) = unStripeRowBlks bvT a.ou
	    and (blvs, blvs1) = unStripeRowBlks bvD a.ol in
	    let bu = getBs buvs vR
	    and bl = getBs blvs vL in
	    let (rl, rr, h, bvs) =
	      let sRows a =
		match a with
		| Row (_, a) -> noOfRows a
		| _ -> failwith "Row expected"
	      in
	      let svht = mergeStrips svhtL svhtR in
	      let (hInts, h) =
		match svht with
		| Row (hInts, h) -> (hInts, h)
		| _ -> failwith "huh?"
	      in
	      let (r, s, vt) = svdE' tol h in
	      rowScale_real' s vt;
	      let (rl, rr) = partition2x1 (sRows svhtL) (sRows svhtR) r in
	      let svtnew = Row (hInts, vt) in
	      let (bvs, exclR) =
		let bs = getNodeData rBlks in
		(Node (getRowIntervalBlks svtnew bs, buvs1, blvs1),
		 mergeIntervals bs [])
	      in
	      (rl, rr, stripIntervals svtnew exclR, bvs)
	    in
	    let (wl, wr, g, v) =
	      let sCols a =
		match a with
		| Col (_, a) -> noOfCols a
		| _ -> failwith "Col expected"
	      in
	      let usg = mergeStrips usgL usgR in
	      let (gInts, gb) =
		match usg with
		| Col (gInts, g) -> (gInts, g)
		| _ -> failwith "huh?"
	      in
	      let (u, s, wt) = svdE' tol gb in
	      colScale_real' u s;
	      let usgnew = Col (gInts, u) in
	      let (wlt, wrt) = partition1x2 (sCols usgL) (sCols usgR) wt in
	      let v = rawMatrix (a.nL + a.nR) (noOfRows wt) in
	      matrix2x1 a.nL (getNodeData vL |*~| wlt)
		        a.nR (getNodeData vR |*~| wrt) v;
	      let exclC = mergeIntervals (getNodeData cBlks) [] in
	      (transp wlt, transp wrt, stripIntervals usgnew exclC,
	       Node (v, vL, vR))
	    in
	    (Node ({ml = a.mL; mr = a.mR; nl = a.nL; nr = a.nR;
		    bu = bu; bl = bl;
		    rl = rl; rr = rr; wl = wl; wr = wr},
		   fmmL, fmmR),
	     h, g, bvs, v)
      in
      let (fmm, _, _, _, _) = split (fillFmmSpine fmmSpine) 0 0
	  (Leaf []) (Leaf []) in
      fmm

    let arrayWith a i ai =
      let b = Array.copy a in
      b.(i) <- ai;
      b

    type point = float array
    type nestedDissectionTree = (point * point * point list, point * point) btree

    let nestedDissection p xs : nestedDissectionTree =
      let d = Array.length (List.hd xs) in
      assert (List.for_all (fun x -> Array.length x = d) xs);
      let bl = Array.init d (fun i -> List.fold_left (fun m x -> min m x.(i))
	  infinity xs) in
      let tr = Array.init d (fun i -> List.fold_left (fun m x -> max m x.(i))
	  neg_infinity xs) in
      let rec nd bl tr xs =
	if List.length xs <= p then begin
	  Leaf (bl, tr, xs)
	end else begin
	  let i =
	    let mv = ref neg_infinity
	    and m = ref (-1) in
	    Array.iteri (fun j blj -> if tr.(j) -. blj > !mv then begin
	      mv := tr.(j) -. blj; m := j end) bl;
	    !m
	  in
	  let mi = (tr.(i) +. bl.(i)) /. 2.0 in
	  let (xs1, xs2) = List.partition
	      (fun x -> assert (bl.(i) <= x.(i) && x.(i) <= tr.(i));
		x.(i) <= mi) xs
	  in
	  Node ((bl, tr),
		nd bl (arrayWith tr i mi) xs1, nd (arrayWith bl i mi) tr xs2)
	end
      in
      nd bl tr xs

    let rec pointsFromNDtree ndTree =
      match ndTree with
      | Leaf (_, _, a) -> a
      | Node (_, lNDtree, rNDtree) ->
	  pointsFromNDtree lNDtree @ pointsFromNDtree rNDtree

    let tooClose (bl1, tr1) (bl2, tr2) =
      let dim = Array.length bl1 in
      let l = Array.make dim 0.0
      and r = Array.make dim 0.0 in
      for i = 0 to dim-1 do
	if tr1.(i) <= bl2.(i) then begin
	  l.(i) <- tr1.(i);
	  r.(i) <- bl2.(i)
	end else if tr2.(i) <= bl1.(i) then begin
	  l.(i) <- bl1.(i);
	  r.(i) <- tr2.(i)
	end else begin
	  l.(i) <- 0.0;
	  r.(i) <- 0.0
	end
      done;
      let dist = ref 0.0 in
      for i = 0 to dim-1 do
	dist := !dist +. (l.(i) -. r.(i)) ** 2.0
      done;
      dist := sqrt !dist;
      let tc = ref false
      and i = ref 0 in
      while not !tc && !i < dim do
	tc := tr1.(!i) -. bl1.(!i) > !dist || tr2.(!i) -. bl2.(!i) > !dist;
	incr i
      done;
      !tc

    let rec makeBs lNDtree rNDtree =
      match lNDtree, rNDtree with
      | Leaf _, Leaf _ -> NoSplit
      | Leaf _, Node _ -> NoSplit
      | Node _, Leaf _ -> NoSplit
      | Node ( lRect, ll, lr), Node (rRect, rl, rr) ->
	  if tooClose lRect rRect then
	    Split (makeBs ll rl, makeBs ll rr,
		   makeBs lr rl, makeBs lr rr)
	  else
	    NoSplit

    let rec fmmSpineFromND ndTree =
      match ndTree with
      | Leaf (bl, tr, xs) -> Leaf { m = List.length xs; n = List.length xs }
      | Node ((bl, tr), lNDtree, rNDtree) ->
	  let lFmmSpine = fmmSpineFromND lNDtree
	  and rFmmSpine = fmmSpineFromND rNDtree in
	  Node ({mL= -1; mR= -1; nL= -1; nR= -1;
		 ou = makeBs lNDtree rNDtree;
		 ol = makeBs rNDtree lNDtree},
		lFmmSpine, rFmmSpine)

    let ppFmmRanks fmm =
      ppBtree
	(fun a -> pp2Tuple Format.print_int Format.print_int
	    (noOfCols a.u, noOfCols a.v))
	(fun a -> pp2Tuple Format.print_int Format.print_int
	    (noOfCols a.rl, noOfCols a.wl))
	fmm

    let rec ppBs b =
      match b with
      | NoSplit ->
	  Format.open_box 0;
	  Format.print_string "NoSplit";
	  Format.close_box ()
      | Split (oo, oi, io, ii) ->
	  Format.open_box 0;
	  Format.print_string "Split";
	  Format.print_space ();
	  Format.open_vbox 0;
	  Format.print_char '(';
	  Format.open_box 0;
	  ppBs oo;
	  Format.print_char ',';
	  Format.print_space ();
	  ppBs oi;
	  Format.print_char ',';
	  Format.close_box ();
	  Format.print_space ();
	  Format.open_box 0;
	  ppBs io;
	  Format.print_char ',';
	  Format.print_space ();
	  ppBs ii;
	  Format.close_box ();
	  Format.print_char ')';
	  Format.close_box ()

    (* Least common of trees *)
    let rec mergeTrees b1 b2 =
      if structurallyEqual b1 b2 then
	b1
      else
	match b1, b2 with
	| Leaf _, Node _ -> b2
	| Node _, Leaf _ -> b1
	| Node (d, b1l, b1r), Node (_, b2l, b2r) ->
	    Node (d, mergeTrees b1l b2l, mergeTrees b1r b2r)
	| Leaf _, Leaf _ -> prerr_endline "Tree = failed"; b1

    let mergeHss lHss rHss =
      match lHss, rHss with
      | Leaf l, Leaf r ->
	  Node ({Hss.ml = noOfRows l.Hss.d; Hss.mr = noOfRows r.Hss.d;
		 Hss.nl = noOfCols l.Hss.d; Hss.nr = noOfCols r.Hss.d;
		 Hss.rl = zeros (noOfCols l.Hss.u) 0;
		 Hss.rr = zeros (noOfCols r.Hss.u) 0;
		 Hss.wl = zeros (noOfCols l.Hss.v) 0;
		 Hss.wr = zeros (noOfCols r.Hss.v) 0;
		 Hss.bu = zeros 0 0; Hss.bl = zeros 0 0},
		lHss, rHss)
      | Leaf l, Node (r, _, _) ->
	  Node ({Hss.ml = noOfRows l.Hss.d; Hss.mr = r.Hss.ml + r.Hss.mr;
		 Hss.nl = noOfCols l.Hss.d; Hss.nr = r.Hss.nl + r.Hss.nr;
		 Hss.rl = zeros (noOfCols l.Hss.u ) 0;
		 Hss.rr = zeros (noOfCols r.Hss.rl) 0;
		 Hss.wl = zeros (noOfCols l.Hss.v ) 0;
		 Hss.wr = zeros (noOfCols r.Hss.wl) 0;
		 Hss.bu = zeros 0 0; Hss.bl = zeros 0 0},
		lHss, rHss)
      | Node (l, _, _), Leaf r ->
	  Node ({Hss.ml = l.Hss.ml + l.Hss.mr; Hss.mr = noOfRows r.Hss.d;
		 Hss.nl = l.Hss.nl + l.Hss.nr; Hss.nr = noOfCols r.Hss.d;
		 Hss.rl = zeros (noOfCols l.Hss.rl) 0;
		 Hss.rr = zeros (noOfCols r.Hss.u ) 0;
		 Hss.wl = zeros (noOfCols l.Hss.wl) 0;
		 Hss.wr = zeros (noOfCols r.Hss.v ) 0;
		 Hss.bu = zeros 0 0; Hss.bl = zeros 0 0},
		lHss, rHss)
      | Node (l, _, _), Node (r, _, _) ->
	  Node ({Hss.ml = l.Hss.ml + l.Hss.mr; Hss.mr = r.Hss.ml + r.Hss.mr;
		 Hss.nl = l.Hss.nl + l.Hss.nr; Hss.nr = r.Hss.nl + r.Hss.nr;
		 Hss.rl = zeros (noOfCols l.Hss.rl) 0;
		 Hss.rr = zeros (noOfCols r.Hss.rl) 0;
		 Hss.wl = zeros (noOfCols l.Hss.wl) 0;
		 Hss.wr = zeros (noOfCols r.Hss.wl) 0;
		 Hss.bu = zeros 0 0; Hss.bl = zeros 0 0},
		lHss, rHss)

    let rec uTreeOfB b =
      match b with
      | Reg _ -> Leaf ()
      | Sup b -> Node ((),
		       mergeTrees (uTreeOfB b.oo) (uTreeOfB b.oi),
		       mergeTrees (uTreeOfB b.io) (uTreeOfB b.ii))

    let rec vTreeOfB b =
      match b with
      | Reg _ -> Leaf ()
      | Sup b ->
	  Node ((),
		mergeTrees
		  (vTreeOfB b.oo)
		  (vTreeOfB b.io),
		mergeTrees
		  (vTreeOfB b.oi)
		  (vTreeOfB b.ii))

    let rec collapseBasisTree us =
      match us with
      | Leaf a -> a
      | Node ((rl, rr), uL, uR) ->
	  let u1 = collapseBasisTree uL
	  and u2 = collapseBasisTree uR in
	  let m1 = noOfRows u1
	  and m2 = noOfRows u2
	  and n  = noOfCols rl in
	  let u  = rawMatrix (m1 + m2) n in
	  matrix2x1 m1 (u1 |*| rl)
	            m2 (u2 |*| rr) u;
	  u

    let rec pruneBasisTree us uTree =
      if structurallyEqual us uTree then
	us
      else
	match us, uTree with
	| Node _, Leaf _ -> Leaf (collapseBasisTree us)
	| Node (r, uL, uR), Node (_, uTreeL, uTreeR) ->
	    Node (r, pruneBasisTree uL uTreeL, pruneBasisTree uR uTreeR)
	| Leaf _, Node _ -> failwith "premature pruning?"
	| Leaf _, Leaf _ -> failwith "structurallyEqual failed"

    let rec collapseHssTree vs fmm =
      match vs, fmm with
      | Leaf a, _ -> vs
      | Node (av, vL, vR), Node (af, lFmm, rFmm) ->
	  let v1 = collapseHssTree vL lFmm
	  and v2 = collapseHssTree vR rFmm in
	  begin match v1, v2 with
	  | Leaf a1, Leaf a2 ->
	      let m1 = noOfRows a1.Hss.d
	      and m2 = noOfRows a2.Hss.d
	      and n  = noOfCols af.wl in
	      let d  = rawMatrix (m1 + m2) n
	      and u  = rawMatrix (m1 + m2) (noOfCols av.Hss.rl)
	      and v  = rawMatrix n (noOfCols av.Hss.wl)
	      and vl = af.wl *~$ a1.Hss.v
	      and vr = af.wr *~$ a2.Hss.v in
	      matrix2x1 m1 (a1.Hss.d *$ af.wl |+| a1.Hss.u *$ av.Hss.bu *$~ vr)
	                m2 (a2.Hss.u *$ av.Hss.bl *$~ vl |+| a2.Hss.d *$ af.wr)
		d;
	      matrix2x1 m1 (a1.Hss.u |*| av.Hss.rl)
		        m2 (a2.Hss.u |*| av.Hss.rr) u;
	      v <-| ((vl *$ av.Hss.wl) |+| (vr *$ av.Hss.wr));
	      Leaf {Hss.d = d; Hss.u = u; Hss.v = v}
	    | _, _ -> failwith "Returned non-Leaf"
	  end
      | _, _ -> failwith "Mismatched vHss and fmm trees"
	      
    let resetRowColumnPartitions vs =
      let rec loop vs =
	match vs with
	| Leaf a -> (noOfRows a.Hss.d, noOfCols a.Hss.d, vs)
	| Node (a, lV, rV) ->
	    let (m1, n1, vL) = loop lV
	    and (m2, n2, vR) = loop rV in
	    (m1 + m2, n1 + n2, Node ({a with Hss.ml = m1; Hss.mr = m2;
				      Hss.nl = n1; Hss.nr = n2},
				     vL, vR))
      in
      let _, _, vNew = loop vs in
      vNew

    let rec pruneHssTree vs vTree fmm =
      if structurallyEqual vs vTree then
	vs
      else
	let vNew =
	  match vs, vTree, fmm with
	  | Node _, Leaf _, Node _ -> collapseHssTree vs fmm
	  | Node (a, vL, vR), Node (_, vTreeL, vTreeR), Node (_, lFmm, rFmm) ->
	      Node (a, pruneHssTree vL vTreeL lFmm, pruneHssTree vR vTreeR rFmm)
	  | Leaf _, Node _, _ -> failwith "Premature hss pruning?"
	  | Leaf _, Leaf _, _ -> failwith "structurallyEqual failed"
	  | _, _, _ -> failwith "Wooah"
	in
	resetRowColumnPartitions vNew

    let rec basisNoOfRows uv =
      match uv with
      | Leaf uv -> noOfRows uv
      | Node (_, ul, ur) -> basisNoOfRows ul + basisNoOfRows ur

    let rec collapseWtree fmm v =
      match v, fmm with
      | Leaf av, _ -> eye (noOfCols av.Hss.d) (noOfCols av.Hss.d)
      | Node (av, vL, vR), Node (a, fmmL, fmmR) ->
	  let wL = collapseWtree fmmL vL
	  and wR = collapseWtree fmmR vR in
	  let m1 = noOfRows wL
	  and m2 = noOfRows wR in
	  let w = rawMatrix (m1 + m2) (noOfCols a.wl) in
	  matrix2x1 m1 (wL |*| a.wl)
	            m2 (wR |*| a.wr) w;
	  w
      | _, _ -> failwith "Bad v/fmm trees"

    let u_b_vt u b rFmm v ubvt =
      let rec u_b_wt u b rFmm v ubwt =
	match b with
	| Reg b -> ubwt <-| ((collapseBasisTree u *$ b) |*~|
	                                             (collapseWtree rFmm v))
	| Sup b ->
	    begin match u, v, rFmm with
	    | Node (_, ul, ur), Node (_, vl, vr), Node (a, fmmL, fmmR) ->
		let m1 = basisNoOfRows ul
		and m2 = basisNoOfRows ur
		and n1 = Hss.no_of_cols vl
		and n2 = Hss.no_of_cols vr in
		matrix2x2 n1 n2
		  m1 (u_b_wt ul b.oo fmmL vl) (u_b_wt ul b.oi fmmR vr)
		  m2 (u_b_wt ur b.io fmmL vl) (u_b_wt ur b.ii fmmR vr) ubwt;
	    | _, _, _ -> failwith "Bad u, fmm in u_b_wt"
	    end
      in
      let ubwt = rawMatrix (basisNoOfRows u) (Hss.no_of_cols v) in
      u_b_wt u b rFmm v ubwt;
      let temp = rawMatrix (noOfCols ubvt) (noOfRows ubvt) in
      Hss.fastMul v (transp ubwt) temp;
      transp' temp ubvt

    let rec u_f u f nrhs =
      match u, f with
      | Leaf u, Leaf f ->  u *$ f
      | Node ((rl, rr), ul, ur), Node (f, fl ,fr) ->
	  let uf1 = u_f ul (add_f (Leaf (mulOpt rl f nrhs)) fl) nrhs
	  and uf2 = u_f ur (add_f (Leaf (mulOpt rr f nrhs)) fr) nrhs in
	  let m1 = noOfRows uf1
	  and m2 = noOfRows uf2 in
	  let uf = rawMatrix (m1 + m2) nrhs in
	  matrix2x1 m1 (iD uf1)
	            m2 (iD uf2) uf;
	  uf
      | Node ((rl, rr), ul, ur), Leaf f ->
	  let uf1 = u_f ul (Leaf (rl *$ f)) nrhs
	  and uf2 = u_f ur (Leaf (rr *$ f)) nrhs in
	  let m1 = noOfRows uf1
	  and m2 = noOfRows uf2 in
	  let uf = rawMatrix (m1 + m2) nrhs in
	  matrix2x1 m1 (iD uf1)
	            m2 (iD uf2) uf;
	  uf
      | _, _ -> failwith "Mismatched u, f"

    let rec compressU uTree =
      match uTree with
      | Leaf u ->
	  let m = noOfRows u
	  and n = noOfCols u in
	  assert (m >= n);
	  if m > n then (* compress *) begin
	    let (q, l_u) = ql' u in
	    ([(1, m, m-n, q)], Leaf (copyL l_u))
	  end else
	    ([(1, m, 0, eyeTransform)], uTree)
      | Node (r, uTreeL, uTreeR) ->
	  let (lems, luTree) = compressU uTreeL
	  and (rems, ruTree) = compressU uTreeR
	  and m1 = basisNoOfRows uTreeL in
	  (lems @ List.map (fun (x, y, z, w) -> (x+m1, y, z, w)) rems,
	   Node (r, luTree, ruTree))

    let applyNsplit imkq d =
      List.split
	(List.map (fun (sr, m, k, q) ->
	  if k > 0 then (* something to do *)
	    let dm = d $ (sr <-> sr + m - 1, All) in
	    q @~* dm;
	    let d1 = d $ (sr <-> sr + k - 1, All)
	    and d2 = d $ (sr + k <-> sr + m - 1, All) in
	    (d1, d2)
	  else
	    (d $ (sr <-> sr - 1, All), d $ (sr <-> sr + m - 1, All)))
	   imkq)

    let rowMerge bs =
      match bs with
      | [] -> zeros 0 0
      | _ ->
	  let ms = List.map noOfRows bs in
	  let m = List.fold_left (+) 0 ms in
	  let d = rawMatrix m (noOfCols (List.hd bs)) in
	  let ds = partitionInfx1 ms d in
	  List.iter2 (<-$) ds bs;
	  d

    (* Need to re-write Q of LQ as I - VXV^T = I - UGU^T, U^TU = I. *)
    let lowRankFormOfQ k q =
      let (tau, vs) = reflectors q in
      let m = noOfRows vs
      and n = noOfCols vs in
      assert (m >= n);
      if m = 0 then
	(zeros k 0, zeros 0 0)
      else begin
	let (uv, rv) = qr' vs in
	let u = eye m n (* top-left of eye m m *)
	and g = zeros n n in
	uv @* u;
	let r = copyU rv in
	(muln (muln (r $@ (1,1)) (r $@ (1,1))) tau.{1} => (1,1)) g;
	for k = 1 to n-1 do
	  let gk = g $ (1 <-> k, 1 <-> k) in
	  let rkp1 = r $ (1 <-> k, k+1 <-> k+1) in
	  let tk = gk *~$ rkp1 -$ rkp1 in
	  scale' tau.{k+1} tk;
	  let alphakp1 = r $@ (k+1, k+1) in
	  let alphaTaukp1 = muln alphakp1 tau.{k+1} in
	  let gkp1 = g $ (1 <-> k+1, 1 <-> k+1) in
	  let (_, g12, g21, _) =
	    partition2x2 k 1
	      k
	      1  gkp1 in
	  gk  <-| (rkp1 |*~--| tk);
	  g12 <-| (iD rkp1);
	  scale' alphaTaukp1 g12;
	  g21 <-| transp' tk;
	  scale' (muln (fromInt (-1)) alphakp1) g21;
	  (muln alphakp1 alphaTaukp1 => (k+1, k+1)) gkp1;
	done;
	(u, g)
      end

    let applyTransform w vHss =
      let (wu, wg) = lowRankFormOfQ (Hss.no_of_rows vHss) w in
      let vHssTwu = rawMatrix (Hss.no_of_cols vHss) (noOfCols wu) in
      Hss.fastMulTransp vHss wu vHssTwu;
      let minusg = copy wg in
      scale' (fromInt (-1)) minusg;
      Hss.addLowRank vHss wu minusg vHssTwu

    let rec makeGtreeLeaves vHss g =
      match vHss with
      | Leaf _ -> Leaf g
      | Node (a, lHss, rHss) ->
	  let n1 = Hss.no_of_cols lHss
	  and n2 = Hss.no_of_cols rHss in
	  let (g1, g2) = partition2x1 n1 n2 g in
	  let gL = makeGtreeLeaves lHss g1
	  and gR = makeGtreeLeaves rHss g2 in
	  Node ((), gL, gR)

    let rec dropFirstRows k hss =
      let rec zeroAllms hss =
	match hss with
	| Leaf a -> Leaf {a with Hss.d = zeros 0 (noOfCols a.Hss.d);
			  Hss.u = zeros 0 (noOfCols a.Hss.u)}
	| Node (a, lHss, rHss) ->
	    Node ({a with Hss.ml = 0; Hss.mr = 0},
		  zeroAllms lHss,
		  zeroAllms rHss)
      in
      if k <= 0 then
	hss
      else
	match hss with
	| Leaf a ->
	    let m = noOfRows a.Hss.d in
	    assert (k <= m);
	    Leaf {Hss.d = a.Hss.d $ (k+1 <-> m, All);
		  Hss.u = a.Hss.u $ (k+1 <-> m, All);
		  Hss.v = a.Hss.v}
	| Node (a, lHss, rHss) ->
	    if k < a.Hss.ml then
	      Node ({a with Hss.ml = a.Hss.ml - k},
		    dropFirstRows k lHss,
		    rHss)
	    else
	      Node ({a with Hss.ml = 0; Hss.mr = a.Hss.mr - k + a.Hss.ml},
		    (zeroAllms lHss),
		    dropFirstRows (k - a.Hss.ml) rHss)

    let rec checkHssRows hss =
      match hss with
      | Leaf _ -> true
      | Node (a, lHss, rHss) ->
	  (a.Hss.ml = Hss.no_of_rows lHss) &&
	  (a.Hss.mr = Hss.no_of_rows rHss) &&
	  checkHssRows lHss && checkHssRows rHss

    let rec fillGtree gTree fmm =
      match gTree, fmm with
      | Leaf g, _ -> Leaf g
      | Node (_, gLtree, gRtree), Node (a, lFmm, rFmm) ->
	  let gL = fillGtree gLtree lFmm
	  and gR = fillGtree gRtree rFmm in
	  Node (a.wl *~$ getNodeData gL +$ a.wr *~$ getNodeData gR, gL, gR)
      | Node _, Leaf _ -> failwith "Bad g/fmm trees"

    (* Preference for g1's tree *)
    let rec add_g g1 g2 =
      match g1 with
      | Leaf g1 -> Leaf (g1 +$ getNodeData g2)
      | Node (g1, g1L, g1R) ->
	  match g2 with
	  | Leaf g2 -> Node (g1 +$ g2, g1L, g1R)
	  | Node (g2, g2L, g2R) -> Node (g1 +$ g2, add_g g1L g2L, add_g g1R g2R)

    type wChiLeaf = { w : transform; chi : matrix }
    and wChiNode =
	{ wN : transform; chiN : matrix; nm0 : int; nm1 : int }


    let fastSolve' fmm b x =
      let nrhs = noOfCols b in
      assert (nrhs = noOfCols x);
      let eliminateLeaf a b =
	if noOfCols a.u >= noOfRows a.u then (* Nothing to compress *)
	  (Leaf {w = eyeTransform; chi = zeros 0 nrhs},
	   Leaf (zeros (noOfCols a.v) nrhs),      (*           g tree *)
	   b,                                     (* remaining b      *)
	   Leaf a.u,                              (* remaining u tree *)
	   Leaf {Hss.d = a.v; Hss.u = zeros (noOfRows a.v) 0;
		 Hss.v = zeros (noOfCols a.v) 0}, (* remaining v hss  *)
	   a.d)                               (* remaining d      *)
	else (* Compress *)
	  let ku = noOfCols a.u
	  and m = noOfRows a.u
	  and n = noOfCols a.d 
	  and (q, l_u) = ql' a.u in
	  q @~* b;
	  q @~* a.d;
	  let (b0, b1) = partition2x1 (m-ku) ku b
	  and (d0, d1) = partition2x1 (m-ku) ku a.d in
	  let (d00, w) = lq' d0 in
	  d1 *~@ w;
	  w @* a.v;
	  subs' d00 b0;
	  let (d10, d11) = partition1x2 (m-ku) (n-m+ku) d1
	  and (v00, v01) = partition2x1 (m-ku) (n-m+ku) a.v
	  and chi = b0 in
	  (Leaf { w = w; chi = chi },
	   Leaf (v00 *~$ chi),                    (*           g tree *)
	   b1 -$ d10 *$ chi,                      (* remaining b      *)
	   Leaf (copyL l_u),                      (* remaining u tree *)
	   Leaf {Hss.d = v01; Hss.u = zeros (noOfRows v01) 0;
                 Hss.v = zeros (noOfCols v01) 0}, (* remaining v hss  *)
	   d11)                              (* remaining d      *)
      in
      let eliminateMergedLeaf u v d b =
	let (ems, l_u_tree) = compressU u in
	let (b1s, b2s) = applyNsplit ems b
	and (d1s, d2s) = applyNsplit ems d in
	let (d11, w) = lq' (rowMerge d1s) in
	let d2 = rowMerge d2s in
	d2 *~@ w;
	let wv = applyTransform w v in (* w @* v *)
	let b1 = rowMerge b1s
	and b2 = rowMerge b2s in
	subs' d11 b1;
	let chi = b1 in
	let k1 = noOfRows chi in
	let (d21, d22) = partition1x2 k1 (noOfCols d2 - k1) d2 in
	let xz = zeros (noOfCols d2) nrhs in
	(xz $ (1 <-> k1, All)) <-$ chi;
	let gs = rawMatrix (Hss.no_of_cols wv) nrhs in
	Hss.fastMulTransp wv xz gs;
	let wv12 = dropFirstRows k1 wv in
	assert (k1 = noOfRows (rowMerge d1s));
	({w=w; chi=chi}, makeGtreeLeaves wv12 gs, b2 -$ d21 *$ chi, l_u_tree,
	 wv12, d22)
      in
      let rec ulSweep fmm b f uMerge vMerge =
	match fmm with
	| Leaf a ->
	    let bnew =
	      match f with
	      | Leaf f -> b -$ a.u *$ f
	      | Node _ -> failwith "Bad f tree"
	    in
	    eliminateLeaf a bnew
	| Node (a, lFmm, rFmm) ->
	    let (b0, b1) = partition2x1 a.ml a.mr b in
	    let (foL, foR) =
	      match f with
	      | Leaf f -> (Leaf (a.rl *$ f), Leaf (a.rr *$ f))
	      | Node (f, fL, fR) -> (add_f (Leaf (mulOpt a.rl f nrhs)) fL,
				     add_f (Leaf (mulOpt a.rr f nrhs)) fR)
	    in
	    let (uMergeL, uMergeR) =
	      let uMergeL1 = uTreeOfB a.bu
	      and uMergeR1 = uTreeOfB a.bl in
	      match uMerge with
	      | Leaf _ -> (uMergeL1, uMergeR1)
	      | Node (_, uMergeL0, uMergeR0) ->
		  (mergeTrees uMergeL0 uMergeL1, mergeTrees uMergeR0 uMergeR1)
	    in
	    let (vMergeL, vMergeR) =
	      let vMergeL1 = vTreeOfB a.bl
	      and vMergeR1 = vTreeOfB a.bu in
	      match vMerge with
	      | Leaf _ -> (vMergeL1, vMergeR1)
	      | Node (_, vMergeL0, vMergeR0) ->
		  (mergeTrees vMergeL0 vMergeL1, mergeTrees vMergeR0 vMergeR1)
	    in
	    let (lWchiTree, gl, bl, ul, vl, dl) = ulSweep lFmm b0
		foL uMergeL vMergeL in
	    let (rWchiTree, gr, br, ur, vr, dr) = ulSweep rFmm b1
		(add_f (b_x_g a.bl gl) foR) uMergeR vMergeR in
	    let m0 = noOfRows dl
	    and n0 = noOfCols dl
	    and m1 = noOfRows dr
	    and n1 = noOfCols dr in
	    let d = rawMatrix (m0+m1) (n0+n1)
	    and bm = rawMatrix (m0+m1) nrhs in
	    matrix2x2   n0            n1
	           m0 (iD dl) (u_b_vt ul a.bu rFmm vr)
	           m1 (u_b_vt ur a.bl lFmm vl) (iD dr) d;
	    matrix2x1 m0 (bl |-| (u_f ul (b_x_g a.bu gr) nrhs))
	              m1 (iD br                               ) bm;
	    let uTree = pruneBasisTree (Node ((a.rl, a.rr), ul, ur)) uMerge in
	    let vTree = Hss.toProperForm' (pruneHssTree (mergeHss vl vr)
					     vMerge fmm) in
	    let (wCh, gTree, bmr, umr, vmr, dmr) =
	      eliminateMergedLeaf uTree vTree d bm in
	    let g = fillGtree gTree fmm in
	    (Node ({wN = wCh.w; chiN = wCh.chi; nm0 = n0; nm1 = n1},
		   lWchiTree, rWchiTree),
	     add_g g (Node (a.wl *~$ getNodeData gl +$ a.wr *~$ getNodeData gr,
			    gl, gr)),
	     bmr,
	     umr,
	     vmr,
	     dmr)
      in
      let rec vSweep fmm wChiTree xp x =
	match fmm, wChiTree with
	  (Leaf _, Leaf {w=w; chi=chi}) ->
	    matrix2x1 (noOfRows chi) (iD chi)
	              (noOfRows xp ) (iD xp ) x;
	    w @~* x
	| (Node (a, lFmm, rFmm),
	   Node ({wN=w;chiN=chi;nm0=n0;nm1=n1}, lWchi, rWchi)) ->
	     let mc = noOfRows chi
	     and mx = noOfRows xp in
	     let chip = rawMatrix (mc+mx) nrhs in
	     matrix2x1 mc (iD chi)
	               mx (iD xp ) chip;
	     w @~* chip;
	     let (xp0, xp1) = partition2x1 n0 n1 chip in
	     matrix2x1 a.nl (vSweep lFmm lWchi xp0)
	               a.nr (vSweep rFmm rWchi xp1) x
	| _ -> failwith "fastSolve': Mismatched trees."
      in
      match fmm with
	Leaf a ->
	  x <-$ b;
	  a.d |/| x
	| _ ->
	    let (wChiTree, _, _, _, _, _) = ulSweep fmm b
		(Leaf (zeros 0 nrhs)) (Leaf ()) (Leaf ()) in
	    vSweep fmm wChiTree (zeros 0 nrhs) x


  end (* module Fmmer (Blop) (Sparse) *)


module DFmm = Fmmer (DBlop) (DSparse) (DHss)
module SFmm = Fmmer (SBlop) (SSparse) (SHss)
module ZFmm = Fmmer (ZBlop) (ZSparse) (ZHss)
module CFmm = Fmmer (CBlop) (CSparse) (CHss)
