(*
  File: mInterpolation.ml
  Author: Shivkumar Chandrasekaran
  Date: 9.25.2007
  Purpose: Try multi-level-interval polynomial interpolation

  $Id: mInterpolation.ml,v 1.1 2007/09/27 04:59:06 shiv Exp shiv $
*)


open Bigarray
open Nla
open Chub


open DBlop
open DChub
  
      
(* From Dym and McKean *)
let smoothStep x =
  if x <= 0.0 then
    0.0
  else if x >= 1.0 then
    1.0
  else
    exp ( (-1.0 /. x) *. exp (1.0 /. (x -. 1.0)) )
      
      
(* 1 on [a',b']. support [a,b] *)
let sPulse a a' b' b x = smoothStep ((x -. a) /. (a' -. a)) *.
    smoothStep ((b -. x) /. (b -. b'))
    
    
type interval = { l : float; r : float }
type window = float -> float
type atom = { d : interval; fst : int; lst : int; w : window }


let breakupPoints a b xs p =
  let window = fun x -> 1.0 in
  let n = Array.length xs in
  assert (0 < p & p <= n);
  let m = n - p + 1 in
  Array.init m (fun i ->
    let l =
      if i = 0 then
	a
      else
	(xs.(i-1) +. xs.(i)) /. 2.0
    and r =
      if i = m-1 then
	b
      else
	(xs.(i+p-1) +. xs.(i+p)) /. 2.0
    in
    {d = {l = l; r = r}; fst = i; lst = i+p-1; w = window})
    

let setWindows xs molecule =
  let n = Array.length molecule in
  let psi = Array.init n (fun i ->
    let d = molecule.(i) in
    let a  = if i = 0   then 2.0 *. d.d.l -. xs.(d.fst) else d.d.l
    and a' = if i = 0   then d.d.l                      else xs.(d.fst)
    and b  = if i = n-1 then 2.0 *. d.d.r -. xs.(d.lst) else d.d.r
    and b' = if i = n-1 then d.d.r                      else xs.(d.lst) in
    sPulse a a' b' b)
  in
  let sum x =
    let acc = ref 0.0 in
    for i = 0 to n-1 do
      acc := !acc +. psi.(i) x
    done;
    !acc
  in
  for i=0 to n-1 do
    molecule.(i) <- {molecule.(i) with w = fun x -> (psi.(i) x) /. sum x}
  done
    

let makeWindowWeights xs d =
  let m = d.lst - d.fst + 1 in
  let w = Array1.create precision fortran_layout m in
  for i = 1 to m do
    w.{i} <- d.w xs.(d.fst+i-1)
  done;
  w


let makeSobolevScale ~sobolev ~preserve ~orderRate d =
  let l = d.d.r -. d.d.l
  and m = int_of_float (orderRate *. float (d.lst - d.fst + 1)) in
  let s = Array1.create precision fortran_layout m in
  if m > 0 then begin
    let preserve = min preserve m in
    if preserve >= 1 then
      s.{1} <- (2.0 ** 23.0) /. sqrt (l *. float m)
    else
      s.{1} <- 1.0 /. sqrt (l *. float m);
    for i = 2 to preserve do
      s.{i} <- (2.0 ** 23.0) /. (sqrt (l *. float m /. 2.0));
    done;
    for i = (max 1 preserve) + 1 to m do
      s.{i} <- 1.0 /. (sqrt (l *. float m /. 2.0) *. (float (i-1) ** sobolev))
    done;
  end;
  s


(* Both of these are inefficient right now. *)
let bigArrayFromList xs =
  Array1.of_array precision fortran_layout (Array.of_list xs)
    
let bigArrayToList s =
  Array.to_list (Array.init (Array1.dim s) (fun i -> s.{i+1}) )


let makeConstraints ~orderRate ~preserve ~sobolev xs molecule =
  let k = Array.length molecule in
  let m = Array.length xs in
  let order d = int_of_float (orderRate *. float (d.lst - d.fst + 1)) in
  let n = Array.fold_left (fun acc d -> acc + order d) 0 molecule in
  let v = zeros m n
  and sList = ref []
  and ci = ref 1 in
  for i = 0 to k-1 do
    let di = molecule.(i) in
    let mi = di.lst - di.fst + 1 in
    let ni = order di in
    let zi = Array.sub xs di.fst mi in
    fromInterval' di.d.l di.d.r zi;
    let vi = chebVand zi ni in
    let wi = makeWindowWeights xs di
    and si = makeSobolevScale ~sobolev ~preserve ~orderRate di in
    colScale' vi si;
    rowScale' wi vi;
    (v $ (di.fst+1 <-> di.lst+1, !ci <-> !ci + ni - 1)) <-$ vi;
    sList := !sList @ bigArrayToList si;
    ci := !ci + ni
  done;
  Printf.printf "cond = %g %i %i\n" (cond v) m n;
  (v, bigArrayFromList !sList)


let pwVand orderRate molecule c x =
  let k = Array.length molecule in
  let order d = int_of_float (orderRate *. float (d.lst - d.fst + 1)) in
  let acc = ref zero
  and ni = ref 1 in
  for i = 0 to k-1 do
    let di = molecule.(i) in
    let n = order di in
    if di.d.l <= x & x <= di.d.r then begin
      acc := addn !acc
	  (di.w x *. ((chebVand [|fromInterval di.d.l di.d.r x|] n *$
		       (c $ (!ni <-> !ni+n-1, All))) $@ (1,1)))
    end;
    ni := !ni + n
  done;
  !acc


let interpolateOL ?(orderRate=1.5) ?(preserve=2) ?(sobolev=2.0) p f a b xs =
  let m = Array.length xs in
  let molecule = breakupPoints a b xs p in
  setWindows xs molecule;
  let (v, s) = makeConstraints ~orderRate ~preserve ~sobolev xs molecule in
  let r = fun2mat (fun i _ -> f xs.(i-1)) m 1 in
  let c = v /$ r in
  rowScale' s c;
  let p x = pwVand orderRate molecule c x in
  (p, molecule, c)
