(*
  File: nestedDissection.ml
  Author: Shivkumar Chandrasekaran
  Date: 9.11.2003
  Email: shiv@ece.ucsb.edu
  Package: CamlFloat
  Purpose: convert points in R^n to HSS trees efficiently.

  $Id: nestedDissection.ml,v 1.1 2005/11/03 17:29:44 shiv Exp shiv $
*)


(* Points in R^n are tuples. We do not assume any ordering as that
   will not help if n > 1. *)

open Utils

(* [a0; a1; ...] -> [(n,a0); (n+1,a1); ...] *)
let rec index_list n xs =
  match xs with
      [] -> []
    | x :: rs -> (n, x) :: index_list (n+1) rs

let dissect1D xs a b =
  let rec dis xis a b =
    if (b <= a) then Leaf []
    else
      match xis with
	  [] -> Leaf []
	| [(i,x)] -> assert (a <= x && x <= b); Leaf [i]
	| _ ->
	    let m = (a +. b) /. 2.0 in
	    let (xl, xr) = List.partition (fun (_, x) -> x <= m) xis in
	    Node ((List.length xl, List.length xr), dis xl a m, dis xr m b)
  in
  dis (index_list 1 xs) a b

(* width must be a fraction: 0 <= width <= 1 *)
let dissect1DWithSeparators ~width xs a b =
  assert (0.0 <= width && width <= 1.0);
  let rec dis xis a b =
    if b <= a then Leaf []
    else match xis with
    | [] -> Leaf []
    | [(i,x)] -> assert (a <= x && x <= b); Leaf [i]
    | _ ->
	let m = (a +. b) /. 2.0 in
	let w = width *. (b -. a) /. 2.0 in
	let ml = m -. w
	and mr = m +. w in
	assert (a <= ml && ml <= mr && mr <= b);
	let (xl, xmr) = List.partition (fun (_,x) -> x <= ml) xis in
	let (xm,  xr) = List.partition (fun (_,x) -> x <= mr) xmr in
	let left = dis xl a ml
	and right = dis xr mr b in
	Node ((List.length xl + List.length xr, List.length xm),
	      Node ((List.length xl, List.length xr), left, right),
	      dis xm ml mr)
  in
  dis (index_list 1 xs) a b

let dissectND xs a b =
  let dim = Array.length a in
  assert (Array.length b = dim);
  let rec dis xis a b =
    let c = Array.init dim (fun i -> b.(i) -. a.(i)) in
    let minc = Array.fold_left (fun m ci -> min m ci) 1.0 c in
    if minc <= 0.0 then Leaf []
    else match xis with
	[] -> Leaf []
      | [(i,x)] -> assert (a <= x && x <= b); Leaf [i]
      | _ ->
	  let k =
	    let mm = ref 0.0
	    and k = ref (-1) in
	    let maxpos i ci = if !mm < ci then begin
	      mm := ci; k := i
	    end in
	    Array.iteri maxpos c;
	    !k
	  in
	  let m = (a.(k) +. b.(k)) /. 2.0 in
	  let ma = Array.copy b
	  and mb = Array.copy a 
	  and (xl, xr) = List.partition (fun (_, x) -> x.(k) <= m) xis
	  in
	  ma.(k) <- m;
	  mb.(k) <- m;
	  Node ((List.length xl, List.length xr), dis xl a ma, dis xr
	    mb b)
  in
  dis (index_list 1 xs) a b

let rec breadthFirstLeafData tree =
  match tree with
      Leaf a -> a
    | Node (_, ltree, rtree) -> breadthFirstLeafData ltree @
	breadthFirstLeafData rtree

let rec prune_tree hss_spine p =
  match hss_spine with
      Leaf _ -> hss_spine
    | Node ((m1,m2), lSpine, rSpine) ->
	if m1 + m2 <= p then Leaf (breadthFirstLeafData hss_spine)
	else Node ((m1,m2), prune_tree lSpine p, prune_tree rSpine p)

(* For vector equations I need to change the final indexes. *)
let rec gen first last =
  if last < first then []
  else first :: gen (first+1) last

(* [3, 4, 1, 2] -> [2p+1, ..., 3p, 3p+1, ..., 4p, 1, ..., p, p+1, ...,
   2p] *)
let rec inflate_indices ns p =
  match ns with
      [] -> []
    | n :: nrs -> gen ((n-1)*p+1) (n*p) @ inflate_indices nrs p

let rec inflate_partition_tree tree p =
  match tree with
      Leaf a -> Leaf (inflate_indices a p)
    | Node ((m1,m2), ltree, rtree) ->
	Node ((m1*p, m2*p), inflate_partition_tree ltree p,
	inflate_partition_tree rtree p)
