(*
  File: nla.mli
  Author: Shivkumar Chandrasekaran
  Email: shiv@ece.ucsb.edu
  Date: 3.26.2003
  Purpose: Uniform interface to single/double Lapack and Blas

  $Id: nla.mli,v 1.8 2008/01/19 03:15:16 shiv Exp shiv $
*)

(** {1 Numerical linear algebra routines.}

  This module provides a stylized high-level interface to the Fortran
  library Lapack and its associated C library CBLAS. This interface
  tries to make numerical linear algebra programs look like they were
  written in the conventional linear algebra mathematics notation,
  without sacrificing speed or memory efficiency. It also tries to
  insulate the user from the convoluted and bug-prone interface
  provided by Lapack. For that we depend to a great extent on the
  static type checking of OCaml, and, to a lesser extent dynamic
  runtime type checking. The design borrows a lot from Matlab, but
  tries to stick to the OCaml way of doing things. Some parts might
  take some getting used to. The library has been used extensively to
  code, test and time my own research algorithms. So it is
  battle-hardened to some extent. User feedback is welcome. You can
  catch me at shiv\@ece.ucsb.edu.

  There is a lower-level interface that mimics the exact calling
  interface of Lapack and CBLAS, but that is not documented here. In
  fact that interface will hopefully become Lacaml at some point. A
  tutorial to help familiarize yourself with this library is also
  available. This package is very similar to the library CleanFloat
  for the Clean functional programming language.

  Of course efficiency depends a great deal on controlling memory
  usages. So CamlFloat tries to be faithful to the underlying
  philosophy of Lapack. Almost all computations destroy incoming
  matrices (two-dimensional arrays in essence). This is quite unlike
  Matlab where almost always copies are made internally. This is
  rather nice for newbies, but rather useless for bigger
  projects. However, to warn users of potential side-effects, we try
  to follow one of two conventions. If a function destroys any
  incoming date it will either have a [unit] return type, or, it will
  have a name that is terminated by a single open quote. Both of these
  conventions are common in the OCaml-world.

  CamlFloat also allows you to write generic code that is capable of
  being executed in real single, real double, complex single and
  complex double precision arithmetic. In standard OCaml form we
  achieve this using functors. To help in this regard all routines
  are placed in four modules that share the common signature
  [BLOP]. Examples of such generic functors can be found by digging
  around in the source files. Interestingly enough the package
  CleanFloat achieves the same effect by using Clean's type classes.


  @author Shivkumar Chandrasekaran
  @author William Lyons

*)

(** A signature that is common to real and complex, single and double
  precision routines.

  This signature is supposed to facilitate generic
  programming: write one code and execute it in four different
  ways. [BLOP] itself is an acronym for Basic Linear OPerators.

  There are four modules [DBlop], [SBlop], [ZBlop] and [CBlop] that
  have this signature. The common approach is to write your code as a
  functor that requires a [BLOP] argument, and then to instantiate
  your code in four different ways with [DBlop], [SBlop], [ZBlop] and
  [CBlop].

  To facilitate this, the [BLOP] signature also has a rudimentary set
  of generic arithmetic operations. In general calling such generic
  arithmetic operators in a tight [for] loop is not efficient in the
  current verions of OCaml, since functor calls are not expanded at
  compile-time. One way around this is to encapsulate such code in
  separate modules and pass them into your functor. If you feel that
  something should be added to [BLOP] that would alleviate this
  problem please let me know. Interestingly enough this problem can be
  solved manually in Clean by using macros and instantiating type
  classes. Ugly, but it works.
*)
module type BLOP =
sig
  (** Matrices use two-dimensional [Bigarray] arrays
    internally. Unfortunately these are not powerful enough to serve as
    matrices because we cannot access arbitrary, but contiguous,
    sub-matrices, without a lot of trickery. Psilab got around this by
    changing the bigarray code. Unfortunately that change never made it
    into the main OCaml distribution. Rather than attempt to change the
    [Bigarray] implementation again I decided to wrap a new [matrix]
    type around the existing [Bigarray] implementation. This has some
    drawbacks. For one, element-by-element operations on [matrix] types
    could be slower than equivalent [Bigarray] operations, especially
    when functors prevent the OCaml compiler from inlining the access
    calls. However, this can be overcome by using the function [matrix]
    to efficiently convert native OCaml arrays to matrices, and
    performing all element by element operations on native OCaml arrays
    instead. *)
  open Bigarray

  (** The type [t] is either [float] in modules [DBlop] and [SBlop],
    or [Complex.t] in modules [ZBlop] and [CBlop].*)
  type t

  (** The internal type that is used to represent the elements of the
    matrix. It is [float64_elt] in [DBlop], [float32_elt] in [SBlop],
    [complex64_elt] in [ZBlop] and [complex32_elt] in [CBlop].*)
  type precision_elt

  (** The internal type used for storing arrays of real numbers, like
    singular values, which are real even for complex matrices. It is
    [float64_elt] in modules [DBlop] and [ZBlop], and [float32_elt] in
    modules [SBlop] and [CBlop]. *)
  type precision_elt_real

  (** The mechanism for specifying the type of array you want from
    [Bigarray]. The usual usage is [Array2.create precision
    fortran_layout m n]. *)
  val precision : (t, precision_elt) kind

  (** Same as above except when you need a real array even inside a
    complex module. *)
  val precision_real : (float, precision_elt_real) kind

  (** The machine precision. It is approximately [10^{-16}] in the
    [DBlop] and [ZBlop] modules, and approximately [10^{-8}] in the
    [SBlop] and [CBlop] modules. *)
  val machine_epsilon : float

  (** The next few functions are generic arithmetic functions which
    should only be used in non-critical pieces of your code. They are
    meant to be used inside functors which are meant to work for both
    the real and complex modules at the same time. Much of my research
    work is implemented in this fashion and I have not had a problem.
  *)
 (** [re z] returns the real part of the (possibly) complex number
   [z].*)
  val re : t -> float

  (** [im z] returns the imaginary part of the (possibly) complex
    number [z]. *)
  val im : t -> float

  (** [conjugate z] returns the complex conjugate of the (possibly)
    complex number [z]. *)
  val conjugate : t -> t

  (** Either the real or complex zero. *)
  val zero : t

  (** Either the real or complex one. *)
  val one : t

  (** Either the real or complex minus one. *)
  val minus_one : t

  (** [fromInt n] returns either the floating-point version of [n] or
    the [Complex.t] version of [n]. *)
  val fromInt : int -> t

  (** [toInt z] returns the integer part of the real part of
    (possibly) complex number [z]. *)
  val toInt : t -> int

  (** Converts a floating point number to floating point number or
    [Complex.t] as the case may be. *)
  val fromFloat : float -> t

  (** [randn ()] returns a random possibly complex number. *)
  val randn : unit -> t

  (** Prints a possibly complex number to [stdout]. *)
  val printfn : t -> unit

  (** Converts a possibly complex number to string. *)
  val sprintfn : t -> string

  (** Adds two possibly complex numbers. *)
  val addn : t -> t -> t

  (** Subtracts two possibly complex numebrs. *)
  val subn : t -> t -> t

  (** Multiplies two possibly complex numbers. *)
  val muln : t -> t -> t

  (** [mulr r z] multiplies the possibly complex number [z] by the
    real number [r]. *)
  val mulr : float -> t -> t
    (** With that we are done with generic arithmetic
      operators. Please note that [t] is not a union type. In any
      instance [t] is either [float] or [Complex.t]. For example inside
      module [DBlop] the type [t] always refers to [float]. Now, on to
      the matrix stuff. *)

  (** The basic data type. It is a wrapper for [Array2.t]. *)
  type matrix

  (** CamlFloat encourages a style of programming I call {b storage
    passing}. Such a style is more useful in a pure language like Clean
    (see CleanFloat). However it continues to have its merits in
    OCaml. The basic idea is that we frequently need to modify the
    contents of a matrix. The type [transformer] encapsulates the most
    common type of such functions. Furthermore most algorithms can be
    implemented in a manner in which the destination matrix is the
    last argument. This enables one to make the code look like the
    mathematical specification of the algorithm (not the Matlab
    specification). See the tutorial. *)
  type transformer = matrix -> unit

  (** Usually an object of type [transform] (not [transformer]) is a
    unitary matrix computed by a call to one of the unitary
    factorization routines: {!Nla.BLOP.ql'} or {!Nla.BLOP.lq'}. It can
    also be a product of such unitary transforms. *)
  type transform

  (** An object of type [factor] is usually the triangular part from
    a QL or LQ factorization. This type alone will save you from most
    of the bugs encountered when programming on top of Lapack. *)
  type factor

  (** [a $@ (i,j)] returns the [(i,j)]-th entry of the matrix [a]. *)
  val ( $@ ) : matrix -> int * int -> t

  (** The weird syntax [(x => (i,j)) a] has the effect of making [a $@
    (i,j)] be equal to [x]. This weird syntax is part of the {b
    storage passing} style I talked about earler, and its use will
    become clear soon. Meanwhile, do not use these element-by-element
    operations in critical [for] loops. You might take a speed-hit. *)
  val ( => ) : t -> int * int -> matrix -> unit

  (** Returns the number of rows in the matrix. *)
  val noOfRows : matrix -> int

  (** Returns the number of columns in the matrix. *)
  val noOfCols : matrix -> int

  (** Determine the number of rows in a matrix that will be printed by
    [printMatrix]. *)
  val printMatrixHeight : int ref

  (** Determines the number of columns of a matrix that will be
    printed by [printMatrix]. *)
  val printMatrixWidth : int ref

  (** Prints a matrix to [stdout]. *)
  val printMatrix : matrix -> unit

  (** [rawMatrix m n] creates an [m] by [n] matrix whose entries are
     uninitialized. Setting [~ooc:true] will create an out-of-core matrix. *)
  val rawMatrix : ?ooc:bool -> int -> int -> matrix

  (** [consts x m n] creates an [m] by [n] matrix with every entry set
    to [x]. See also [const]. Setting [~ooc:true] will create an out-of-core
     matrix. *)
  val consts : ?ooc:bool -> t -> int -> int -> matrix

  (** [const x a] sets every entry of the matrix [a] to [x]. See also
    [consts]. *)
  val const : t -> matrix -> unit

  (** [zeros m n] creates a [m] by [n] zero matrix. *)
  val zeros : ?ooc:bool -> int -> int -> matrix

  (** [ones m n] creates a [m] by [n] matrix of ones. *)
  val ones : ?ooc:bool -> int -> int -> matrix

  (** [randomMatrix m n] creates a [m] by [n] random matrix whose
    entries (real and imaginary parts) are uniformly distributed
    between -1.0 and 1.0. *)
  val randomMatrix : ?ooc:bool -> int -> int -> matrix

  (** [fun2mat fn m n] creates an [m] by [n] matrix whose [(i,j)]-th
    entry is [fn i j]. See also [fun2mat']. *)
  val fun2mat : ?ooc:bool -> (int -> int -> t) -> int -> int -> matrix

  (** [fun2mat' fn a] modifies all the entries of the matrix [a] such
    that the [(i,j)]-th entry becomes [fn i j]. See also [fun2mat]. *)
  val fun2mat' : (int -> int -> t) -> matrix -> unit

  (** [eye m n] returns the top-left [m] by [n] submatrix of the [max
    m n] by [max m n] identity matrix. *)
  val eye : ?ooc:bool -> int -> int -> matrix

  (** Efficiently converts a standard two-dimensional OCaml array into
    a [matrix]. Element-by-element operations are better done on
    regular OCaml arrays. *)
  val matrix : t array array -> matrix

  (** Map a matrix to a file on disk. More memory efficient if you are not
     going to access majority of the elements. Note, this is not a way to
     do out-of-core algorithms. *)
  val mmapdMatrix : ?shared:bool -> string -> int -> int -> matrix

  (** Returns true if the matrix is out-of-core. You obtain such matrices
     by using setting the optional parameter [~ooc:true] in the matrix
     creation routines. *)
  val is_ooc : matrix -> bool

  (** Checks if the bigarray is in-core memory. *)
  val is_in : matrix -> bool

  (** Converts a matrix to out-of-core. *)
  val to_ooc : matrix -> matrix

  (** Converts a matrix to in-core. *)
  val from_ooc : matrix -> matrix

  (** An intermediate data type for creating sub-matrices. See [$]. *)
  type range =
      Range of int * int
    | All

  (** This operator is used to pick out sub-matrices. See [<->]. *)
  val ( $ ) : matrix -> range * range -> matrix

  (** Equivalent to Matlab's colon operator. [a $ (sr <-> lr, sc <->
    lc)] returns the contiguous submatrix of a that includes rows [sr]
    to [lr], and columns [sc] to [lc]. In Matlab you would have
    written this as [a(sr:lr,sc:lc)]. If you write instead [a $ (All,
    sc <-> lc)] then this would be the submatrix of [a] that includes
    all the rows of [a], and the columns from [sc] to
    [lc]. In Matlab this would have been [a(::,sc:lc)]. However, there
    is {b one big difference} from Matlab. [a $ (sr <-> lr, sc <->
    lc)] and [a] share the same storage! Please be careful. The type
    system will not keep track of this sharing for you. However, in
    CleanFloat the uniqueness type attribute will keep track of
    sharing. *)
  val ( <-> ) : int -> int -> range

  (** In CamlFloat I strongly discourage you from programming as if
    you were in Matlab. Rather I strongly encourage you to program as
    if you were specifying the algorithm in standard mathematical
    notation. For example, matrix block partitioning is one of the key
    concepts in matrix analysis. However, in Matlab this can only be
    achieved by specifying multiple sub-matrices, all of which will be
    copied out. However, in CamlFloat the call [partition2x1 m1 m2 a]
    will return the tuple [(a1, a2)], where, in Matlab notation, [a1]
    is equal to [a(1:m1,::)] and [a2] is equal to
    [a(m1+1:m1+m2,::)]. Note the index arithmetic that you must
    perform correctly in Matlab. Furthermore, note that in CamlFloat
    [a1] and [a2] share their storage with [a], unlike Matlab. *)
  val partition2x1 : int -> int -> matrix -> matrix * matrix

  (** [partition1x2 n1 n2 a] returns the tuple [(a1,a2)] obtained by
    column partitioning [a], with first block column [a1] having the first
    [n1] columns, and the second block column [a2] having the last
    [n2] columns. Note that [a1] and [a2] share storage with [a]. *)
  val partition1x2 : int -> int -> matrix -> matrix * matrix

  (** [partition2x2 n1 n2 m1 m2 a] returns the tuple [(a11, a12, a21,
    a22)] obtained by block partitioning [a] into a 2 by 2 matrix,
    with [n1] columns in the first column partition and [m1] rows in
    the first row partition. I usually format the call as follows
    {[ partition2x2 n1 n2
           m1
           m2     a
    ]} so that it looks more like the mathematical notation. *)
  val partition2x2 : int -> int -> int -> int -> matrix ->
    matrix * matrix * matrix * matrix

  (** [partition1x3 n1 n2 n3 a] returns [(a1,a2,a3)] obtained by
    doing a 1 by 3 block column partition of [a]. *)
  val partition1x3 : int -> int -> int -> matrix -> matrix * matrix *
    matrix

  (** [partition3x1 m1 m2 m3 a] returns [(a1,a2,a3)] obtained by doing
    a 3 by 1 block row partition of [a]. *)
  val partition3x1 : int -> int -> int -> matrix -> matrix * matrix *
    matrix

  (** [partition3x3 n1 n2 n3 m1 m2 m3 a] returns
    [(a11,a12,a13,a21,a22,a23,a31,a32,a33)] obtained by doing a 3 by 3
    block partitioning of [a]. I usually format the call as
    {[ partition3x3 n1 n2 n3
           m1
           m2
           m3        a
    ]} so that it looks more like the mathematical notation. *)
  val partition3x3 :
    int -> int -> int -> int -> int -> int -> matrix ->
    matrix * matrix * matrix *
    matrix * matrix * matrix *
    matrix * matrix * matrix

  (** Sometimes we cannot predict the order of the block
    partitions. To aid in this situation we have the following
    functions. *)
  (** [partitionInfx1 ms a] returns a list of block row
    partitions of [a], where the number of rows in the [i]-th
    partition is determined by the number in the [i]-th position in
    the list [ms]. *)
  val partitionInfx1 : int list -> matrix -> matrix list

  (** [partition1xInf ns a] returns a list of block column partitions
    of [a], where the number of columns in the [j]-th partition is
    determined by the number in the [j]-th position in the list [ns]. *)
  val partition1xInf : int list -> matrix -> matrix list

  (** [partitionNx1 ms a] is identical to [partitionInfx1 ms a],
    except tha [ms] must be an array now. *)
  val partitionNx1 : int array -> matrix -> matrix array

  (** [partition1xN ns a] is identical to [partition1xInf ns a],
    except that [ns] must be an array now. *)
  val partition1xN : int array -> matrix -> matrix array

  (** We now present a set of functions that are most useful in the
    storage-passing style of programming. *)
  (** The function [ooo] is a nop. Very useful for subsequent
    functions. *)
  val ooo : matrix -> unit

  (** [matrix2x1 m1 a1 m2 a2 a] modifies the entries of the matrix
    [a], by applying the function ([transformer]) [a1] to [a $ (1<->
    m1, All)] and the [transformer] [a2] to [a $ (m1+1 <-> m1+m1,
    All)]. I usually format the call as
    {v matrix2x1 m1 a1
          m2 a2 a
    v} so that it looks more like the mathematical notation. If you
    use a higher-order style of programming you can even drop the last
    argument, the destination matrix [a], from the call, and make it
    look even more like the mathematical notation. See the tutorial
    for a good example of this style. Again observe how this call
    saves you from index gymnastics. *)
  val matrix2x1 :
    int -> transformer -> int -> transformer -> matrix -> unit

  (** [matrix3x1 m1 a1 m2 a2 m3 a3 a] modifies the entries of [a] by
    first applying the [transformer] [a1] to the first block row of the 3 by
    1 block row partitioning of [a], then applying the [transformer]
    [a2] to the second block row partition of [a], and so on. I
    usually format the call a bit more suggestively as
    {[ matrix3x1 m1 a1
           m2 a2
           m3 a3 a
    ]} so that it looks more like the mathematical notation.
  *)
  val matrix3x1 : int -> transformer -> int -> transformer ->
    int -> transformer -> matrix -> unit

  (** [matrix1x2 n1 n2 a1 a2 a] modifies the entries of [a] by
    applying the two [transformer]s [a1] and [a2] to the two block
    column partitions of [a]. Please observe that [matrix1x2] has a
    very different type from [matrix2x1]. This is so that I can format
    the call as
    {[ matrix1x2 n1 n2
           a1 a2 a
    ]} so that it looks more like the mathematical notation.
  *)
  val matrix1x2 :
    int -> int -> transformer -> transformer -> matrix -> unit

  (** [matrix1x3 n1 n2 n3 a1 a2 a3 a] modifies the entries of the
    matrix [a] by applying the [transformer]s [a1], [a2] and [a3] to
    the corresponding block column partitions of [a]. Note that
    [matrix1x3] has a different type from [matrix3x1]. This is so that
    I can format the call as
    {[ matrix1x3 n1 n2 n3
           a1 a2 a3 a
    ]} so that it looks more like the mathematical notation.
  *)
  val matrix1x3 : int -> int -> int ->
    transformer -> transformer -> transformer -> matrix -> unit

  (** [matrix2x2 n1 n2 m1 a11 a12 m2 a21 a22 a] modifies the entries
    of the [matrix] [a] by applying the [transformer]s [a11], [a12],
    [a21] and [a22] to the corresponding sub-matrices of [a] obtained
    by a block 2 by 2 partitioning. I usually format the call as
    {[ matrix2x2 n1  n2
        m1 a11 a12
        m2 a21 a22 a
    ]} so that it looks more like the mathematical notation.*)
  val matrix2x2 : int -> int ->
    int -> transformer -> transformer ->
    int -> transformer -> transformer -> matrix -> unit

  (** [matrix2x3 n1 n2 n3 m1 a11 a12 a13 m2 a21 a22 a23 a] modifies
    the entries of the [matrix] [a] by applying the [transformer]s
    [a11], [a12], [a13], [a21], [a22] and [a23] to the corresponding
    sub-matrices of
    [a] obtained by a block 2 by 3 partitioning. I usually format the
    call as
    {[ matrix2x3 n1  n2  n3
        m1 a11 a12 a13
        m2 a21 a22 a23 a
    ]} so that it looks more like the mathematical notation.*)
  val matrix2x3 : int -> int -> int ->
    int -> transformer -> transformer -> transformer ->
    int -> transformer -> transformer -> transformer -> matrix -> unit

  (** [matrix4x2 n1 n2 m1 a11 a12 m2 a21 a22 m3 a31 a32 m4 a41 a42 a]
    modifies the entries
    of the [matrix] [a] by applying the [transformer]s [aij] to
    the corresponding sub-matrices of [a] obtained
    by a block 4 by 2 partitioning. I usually format the call as
    {[ matrix4x2 n1  n2
        m1 a11 a12
        m2 a21 a22
        m3 a31 a32
        m4 a41 a42 a
    ]} so that it looks more like the mathematical notation.*)
  val matrix4x2 : int -> int ->
    int -> transformer -> transformer ->
    int -> transformer -> transformer ->
    int -> transformer -> transformer ->
    int -> transformer -> transformer -> matrix -> unit

  (** [matrixInfx1 ms ais a] applies the [transformer]s in the list
    [ais] to the corresponding row partitions of [a] as determined by
    the block sizes in the list [ms]. *)
  val matrixInfx1 : int list -> transformer list -> matrix -> unit

  (** [matrixNx1 ms ais a] applies the [transformer]s in the array
    [ais] to the corresponding row partitions of [a] as determined by
    the block sizes in the array [ms]. *)
  val matrixNx1 : int array -> transformer array -> matrix -> unit

  (** Since the [$] operator produces sub-matrices that share storage
    with the parent matrix, we need a way to make copies of
    matrices. [copy a] produces a copy of the [matrix] [a] such that
    they do not share storage. *)
  val copy : ?ooc:bool -> matrix -> matrix

  (** [iD a b] replaces the contents of the [matrix] [b] with that of
    [a]. *)
  val iD : matrix -> matrix -> unit

  (** [(a $=> (sr<->lr,sc<->lc)) b] will replace the contents of [b $
    (sr<->lc,sc<->lc)] with that of [a]. *)
  val ( $=> ) : matrix -> range * range -> matrix -> unit

  (** People who do not like to write the destination storage on the
    right-hand side, need this operator [<-|]. Therefore instead of
    [iD a b] you can say [b <-| (iD a)]. This is not needed, however,
    if you use the [matrixXxY] functions. *)
  val ( <-| ) : matrix -> transformer -> unit

  (** People who do not like to write [(a $=> (sr<->lr,All)) b] can
    now write [(b $ (sr<->lr,All)) <-$ a] instead. In other words, if
    you don't like my notation, define your own! After all this is
    OCaml. *)
  val ( <-$ )  : matrix -> matrix -> unit

  (** We will now specify the basic matrix arithmetic operators. You
    will find the Matlab equivalents here. However it is probably more
    important to realize that you can find the BLAS3 equivalents here
    too. I have tried to wrap the BLAS3 calls to facilitate the
    storage-passing style. That is the destination matrix is always
    the last argument, so by currying these operators you can embed
    them inside [matrix2x2] for example. Some of the operator notation
    might take some getting used to. I find them logical and
    convenient. If you disagree just define your own notation. *)
  (** [(a |+| b) c] will replace the contents of [c] with that of [a +
    b]. *)
  val ( |+| ) : matrix -> matrix -> matrix -> unit

  (** [(a |-| b) c] will replace the contents of [c] with that of [a -
    b]. *)
  val ( |-| ) : matrix -> matrix -> matrix -> unit

  (** [a |++| b] will replace the contents of [b] with that of [a + b]. *)
  val ( |++| ) : matrix -> matrix -> unit

  (** [a |--| b] will replace the contents of [b] with that of [b - a]. *)
  val ( |--| ) : matrix -> matrix -> unit

  (** [a +$ b] returns a new [matrix] whose entries are those of [a +
    b]. *)
  val ( +$ ) : matrix -> matrix -> matrix

  (** [a -$ b] returns a new [matrix] whose entries are those of [a -
    b]. *)
  val ( -$ ) : matrix -> matrix -> matrix

  (** [transp a] returns a new matrix formed by conjugate-transposing
    [a]. *)
  val transp : ?ooc:bool -> matrix -> matrix

  (** [transp' a b] replaces the contents of [b] with the
    conjugate-transpose of [a]. *)
  val transp' : matrix -> matrix -> unit

  (** [conjg' a] replaces the entries of [a] by their conjugates. *)
  val conjg' : matrix -> unit

  (** [(a |*| b) c] will replace the contents of [c] with that of [a *
    b]. *)
  val ( |*| ) : matrix -> matrix -> matrix -> unit

  (** [(a |*~| b) c] will replace the contents of [c] with that of [a *$
    transp b]. Note that we use [~] to denote transposition and that
    it is attached to the operator rather than the argument. *)
  val ( |*~| ) : matrix -> matrix -> matrix -> unit

  (** [(a |~*| b) c] will replace the contents of [c] with that of
    [transp a *$ b]. Note that we use [~] to denote transposition and that
    it is attached to the operator rather than the argument. Also note
    that the position of [~] tells you to which argument the transpose
    is attached. *)
  val ( |~*| ) : matrix -> matrix -> matrix -> unit

  (** [a *$ b] will create a new matrix with the contents of [a * b]. *)
  val ( *$ ) : matrix -> matrix -> matrix

  (** [a *~$ b] will create a new matrix with the contents of [transp
    a *$ b]. *)
  val ( *~$ ) : matrix -> matrix -> matrix

  (** [a *$~ b] will create a new matrix with the contents of [a *$
    transp b]. *)
  val ( *$~ ) : matrix -> matrix -> matrix

  (** [(a |*++| b) c] will replace the contents of [c] with that of [c + a * b]. *)
  val ( |*++| ) : matrix -> matrix -> matrix -> unit

  (** [(a |*--| b) c] will replace the contents of [c] with that of [c - a * b]. *)
  val ( |*--| ) : matrix -> matrix -> matrix -> unit

  (** [(a |*~++| b) c] will replace the contents of [c] with that of
    [c + a * transp b]. *)
  val ( |*~++| ) : matrix -> matrix -> matrix -> unit

  (** [(a |*~--| b) c] will replace the contents of [c] with that of [c - a *
      transp b]. *)
  val ( |*~--| ) : matrix -> matrix -> matrix -> unit

  (** [(a |~*++| b) c] will replace the contents of [c] with that of
    [c + transp a * b]. *)
  val ( |~*++| ) : matrix -> matrix -> matrix -> unit

  (** [(a |~*--| b) c] will replace the contents of [c] with that of
    [c - transp a * b]. *)
  val ( |~*--| ) : matrix -> matrix -> matrix -> unit

  (** [norm1 a] returns the 1-norm of [a]. *)
  val norm1 : matrix -> float

  (** [normInf a] returns the infinity-norm of [a]. *)
  val normInf : matrix -> float

  (** [normF a] returns the Frobenius-norm of [a]. *)
  val normF : matrix -> float

  (** [normMax a] returns the largest entry of [a] in magnitude. *)
  val normMax : matrix -> float

  (** We now move on to the Lapack routines. Currently this mainly
    includes interfaces to the orthogonal factorization and
    (non-square) linear system solver routines. As need arises more
    can be added. Let me know. *)

  (** The identity matrix as a [transform] type. *)
  val eyeTransform : transform

  (** [lq' a] computes the LQ factorization of [a]. Note that [a] is
    destroyed and its internal storage re-used for the returned [l]
    and [q]. Any extant sub-matrices of [a] will be rendered
    useless. You have been warned. *)
  val lq' : matrix -> factor * transform

  (** [ql' a] computes the QL factorization of [a]. The contents of
    [a] are destroyed and its storage is used to hold [q] and [l]. *)
  val ql' : matrix -> transform * factor

  (** [qr' a] computes the QR factorization of [a]. The contents of
    [a] are destroyed and its storage is used to hold [q] and [r]. *)
  val qr' : matrix -> transform * factor

  (** [rq' a] computes the RQ factorization of [a]. The contents of
    [a] are destroyed and its storage is used to hold [q] and [r]. *)
  val rq' : matrix -> factor * transform

  (** [reflectors q] returns [(tau, vs)] the elementary reflectors
      that make up q, where q = H(k) ... H(2) H(1), and H(i) = I -
      tau(i) * vs(:,i) v(:,i)^T. *)
  val reflectors : transform ->
    (t, precision_elt, fortran_layout) Array1.t * matrix

   (** [plu' a] computes the PLU fatorization of [a]. The contents of
     [a] are detsroyed and its storage is used to hold [l] and [u]. *)
   val plu' : matrix -> transform * factor * factor

  (** Given an L [factor] [l] computed from [ql'], [lq'], or [plu'], [copyL l]
    returns a lower-triangular (or lower-trapezoidal) [matrix] that is
    equivalent to [l]. Many bugs in the usage of Lapack can be traced
    to this one operation. *)
  val copyL : ?ooc:bool -> factor -> matrix

  (** Given an [l] [factor] computed using [ql'], [lq'], or [plu'], [iDL l a]
    copies the lower-trianglar (or lower-trapezoidal) part of [l] into
    [a]. *)
  val iDL : factor -> matrix -> unit

(** Given an [u] [factor] from [plu'], [copyU u] copies the upper-triangular
   part of [u]. *)
    val copyU : ?ooc:bool -> factor -> matrix

(** Given an [u] [factor] from [plu'], [iDU u b] copies the upper-triangular
   part of [u] into the upper-triangular part of [b]. *)
    val iDU : factor -> matrix -> unit

  (** [q @* a] applies a [transform] [q] to the [matrix] [a] changing
    its contents. *)
  val ( @* ) : transform -> matrix -> unit

  (** [q @~* a] applies the conjugate-transpose of the [transform] [q]
    to the [matrix] [a], changing its contents. Note again that the
    transpose symbol [~] is attached to the operator rather than the
    argument. In any operation involving [transform]s the transpose
    symbol is always attached to the [transform], not the [matrix]
    argument. This is not a short-coming.  *)
  val ( @~* ) : transform -> matrix -> unit

  (** [a *@ q] applies the [transform] [q] to the [matrix] [a] from
    the right, changing its contents. *)
  val ( *@ ) : matrix -> transform -> unit

  (** [a *~@ q] applies the conjugate-transpose of the [transform] [q]
    to the [matrix] [a] from the right, changing its contents. *)
  val ( *~@ ) : matrix -> transform -> unit

  (** [subs' l b] replaces the contents of the [matrix] [b] by
    [l]{^-1}[b]. It assumes that [l] is a square triangular
    matrix. That is, if [l] is trapezoidal you will get a
    run-time error. *)
  val subs' : factor -> matrix -> unit

  (** [subsT' l b] replaces the contents of the [matrix] [b] with that
    of (transpose [l]){^-1}[b]. It assumes that [l] is a square
    triangular [factor]. Otherwise you will get a run-time
    error. *)
  val subsT' : factor -> matrix -> unit

  (** [a |/| b] replaces the contents of the [matrix] [b] by the
    contents of the minimum-norm least-squares solution of the system of
    linear equations [ax = b]. In the process the contents of the
    [matrix] [a] are also destroyed. If [a] is a fat [m] by [n] ([m <
    n]) [matrix], then the matrix [b] must have [n] rows, and initially
    the actual contents of the right-hand side must lie in the first
    [m] rows. On the other hand, if [a] is a skinny [m] by [n] ([m >
    n]) [matrix], then the least-squares solution will be found in the
    first [n] rows of [b] after the call returns. *)
  val ( |/| ) : matrix -> matrix -> unit

  (** The same semantics and restrictions as [|/|], except that the
    system being solved now is [(transp a)x=b]. *)
  val ( |~/| ) : matrix -> matrix -> unit

  (** [a /$ b] returns a new matrix obtained by applying the
    pseudo-inverse of [a] to [b]. *)
  val ( /$ ) : matrix -> matrix -> matrix

  (** [a /~$ b] returns a new matrix obtained by applying the
    pseudo-inverse of the conjugate-transpose of [a] to [b]. *)
  val ( /~$ ) : matrix -> matrix -> matrix

  (** [pluSolve' a b] returns the inverse of [a] times [b] in [b],
     destroying [a] in the process. It uses the [plu'] factorization and
     will work only on square matrices. *)
  val pluSolve' : matrix -> matrix -> unit

  (** [scale' x a] multiplies every entry of [a] in place by [x]. *)
  val scale' : t -> matrix -> unit

  (** [rowScale' x a] scales every row of [a] in place by the
    corresponding entry in [x]. *)
  val rowScale' : (t, precision_elt, fortran_layout) Array1.t ->
    matrix -> unit

  (** [colScale' a x] scales every column of [a] in place by the
    corresponding entry in [x]. *)
  val colScale' : matrix -> (t, precision_elt, fortran_layout)
    Array1.t -> unit

  (** Same as [scale'] except that first argument must be a real
    number. *)
  val scale_real' : float -> matrix -> unit

  (** Same as [rowScale'] except that the first argument must be a
    real array. Useful for SVDs of complex matrices. *)
  val rowScale_real' : (float, precision_elt_real, fortran_layout) Array1.t ->
    matrix -> unit

  (** Same as [colScale'] except that the second argument must be a
    real array. Useful for SVDs of complex matrices. *)
  val colScale_real' : matrix -> (float, precision_elt_real, fortran_layout)
    Array1.t -> unit

  (** [svd' a] returns the economy SVD of [a], destroying its contents
    in the process. *)
  val svd' :
      ?ooc:bool -> matrix ->
	matrix *
	  (float, precision_elt_real, fortran_layout) Array1.t *
	  matrix

  (** [svdL' a] is identical [svd' a] except that the right singular
    vectors are not calculated. *)
  val svdL' :
      ?ooc:bool -> matrix ->
	matrix *
	  (float, precision_elt_real, fortran_layout) Array1.t

  (** [svdR' a] is identical to [svd' a] except that the left singular
    vectors are not calculated. *)
  val svdR' :
      ?ooc:bool -> matrix ->
	(float, precision_elt_real, fortran_layout) Array1.t *
	  matrix

  (** [singValues' a] is identical to [svd' a] except that the none of
    the singular vectors are computed. *)
  val singValues' : matrix ->
    (float, precision_elt_real, fortran_layout) Array1.t

  (** [cond a] returns the 2-norm condition number of the [matrix] a. *)
  val cond : matrix -> float

  (** Converts a lower-triangular [matrix] to [factor] type. Note, no
    checking or modifications are done. This is purely a coercion. *)
  val isLowerTriangular : matrix -> factor

  (** Converts a upper-triangular [matrix] to [factor] type. Note, no
    checking or modifications are done. This is purely a coercion. *)
  val isUpperTriangular : matrix -> factor

  (** Given a one-dimensional bigarray of length [n] of permuted
    integers from [1] to [n], it creates a permutation transform. No
    checking is done. *)
  val permutation : (int, int_elt, fortran_layout) Array1.t -> transform

  (** tensor a b = a tensor b *)
  val tensor : ?ooc:bool -> matrix -> matrix -> matrix

  (** vecC a = column scanned version of a *)
  val vecC' : matrix -> matrix -> unit

  (** unvecC (vecC a) = a *)
    val unvecC' : matrix -> matrix -> unit

  (** vecR a = row scanned version of a *)
  val vecR' : matrix -> matrix -> unit

  (** unvecR (vecR a) = a *)
    val unvecR' : matrix -> matrix -> unit

end

(** {1 The Real and Complex Modules} *)

(** Real double precision routines. *)
module DBlop : (BLOP with type t = float
		     and type precision_elt = Bigarray.float64_elt
		     and type precision_elt_real =
		       Bigarray.float64_elt)

(** Real single precision routines. *)
module SBlop : (BLOP with type t = float
		     and type precision_elt = Bigarray.float32_elt
		     and type precision_elt_real =
		       Bigarray.float32_elt)

(** Complex double precision routines. *)
module ZBlop : (BLOP with type t = Complex.t
		     and type precision_elt = Bigarray.complex64_elt
		     and type precision_elt_real =
		       Bigarray.float64_elt)

(** Complex single precision routines. *)
module CBlop : (BLOP with type t = Complex.t
		     and type precision_elt = Bigarray.complex32_elt
		     and type precision_elt_real =
		       Bigarray.float32_elt)


