(*
   File: poisson2D.ml
   Author: Shivkumar Chandrasekaran
   Date: 9.28.2004
   Purpose: Low-order psuedo-spectral Poisson solver on 2D rectangle.
   Package: CamlFloat

   $Id: poisson2D.ml,v 1.5 2005/04/11 21:42:06 shiv Exp shiv $
*)


open Bigarray
open Utils
open Nla
open Chub


module type POISSON2D =
    sig
      type matrix
      val make_pd : int -> matrix
      val make_pn : int -> matrix
      val make_pnd : float -> int -> matrix
      val partialXX : int -> int -> matrix
      val partialYY : int -> int -> matrix
      val partialX : int -> int -> matrix
      val partialY : int -> int -> matrix
      val projectDownX : int -> int -> int -> matrix
      val projectDownY : int -> int -> int -> matrix
    end (* module type POISSON2D *)


module Poissoner2D (Blop : BLOP)
    (Chub : CHUB with type t = Blop.t and type matrix = Blop.matrix) =
  struct

    open Blop
    open Chub

    type matrix = Blop.matrix


    (* Note: x varies with j the column index, and y with i the row index.
       This matches the way we draw the x- and y-axes conventionally. *)
    let discretize2d xs ys f =
      fun2mat (fun i j -> f xs.(j-1) ys.(i-1))
	(Array.length ys) (Array.length xs)


    let cheb2d i j x y = fromFloat (cheb i x *. cheb j y)
	

    let spectDeriv u =
      let du = rawMatrix (noOfRows u - 1) (noOfCols u) in
      spectralDerivative u du;
      du


    let diff1 u =
      let n = noOfRows u in
      let uc = copy u in
      toCheb' uc;
      let dx1u = spectDeriv uc in
      fromCheb' dx1u;
      dx1u


    let diff2 u =
      let n = noOfRows u in
      let uc = copy u in
      toCheb' uc;
      let dx2u = spectDeriv (spectDeriv uc) in
      fromCheb' dx2u;
      dx2u


    let projectDown2 f =
      let fc = copy f
      and n = noOfRows f in
      toCheb' fc;
      chebVand (chebZeros (n - 2)) n *$ fc


    let projectDown k p =
      let f = eye p p
      and g = rawMatrix (p - k) p in
      projectTo' (p - k) f g;
      g


    let laplacian u =
      (transp (projectDown2 (transp (diff2 u)))) +$
      (projectDown2 (transp (diff2 (transp u))))


    let dirichlet u =
      let eval1 uc y =
	let n = noOfCols uc
	and m = noOfRows uc in
	let v = transp (chebVand [|y|] m *$ uc) in
	toCheb' v;
	chebVand (chebZeros (n - 1)) n *$ v
      in
      let uc = copy u in
      toCheb' uc;
      let t = eval1 uc (1.0)
      and b = eval1 uc (-1.0) in
      let uc = transp u in
      toCheb' uc;
      let l = eval1 uc (-1.0)
      and r = eval1 uc (1.0) in
      (t, r, b, l)


    let column a k = a $ (All, k <-> k)


    let unvecC m v =
      let n = noOfRows v / m in
      assert (m * n = noOfRows v && noOfCols v = 1);
      let uv = rawMatrix m n in
      unvecC' v uv;
      uv


    let svdE' tol a =
      let k = min (noOfRows a) (noOfCols a) in
      let (u, s, vt) = svd' a in
      let r = ref 0 in
      while !r < k && s.{!r+1} > s.{1} *. tol do
	r := !r + 1
      done;
      assert (!r <= k && (!r = 0 || s.{!r} > s.{1} *. tol));
      (u $ (All, 1 <-> !r), Array1.sub s 1 !r, vt $ (1 <-> !r, All))


    let psDerivative1 p =
      let ps = eye p p
      and ps1 = rawMatrix (p - 1) p in
      psDerivative1' ps ps1;
      ps1


    (* px.py -> px.(py-1) *)
    let partialY px py =
      tensor (eye px px) (psDerivative1 py)


    (* px.py -> (px-1).py *)
    let partialX px py =
      tensor (psDerivative1 px) (eye py py)


    (* px.py -> px.(py-2) *)
    let partialYY px py =
      tensor (eye px px) (diff2 (eye py py))


    (* px.py -> (px-2).py *)
    let partialXX px py =
      tensor (diff2 (eye px px)) (eye py py)


    (* px.py -> (px-1).(py-1) *)
    let partialXY px py =
      tensor (psDerivative1 px) (psDerivative1 py)


    (* px.py -> px.(py-k) *)
    let projectDownY k px py =
      tensor (eye px px) (projectDown k py)

    
    (* px.py -> (px-k).py *)
    let projectDownX k px py =
      tensor (projectDown k px) (eye py py)


    (* P_D (f; g) = u is computed, where \nabla^2 u = f on
       [-1,1]x[-1,1] and u = g on the edges. p points on each edge,
       (p-1)x(p-1) for f and (p+1)x(p+1) for u. g is clock-wise starting at
       top. f and u are column-wise starting at left. Observe that the
       map is only a pseudo-inverse. If f & g come from a tensor
       Chebyshev polynomial then it is exact. Otherwise not---by quite
       a bit! *)
    let make_pd p =
      let nu = p+1
      and nl = p-1 in
      let u = rawMatrix (nu * nu) (nu * nu)
      and fg = rawMatrix (nu * nu) (nu * nu) in
      let (f, g) = partition2x1 (nl * nl)
                                  (4 * p) fg in
      let (gt, gr, gb, gl) =
	let (gtr, gbl) = partition2x1 (2 * p) (2 * p) g in
	let (gt, gr) = partition2x1 p p gtr
	and (gb, gl) = partition2x1 p p gbl in
	(gt, gr, gb, gl)
      in
      let znu = chebZeros nu
      and zp = chebZeros p in
      for i = 0 to nu - 1 do
	for j = 0 to nu - 1 do
	  let uij = discretize2d znu znu (cheb2d i j) in
	  let fij = laplacian uij
	  and bt = discretize zp 1 (fun x _ -> cheb2d i j x  1.0)
	  and bb = discretize zp 1 (fun x _ -> cheb2d i j x (-1.0))
	  and bl = discretize zp 1 (fun y _ -> cheb2d i j (-1.0) y)
	  and br = discretize zp 1 (fun y _ -> cheb2d i j 1.0  y) in
	  let k = 1 + i * nu + j in
	  (column u k) <-| vecC' uij;
	  (column f k) <-| vecC' fij;
	  (column gt k) <-$ bt;
	  (column gb k) <-$ bb;
	  (column gl k) <-$ bl;
	  (column gr k) <-$ br
	done
      done;
      let (w, s, vt) = svdE' (machine_epsilon ** (3. /. 4.)) (copy fg) in
      let uv = u *$~ vt in
      for i = 1 to Array1.dim s do
	s.{i} <- 1.0 /. s.{i}
      done;
      colScale_real' uv s;
      let pd = uv *$~ w in
      (* Begin checking *)
      let pdfg = pd *$ fg in
      let newfg = rawMatrix (nu * nu) (nu * nu) in
      let newu = rawMatrix nu nu in
      for i = 1 to nu * nu do
	let newlap = newfg $ (1 <-> nl * nl, i <-> i) in
	let newd = newfg $ (nl*nl + 1 <-> nu*nu, i <-> i) in
	let (newtr, newbl) = partition2x1 (2 * p) (2 * p) newd in
	let (newt, newr) = partition2x1 p p newtr
	and (newb, newl) = partition2x1 p p newbl in
	unvecC' (column pdfg i) newu;
	vecC' (laplacian newu) newlap;
	let (t, r, b, l) = dirichlet newu in
	newt <-$ t; newb <-$ b; newl <-$ l; newr <-$ r;
      done;
      Printf.printf "projection error = %g\n" (norm1 (fg -$ newfg) /. norm1 fg);
      (* Done checking *)
      pd


    (* Given a px x py Chebyshev grid with a function u discretized on
       it, return the outward normal derivatives at the boundary on px
       and py Chebyshev points. The ordering is top, right, bottom,
       left for the returned values. *)
    let normalDerivatives u =
      let px = noOfCols u
      and py = noOfRows u in
      let uy = diff1 u
      and uxt = diff1 (transp u) in
      toCheb' uy;
      toCheb' uxt;
      let ndy = chebVand [| 1.0; (-1.0) |] (py - 1) *$ uy
      and ndx = chebVand [| 1.0; (-1.0) |] (px - 1) *$ uxt in
      scale_real' (-1.0) (ndy $ (2 <-> 2, All));
      scale_real' (-1.0) (ndx $ (2 <-> 2, All));
      let normDeriv = rawMatrix (2 * px + 2 * py) 1 in
      let (ndT, ndR, ndB, ndL) =
	let (normDerivTR, normDerivBL) =
	  partition2x1 (px + py) (px + py) normDeriv in
	let (ndT, ndR) = partition2x1 px py normDerivTR
	and (ndB, ndL) = partition2x1 px py normDerivBL in
	(ndT, ndR, ndB, ndL) in
      transp' (ndy $ (1 <-> 1, All)) ndT;
      transp' (ndx $ (1 <-> 1, All)) ndR;
      transp' (ndy $ (2 <-> 2, All)) ndB;
      transp' (ndx $ (2 <-> 2, All)) ndL;
      normDeriv


    (* Same as normalDerivatives except it returns it at (px - 1) and
       (py - 1) Chebyshev boundary points. *)
    let normalDerivatives2 u =
      let px = noOfCols u
      and py = noOfRows u in
      let pdx1 = projectDown 1 px
      and pdy1 = projectDown 1 py in
      let uy = diff1 (u *$~ pdx1)
      and uxt = diff1 (transp (pdy1 *$ u)) in
      toCheb' uy;
      toCheb' uxt;
      let ndy = chebVand [| 1.0; (-1.0) |] (py - 1) *$ uy
      and ndx = chebVand [| 1.0; (-1.0) |] (px - 1) *$ uxt in
      scale_real' (-1.0) (ndy $ (2 <-> 2, All));
      scale_real' (-1.0) (ndx $ (2 <-> 2, All));
      let normDeriv = rawMatrix (2 * (px - 1) + 2 * (py - 1)) 1 in
      let (ndT, ndR, ndB, ndL) =
	let (normDerivTR, normDerivBL) =
	  partition2x1 (px + py - 2) (px + py - 2) normDeriv in
	let (ndT, ndR) = partition2x1 (px-1) (py-1) normDerivTR
	and (ndB, ndL) = partition2x1 (px-1) (py-1) normDerivBL in
	(ndT, ndR, ndB, ndL) in
      transp' (ndy $ (1 <-> 1, All)) ndT;
      transp' (ndx $ (1 <-> 1, All)) ndR;
      transp' (ndy $ (2 <-> 2, All)) ndB;
      transp' (ndx $ (2 <-> 2, All)) ndL;
      normDeriv


    (* Given a function on a 2D Chebyshev grid, evaluate the function
       at the center of the grid. *)
    let midPointValue u =
      let py = noOfRows u
      and px = noOfCols u in
      let uc = copy u in
      toCheb' uc;
      let uct = transp uc in
      toCheb' uct;
      (chebVand [|0.0|] px *$ uct *$~ chebVand [|0.0|] py) $@ (1,1)


    (* int_{-1}^1 int_{-1}^1 u *)
    let integral2D u =
      let py = noOfRows u
      and px = noOfCols u in
      let qx = chebQuadWts px
      and qy = chebQuadWts py in
      (qy *$ u *$~ qx) $@ (1,1)


    let zerothCheb u =
      (* let uc = copy u in
      toCheb2D' uc;
      uc $@ (1,1) *)
      let py = noOfRows u
      and px = noOfCols u in
      (ones 1 py *$ u *$ ones px 1) $@ (1,1)


    (* P_N (f; g; um) = u is computed, where \nabla^2 u = f on
       [-1,1]x[-1,1] and \partial_\nu u = g on edges, and um =
       u(0,0). There are p points on each edge of the boundary for
       g. (p-2)x(p-2) for f and pxp for u.  g is clock-wise starting
       at top. f and u are column-wise starting at left. Observe that
       the map is only a pseudo-inverse. If f & g come from a tensor
       Chebyshev polynomial then it is exact. Otherwise not---by quite
       a bit! *)
    let make_pn1 p =
      let nu = p
      and nl = p-2 in
      let nf = nl * nl
      and ng = 4*p in
      let nfgum = nf + ng + 1 in
      let u = rawMatrix (nu * nu) (nu * nu)
      and fgum = rawMatrix nfgum (nu * nu) in
      let (f, g, um) = partition3x1 nf
                                    ng
	                             1 fgum in
      let znu = chebZeros nu
      and zp = chebZeros p in
      for i = 0 to nu - 1 do
	for j = 0 to nu - 1 do
	  let uij = discretize2d znu znu (cheb2d i j) in
	  let fij = laplacian uij
	  and gij = normalDerivatives uij in
	  let k = 1 + i * nu + j in
	  (column u k) <-| vecC' uij;
	  (column f k) <-| vecC' fij;
	  (column g k) <-$ gij;
	  (cheb2d i j 0. 0. => (1,1)) um;
	done
      done;
      let (w, s, vt) = svdE' (machine_epsilon ** (3. /. 4.)) (copy fgum) in
      let uv = u *$~ vt in
      for i = 1 to Array1.dim s do
	s.{i} <- 1.0 /. s.{i}
      done;
      colScale_real' uv s;
      let pn = uv *$~ w in
      let check = true in
      if check then begin
	let pn_fgum = pn *$ fgum in
	let new_fgum = rawMatrix nfgum (nu * nu) in
	let newu = rawMatrix nu nu in
	for i = 1 to nu * nu do
	  let new_lap = new_fgum $ (1 <-> nl * nl, i <-> i) in
	  let new_n = new_fgum $ (nl*nl + 1 <-> nf+ng, i <-> i) in
	  unvecC' (column pn_fgum i) newu; (* compute u *)
	  vecC' (laplacian newu) new_lap;  (* compute Laplacian of u *)
	  new_n <-$ normalDerivatives newu; (* compute Neumann data of u *)
	  (midPointValue newu => (nfgum, i)) new_fgum; (* u(0,0) *)
	done;
	let p2 = (p-2)*(p-2) in
(*	Blop.printMatrixHeight := nfgum;
	printMatrix ((fgum -$ new_fgum) $ (All, 1<-> p2)); print_newline (); *)
	Printf.printf "projection error = %g %g\n"
	  (norm1 (fgum -$ new_fgum) /. norm1 fgum)
	  (norm1 ((fgum -$ new_fgum) $ (1 <-> nfgum-1, All)));
      end;
      pn


    (* P_N (f; g; um) = u is computed, where \nabla^2 u = f on
       [-1,1]x[-1,1] and \partial_\nu u = g on edges, and um = zeroth
       Chebyshev coefficient. There are p points on each edge of the
       boundary for g. (p-2)x(p-2) for f and pxp for u.  g is
       clock-wise starting at top. f and u are column-wise starting at
       left. Observe that the map is only a pseudo-inverse. If f & g
       come from a tensor Chebyshev polynomial then it is
       exact. Otherwise not---by quite a bit! *)
    let make_pn p =
      let nu = p
      and nl = p-2 in
      let nf = nl * nl
      and ng = 4*p in
      let nfgum = nf + ng + 1 in
      let u = rawMatrix (nu * nu) (nu * nu)
      and fgum = rawMatrix nfgum (nu * nu) in
      let (f, g, um) = partition3x1 nf
                                    ng
	                             1 fgum in
      let znu = chebZeros nu
      and zp = chebZeros p in
      for i = 0 to nu - 1 do
	for j = 0 to nu - 1 do
	  let uij = discretize2d znu znu (cheb2d i j) in
	  let fij = laplacian uij
	  and gij = normalDerivatives uij in
	  let k = 1 + i * nu + j in
	  (column u k) <-| vecC' uij;
	  (column f k) <-| vecC' fij;
	  (column g k) <-$ gij;
	  (zerothCheb uij => (1,1)) um;
	done
      done;
      let (w, s, vt) = svdE' (machine_epsilon ** (3. /. 4.)) (copy fgum) in
      let uv = u *$~ vt in
      for i = 1 to Array1.dim s do
	s.{i} <- 1.0 /. s.{i}
      done;
      colScale_real' uv s;
      let pn = uv *$~ w in
      let check = true in
      if check then begin
	let pn_fgum = pn *$ fgum in
	let new_fgum = rawMatrix nfgum (nu * nu) in
	let newu = rawMatrix nu nu in
	for i = 1 to nu * nu do
	  let new_lap = new_fgum $ (1 <-> nl * nl, i <-> i) in
	  let new_n = new_fgum $ (nl*nl + 1 <-> nf+ng, i <-> i) in
	  unvecC' (column pn_fgum i) newu; (* compute u *)
	  vecC' (laplacian newu) new_lap;  (* compute Laplacian of u *)
	  new_n <-$ normalDerivatives newu; (* compute Neumann data of u *)
	  (zerothCheb newu => (nfgum, i)) new_fgum;
	done;
	let p2 = (p-2)*(p-2) in
(*	Blop.printMatrixHeight := nfgum;
	printMatrix ((fgum -$ new_fgum) $ (All, 1<-> p2)); print_newline (); *)
	Printf.printf "projection error = %g %g\n"
	  (norm1 (fgum -$ new_fgum) /. norm1 fgum)
	  (norm1 ((fgum -$ new_fgum) $ (1 <-> nfgum-1, All)));
      end;
      pn


    (* P_ND (f; g) = u is computed, where \nabla^2 u = f on
       [-1,1]x[-1,1] and \partial_\nu u +\gamma u = g on edges. There
       are p points on each edge of the boundary for g. (p-1)x(p-1)
       for f and (p=1)x(p+1) for u.  g is clock-wise starting at
       top. f and u are column-wise starting at left. Observe that the
       map is only a pseudo-inverse. If f & g come from a tensor
       Chebyshev polynomial then it is exact. Otherwise not---by quite
       a bit! *)
    let make_pnd gamma p =
      let nu = p+1
      and nl = p-1 in
      let nf = nl * nl
      and ng = 4*p in
      let nfg = nf + ng in
      let u = rawMatrix (nu * nu) (nu * nu)
      and fg = rawMatrix (nu * nu) (nu * nu) in
      let (f, g) = partition2x1 (nl * nl)
	                          (4 * p) fg in
      let (gt, gr, gb, gl) =
	let (gtr, gbl) = partition2x1 (2 * p) (2 * p) g in
	let (gt, gr) = partition2x1 p p gtr
	and (gb, gl) = partition2x1 p p gbl in
	(gt, gr, gb, gl)
      in
      let znu = chebZeros nu
      and zp = chebZeros p in
      for i = 0 to nu - 1 do
	for j = 0 to nu - 1 do
	  let uij = discretize2d znu znu (cheb2d i j) in
	  let fij = laplacian uij
	  and bt = discretize zp 1 (fun x _ -> cheb2d i j x  1.0)
	  and bb = discretize zp 1 (fun x _ -> cheb2d i j x (-1.0))
	  and bl = discretize zp 1 (fun y _ -> cheb2d i j (-1.0) y)
	  and br = discretize zp 1 (fun y _ -> cheb2d i j 1.0  y)
	  and gij = normalDerivatives2 uij in
	  let k = 1 + i * nu + j in
	  (column u k) <-| vecC' uij;
	  (column f k) <-| vecC' fij;
	  gij |++| (column g k);
	  scale_real' gamma bt; bt |++| (column gt k);
	  scale_real' gamma bb; bb |++| (column gb k);
	  scale_real' gamma bl; bl |++| (column gl k);
	  scale_real' gamma br; br |++| (column gr k);
	done
      done;
      let (w, s, vt) = svdE' (machine_epsilon ** (3. /. 4.)) (copy fg) in
      let uv = u *$~ vt in
      for i = 1 to Array1.dim s do
	s.{i} <- 1.0 /. s.{i}
      done;
      colScale_real' uv s;
      let pnd = uv *$~ w in
      let check = true in
      if check then begin
	let pnd_fg = pnd *$ fg in
	let new_fg = rawMatrix nfg (nu * nu) in
	let newu = rawMatrix nu nu in
	for i = 1 to nu * nu do
	  let new_lap = new_fg $ (1 <-> nl * nl, i <-> i) in
	  let new_nd = new_fg $ (nl*nl + 1 <-> nu*nu, i <-> i) in
	  let (newtr, newbl) = partition2x1 (2 * p) (2 * p) new_nd in
	  let (newt, newr) = partition2x1 p p newtr
	  and (newb, newl) = partition2x1 p p newbl in
	  unvecC' (column pnd_fg i) newu; (* compute u *)
	  vecC' (laplacian newu) new_lap;  (* compute Laplacian of u *)
	  new_nd <-$ normalDerivatives2 newu; (* compute Neumann data of u *)
	  let (t, r, b, l) = dirichlet newu in
	  scale_real' gamma t; t |++| newt;
	  scale_real' gamma b; b |++| newb;
	  scale_real' gamma l; l |++| newl;
	  scale_real' gamma r; r |++| newr;
	done;
	let p2 = (p-1)*(p-1) in
(*	Blop.printMatrixHeight := nfgum;
	printMatrix ((fg -$ new_fg) $ (All, 1<-> p2)); print_newline (); *)
	Printf.printf "projection error = %g\n"
	  (norm1 (fg -$ new_fg) /. norm1 fg);
      end;
      pnd



  end (* module Poissoner2D (Blop : BLOP) (Chub : CHUB) *)


module DPoisson2D = Poissoner2D (DBlop) (DChub)
module SPoisson2D = Poissoner2D (SBlop) (SChub)
module CPoisson2D = Poissoner2D (CBlop) (CChub)
module ZPoisson2D = Poissoner2D (ZBlop) (ZChub)


