(*
   File: poly.ml
   Author: Shivkumar Chandrasekaran
   Date: 9.29.2004
   Purpose: Exact polynomial algebra

   $Id: poly.ml,v 1.5 2004/10/26 18:41:39 shiv Exp shiv $
*)

let zip_with fn def0 def1 a b =
  let rec zw a b acc =
    match a, b with
      [], [] -> List.rev acc
    | _, [] -> List.rev_append acc (List.map (fun ai -> fn ai def1) a)
    | [], _ -> List.rev_append acc (List.map (fun bi -> fn def0 bi) b)
    | a0 :: a1s, b0 :: b1s  -> zw a1s b1s (fn a0 b0 :: acc)
  in
  zw a b []
	
let range n0 n1 =
  assert (n0 <= n1);
  let rec gr i acc =
    if i < n0 then acc
    else gr (i-1) (i :: acc)
  in
  gr n1 []
	
let take n ls =
  let rec tk i acc l =
    if i <= 0 then
      List.rev acc
    else
      tk (i - 1) (List.hd l :: acc) (List.tl l)
  in
  tk n [] ls

let rec drop n l =
  if n <= 0 then
    l
  else
    drop (n - 1) (List.tl l)


(* Poylnomials and inner product spaces over these. *)
module type SCALAR =
    sig
      type t
      val add : t -> t -> t
      val sub : t -> t -> t
      val mul : t -> t -> t
      val div : t -> t -> t
      val minus : t -> t
      val equal : t -> t -> bool
      val one : t
      val zero : t
      val of_int : int -> t
      val of_float : float -> t
      val to_string : t -> string
    end (* module type SCALAR *)


(* Signature for one variable polynomial *)
module type ONE =
  sig
    module Scalar : SCALAR
    type t
    val coeff : t -> int -> Scalar.t
    val degree : t -> int
    val of_array : Scalar.t array -> t
    val of_list : Scalar.t list -> t
    val to_list : t -> Scalar.t list
    val add : t -> t -> t
    val sub : t -> t -> t
    val minus : t -> t
    val scale : Scalar.t -> t -> t
    val prune : t -> t
    val mul : t -> t -> t
    val derivative : t -> t
    val eval : t -> Scalar.t -> Scalar.t
    val integral : t -> t
    val definiteIntegral : Scalar.t -> Scalar.t -> t -> Scalar.t
    val print : t -> string
    val zero : t
    val cheb : int -> t
    val chebs : int -> t list
  end (* module type ONE *)


(* Signature of two variable polynomial. *)
module type TWO =
  sig
    module Scalar : SCALAR
    module One : ONE
    type t
    val zero : t
    val of_list : Scalar.t list list -> t
    val add : t -> t -> t
    val sub : t -> t -> t
    val minus : t -> t
    val scale : Scalar.t -> t -> t
    val prune : t -> t
    val mul : t -> t -> t
    val partialX : t -> t
    val partialY : t -> t
    val laplacian : t -> t
    val evalX : t -> Scalar.t -> One.t
    val evalY : t -> Scalar.t -> One.t
    val integralX : t -> t
    val integralY : t -> t
    val rectSurfaceIntegral : Scalar.t -> Scalar.t -> Scalar.t -> Scalar.t ->
      t -> Scalar.t
    val rectLineIntegral : Scalar.t -> Scalar.t -> Scalar.t -> Scalar.t ->
      t -> Scalar.t
    val print : t -> string
    val harmonic : int -> t list
    val nonHarmonic : int -> t list
    val monomial : int -> t list
    val harmonics : int -> t list
    val nonHarmonics : int -> t list
    val monomials : int -> t list
    val liftX : One.t -> t
    val liftY : One.t -> t
    val cheb : int -> int -> t
    val chebs : int -> int -> t list
  end (* module type TWO *)


module NumScalar =
  struct
    type t = Num.num
    let add = Num.add_num
    let sub = Num.sub_num
    let mul = Num.mult_num
    let div = Num.div_num
    let minus = Num.minus_num
    let equal = Num.eq_num
    let one = Num.num_of_int 1
    let zero = Num.num_of_int 0
    let of_int = Num.num_of_int
    let of_float x = Num.num_of_string (string_of_float x)
    let to_string = Num.string_of_num
  end (* module NumScalar *)


module FloatScalar =
  struct
    type t = float
    let add = ( +. )
    let sub = ( -. )
    let mul = ( *. )
    let div = ( /. )
    let minus = ( ~-. )
    let equal = ( = )
    let one = 1.0
    let zero = 0.0
    let of_int = float
    let of_float x = x
    let to_string = string_of_float
  end (* module FloatScalar *)


module One (Scalar : SCALAR) =
  struct

    module Scalar = Scalar

    type t = Scalar.t list

    let degree p = List.length p - 1

    let zero = []

    let coeff poly i = List.nth poly i
	
    let of_array p = Array.to_list p
	
    let of_list p = p

    let to_list p = p
	
    let add p q = zip_with Scalar.add Scalar.zero Scalar.zero p q
	
    let sub p q = zip_with Scalar.sub Scalar.zero Scalar.zero p q
	
    let minus p = List.map Scalar.minus p
	
    let scale a p = List.map (Scalar.mul a) p
	
    let prune p =
      let rec pp p =
	match p with
	  [] -> p
	| p0 :: ps when Scalar.equal p0 Scalar.zero -> pp ps
	| _ -> List.rev p
      in
      pp (List.rev p)
	
    let rec mul p q =
      match p, q with
	[], _ -> []
      | _, [] -> []
      | p0 :: p1s, q0 :: q1s ->
	  let hot = Scalar.zero :: mul p1s q1s in
	  prune (Scalar.mul p0 q0 ::
		 add (add (scale p0 q1s) (scale q0 p1s)) hot)
	    
    let derivative p =
      match p with
	[] -> []
      | _ ->
	  let pwr = List.map Scalar.of_int (range 0 (List.length p - 1)) in
	  List.tl (zip_with Scalar.mul Scalar.zero Scalar.zero p pwr)
	
    let eval p x =
      List.fold_right (fun coeff acc -> Scalar.add coeff (Scalar.mul x acc))
	p Scalar.zero
	
    let integral p =
      match p with
	[] -> []
      | _ ->
	  let quo = List.map Scalar.of_int (range 1 (List.length p)) in
	  Scalar.zero :: (zip_with Scalar.div Scalar.one Scalar.zero p quo)
	       
    let definiteIntegral a b p =
      let pi = integral p in
      Scalar.sub (eval pi b) (eval pi a)
	
    let print p =
      let n = List.length p in
      if n = 0 then begin
	"0"
      end else begin
	let pa = Array.of_list p in
	let s = ref (Scalar.to_string pa.(0)) in
	for i = 1 to Array.length pa - 1 do
	  s := !s ^ (Printf.sprintf " + %s x^%i" (Scalar.to_string pa.(i)) i)
	done;
	!s
      end

    let chebs n =
      let rec rchebs n =
	if n <= 0 then
	  []
	else if n = 1 then
	  [[Scalar.one]]
	else if n = 2 then
	  [[Scalar.zero; Scalar.one]; [Scalar.one]]
	else
	  let cs = rchebs (n - 1) in
	  sub (mul [Scalar.zero; Scalar.of_int 2] (List.hd cs))
	    (List.hd (List.tl cs)) :: cs
      in
      List.rev (rchebs n)

    let cheb n = List.hd (List.rev (chebs (n+1)))
	  
  end (* module One *)


module Two (One : ONE) =
  struct

    module Scalar = One.Scalar

    module One = One

    type t = One.t list (* Each coeff is a poly in x *)

    let zero = [One.zero]

    let of_list ps = List.map One.of_list ps

    let add p q = zip_with One.add One.zero One.zero p q

    let sub p q = zip_with One.sub One.zero One.zero p q

    let minus p = List.map One.minus p

    let scale a p = List.map (One.scale a) p

    let prune p =
      let rec pp p =
	match p with
	  [] -> p
	| py0 :: ps when One.degree py0 = -1 -> pp ps
	| _ -> List.rev p
      in
      let p1 = List.map One.prune p in
      pp (List.rev p1)

    let rec mul p q =
      let xscale px qxs = List.map (One.mul px) qxs in
      match p, q with
	[], _ -> []
      | _, [] -> []
      | p0 :: p1s, q0 :: q1s ->
	  let hot = One.zero :: mul p1s q1s in
	  prune (One.mul p0 q0 :: add (add (xscale p0 q1s) (xscale q0 p1s)) hot)

    let partialX p = prune (List.map One.derivative p)

    let partialY p =
      match p with
	[] -> []
      | _ ->
	  let pwr = List.map Scalar.of_int (range 0 (List.length p - 1)) in
	  prune (List.tl (zip_with One.scale Scalar.zero One.zero pwr p))

    let laplacian p =
      prune (add (partialX (partialX p)) (partialY (partialY p)))

    let evalX p x = One.of_list (List.map (fun px -> One.eval px x) p)

    let evalY p y =
      List.fold_right (fun coeffX acc -> One.add coeffX (One.scale y acc))
	p One.zero

    let eval p x y =
      One.eval (One.of_list (List.map (fun px -> One.eval px x) p)) y

    let integralX p = List.map One.integral p

    let integralY p =
      let quo = List.map Scalar.of_int (range 1 (List.length p)) in
      One.zero ::
      (zip_with (fun px n -> One.scale (Scalar.div Scalar.one n) px)
	 One.zero Scalar.zero p quo)

    let rectSurfaceIntegral x0 x1 y0 y1 p =
      let px = One.of_list (List.map (One.definiteIntegral x0 x1) p) in
      One.definiteIntegral y0 y1 px

    (* Scalar line integral. Note lack of directionality. *)
    let rectLineIntegral x0 x1 y0 y1 p =
      let pb = One.definiteIntegral x0 x1 (evalY p y0)
      and pt = One.definiteIntegral x0 x1 (evalY p y1)
      and pr = One.definiteIntegral y0 y1 (evalX p x1)
      and pl = One.definiteIntegral y0 y1 (evalX p x0) in
      Scalar.add (Scalar.add (Scalar.add pb pt) pr) pl

    let print p =
      let n = List.length p in
      if n = 0 then begin
	"0"
      end else begin
	let pa = Array.of_list p in
	let s = ref (One.print pa.(0)) in
	for i = 1 to Array.length pa - 1 do
	  s := !s ^ (Printf.sprintf " + (%s) y^%i" (One.print pa.(i)) i)
	done;
	!s
      end

    type z = {re : t; im : t}

    let zmul z1 z2 =
      {re = sub (mul z1.re z2.re) (mul z1.im z2.im);
       im = add (mul z1.re z2.im) (mul z1.im z2.re)}

    let rec zpow z n =
      assert (n >= 0);
      if n = 0 then
	{re = of_list [[Scalar.one]]; im = of_list [[]]}
      else if n mod 2 = 0 then
	let znd2 = zpow z (n/2) in
	zmul znd2 znd2
      else (* n mod 2 = 1 *)
	let znm1 = zpow z (n - 1) in
	zmul z znm1

    (* harmonics of degree n *)
    let harmonic n =
      assert (n >= 0);
      if n = 0 then
	[of_list [[Scalar.one]]]
      else
	let zn = zpow {re = of_list [[Scalar.zero; Scalar.one]];
		       im = of_list [[]; [Scalar.one]]} n
	in
	[zn.re; zn.im]      

    (* r2 = x^2 + y^2 *)
    let r2 = of_list [[Scalar.zero; Scalar.zero; Scalar.one]; [];
		      [Scalar.one]]

    (* monomials of degree n *)
    let rec monomial n =
      if n <= 1 then
	harmonic n
      else
	(harmonic n) @ (List.map (mul r2) (monomial (n-2)))

    (* First n monomials. *)
    let monomials n =
      let rec acr m acc =
	if List.length acc >= n then
	  take n (List.rev acc)
	else
	  acr (m + 1) (List.rev_append (monomial m) acc)
      in
      acr 0 []

    (* harmonic n osum nonHarmonic n = monomial n *)
    let nonHarmonic n =
      if n <= 1 then
	[]
      else
	drop 2 (monomial n)

    (* First n nonHarmonics. *)
    let nonHarmonics n =
      let rec acr m acc =
	if List.length acc >= n then
	  take n (List.rev acc)
	else
	  acr (m + 1) (List.rev_append (nonHarmonic m) acc)
      in
      acr 0 []

    (* The first n harmonic polynomials *)
    let harmonics n =
      if n <= 0 then
	[]
      else
	let z0 = {re = of_list [[Scalar.one]]; im = of_list [[]]}
	and z1 = {re = of_list [[Scalar.zero; Scalar.one]];
		  im = of_list [[]; [Scalar.one]]} in
	let rec zPowers acc i =
	  if i = 0 then
	    List.rev acc
	  else
	    zPowers (zmul z1 (List.hd acc) :: acc) (i - 1)
	in
	let pows = List.tl (zPowers [z0] (n / 2)) in
	z0.re ::
	take (n-1) (List.flatten (List.map (fun z -> [z.re; z.im]) pows))

    (* Make a polynomial in x, into one in (x, y) *)
    let liftX px = [px]

    (* Make a polynomial in y, into one in (x, y) *)
    let liftY py = List.map (fun a -> One.of_list [a]) (One.to_list py)

    (* Two dimensional Chebyshev polynomial. *)
    let cheb i j = mul (liftX (One.cheb i)) (liftY (One.cheb j))

    (* The first nx*ny 2D chebyshev polynomials, y varying fastest. *)
    let chebs nx ny =
      let cx = List.map liftX (One.chebs nx)
      and cy = List.map liftY (One.chebs ny) in
      List.flatten (List.map (fun px -> List.map (mul px) cy) cx)

  end (* module Two (One : ONE) *)


(* Real inner product space signature *)
module type IP =
  sig
    module Scalar : SCALAR
    type t
    val add : t -> t -> t
    val sub : t -> t -> t
    val zero : t
    val scale : Scalar.t -> t -> t
    val ip : t -> t -> Scalar.t
  end


(* Inner product on boundary of square. *)
module Boundary (Two : TWO) =
  struct
    module Scalar = Two.Scalar
    type t = Two.t
    let add = Two.add
    let sub = Two.sub
    let zero = Two.zero
    let scale = Two.scale
    let ip p q = Two.rectLineIntegral (Scalar.of_int (-1))
	Scalar.one (Scalar.of_int (-1)) Scalar.one (Two.mul p q)
  end (* module Boundary (Two : TWO) *)


(* Inner product on interior of square. *)
module Interior (Two : TWO) =
  struct
    module Scalar = Two.Scalar
    type t = Two.t
    let add = Two.add
    let sub = Two.sub
    let zero = Two.zero
    let scale = Two.scale
    let ip p q = Two.rectSurfaceIntegral (Scalar.of_int (-1))
	Scalar.one (Scalar.of_int (-1)) Scalar.one (Two.mul p q)
  end (* module Interior (Two : TWO) *)


(* Functor for Gram-Schmidt process *)
module GS (Ip : IP) =
  struct

    (* Make x perperpendicular to v *)
    let makePerpTo v x =
      Ip.sub x (Ip.scale (Ip.Scalar.div (Ip.ip v x) (Ip.ip v v)) v)
      
    (* assumes vs is orthogonal already *)
    let rec orthogonalize v vs =
      match vs with
	[] -> v
      | v0 :: vrs ->
	  orthogonalize (makePerpTo v0 v) vrs

    let mgs vs =
      let rec domgs acc vs =
	match vs with
	  [] -> List.rev acc
	| v0 :: vrs ->
	    domgs (orthogonalize v0 acc :: acc) vrs
      in
      domgs [] vs

    (* Write x as a linear combination of vs that are orthogonal. Returns
       residual too. *)
    let orthoExpand vs x =
      let coeffs =
	List.map (fun vi -> Ip.Scalar.div (Ip.ip x vi) (Ip.ip vi vi)) vs in
      let projection = List.fold_left Ip.add Ip.zero
	  (zip_with Ip.scale Ip.Scalar.zero Ip.zero coeffs vs) in
      let residual = Ip.sub x projection in
      (coeffs, residual)

  end (* GS (Ip : IP) *)


module OneNum          = One (NumScalar)
module OneFloat        = One (FloatScalar)
module TwoNum          = Two (OneNum)
module TwoFloat        = Two (OneFloat)
module BoundaryNum     = Boundary (TwoNum)
module BoundaryFloat   = Boundary (TwoFloat)
module InteriorNum     = Interior (TwoNum)
module InteriorFloat   = Interior (TwoFloat)
module GSBoundaryNum   = GS (BoundaryNum)
module GSBoundaryFloat = GS (BoundaryFloat)
module GSInteriorNum   = GS (InteriorNum)
module GSInteriorFloat = GS (InteriorFloat)
