(*
   File: sparse.ml
   Author: Shivkumar Chandrasekaran
   Date: 4.26.2003
   Purpose: Block sparse linear algebra routines

   $Id: sparse.ml,v 1.10 2006/03/31 17:49:37 shiv Exp shiv $
*)


open Nla


module type SPARSE =
  sig
    type t
    type matrix
    type sparse
    val get : sparse -> int -> int -> matrix option
    val set : sparse -> int -> int -> matrix option -> unit
    val empty : int -> int -> sparse
    val no_of_row_partitions : sparse -> int
    val no_of_col_partitions : sparse -> int
    val row_partitions : sparse -> int array
    val col_partitions : sparse -> int array
    val no_of_rows : sparse -> int
    val no_of_cols : sparse -> int
    val mul : ?transp:bool -> sparse -> matrix ->
      ?op:[ `AssignTo | `Incr | `Decr ] -> matrix -> unit
    type sparse_transform
    val ql_factorize' : ?ooc:bool -> sparse -> sparse_transform * sparse
    val lq_factorize' : ?ooc:bool -> sparse -> sparse * sparse_transform
    val qr_factorize' : ?ooc:bool -> sparse -> sparse_transform * sparse
    val rq_factorize' : ?ooc:bool -> sparse -> sparse * sparse_transform
    val mul_transform : ?transp:bool -> ?right:bool ->
      sparse_transform -> matrix -> unit
    val fSubs' : sparse -> matrix -> unit
    val bSubs' : sparse -> matrix -> unit
    val linear_solve' : ?use_ql:bool -> sparse -> matrix -> matrix
    val shallow_copy : sparse -> sparse
    val copy : ?ooc:bool -> sparse -> sparse
    val fastAdd : ?ooc:bool -> sparse -> sparse -> sparse
    val fastSub : ?ooc:bool -> sparse -> sparse -> sparse
    val fastMul : ?ooc:bool -> sparse -> sparse -> sparse
    val fastMul_transform : ?ooc:bool -> ?transp:bool -> ?right:bool ->
      sparse_transform -> sparse -> unit
    val fastFsubs' : ?ooc:bool -> sparse -> sparse -> unit
    val fastBsubs' : ?ooc:bool -> sparse -> sparse -> unit
    val fastLinear_solve' : ?use_ql:bool -> ?ooc:bool -> sparse ->
      sparse -> sparse
    val blkNorm1 : sparse -> float
    val spy : sparse -> unit
    val check_sparse : sparse -> bool
  end


module Sparser (Blop : BLOP) =
  struct

    open Blop
    type t = Blop.t
    type matrix = Blop.matrix


    (* A sparse matrix is made up of sub-matrices. Sub-matrices may
       not be square. Each node represents both a row and column. Also
       it has edges that originate at that node and represent each
       non-zero row entries. It also has edges that terminate at that
       node representing each non-zero column entry. *)

    (* The context-ual implementation of sorted lists based on Gerard
       Huet's ideas *)
    module Context : sig
      type 'a t
      val empty : unit -> 'a t
      val get : 'a t -> int -> 'a option
      val set : 'a t -> int -> 'a option -> unit
      val choose : 'a t -> (int * 'a) option
      val first : 'a t -> (int * 'a) option
      val next : 'a t -> (int * 'a) option
      val to_list : 'a t -> (int * 'a) list
      val of_list : (int * 'a) list -> 'a t
      val iter : ((int * 'a) -> unit) -> 'a t -> unit
    end =
      struct

	(* Contexts store pairs in increasing order, with the first
	   part in reverse order in head, and the last part in tail *)
	type 'a t = {mutable head : (int * 'a) list;
		     mutable tail : (int * 'a) list}

	let empty () = {head = []; tail = []}

	let check_context c =
	  let is_incr l =
	    fst (List.fold_left (fun (ok, i) (j, _) -> (ok & i < j, j))
		   (true, min_int) l)
	  in
	  let is_decr l =
	    fst (List.fold_left (fun (ok, i) (j, _) -> (ok & i > j, j))
		   (true, max_int) l)
	  in
	  is_decr c.head & is_incr c.tail

	let get c i =
	  let rec get_from_list head tail =
	    match tail with
	      [] -> raise Not_found
	    | (j,a) as w :: tls ->
		if i < j then raise Not_found
		else if i = j then a, (w :: head), tls
		else get_from_list (w :: head) tls in
	  let rec get_from_rev_list head tail =
	    match head with
	      [] -> raise Not_found
	    | (k,a) as v :: hds ->
		if k < i then raise Not_found
		else if k = i then a, hds, (v :: tail)
		else get_from_rev_list hds (v :: tail) in
	  let git head tail =
	    match head, tail with
	      [], [] -> raise Not_found
	    | _, [] -> get_from_rev_list head tail
	    | [], _ -> get_from_list head tail
	    | ((j,a) as w) :: hds, ((k,b) as v) :: tls ->
		if j < i && i < k then raise Not_found
		else if k < i then get_from_list (v :: head) tls
		else if i < j then get_from_rev_list hds (w :: tail)
		else if k = i then b, (v :: head), tls
		else (assert (j = i); a, hds, (w :: tail))
	  in
	  try
	    let a, new_head, new_tail = git c.head c.tail in
	    c.head <- new_head;
	    c.tail <- new_tail;
	    Some a
	  with
	    Not_found ->
	      None

	let set c i z =
	  let rec set_into_list head tail =
	    match tail with
	      [] -> begin match z with
		None -> raise Not_found
	      | Some zz -> (i, zz) :: head, tail
	      end
	    | (j,a) as w :: tls ->
		if i < j then begin
		  match z with
		    None -> raise Not_found
		  | Some zz -> (i,zz) :: head, tail
		end else if i = j then begin
		  match z with
		    None -> head, tls
		  | Some zz -> (i,zz) :: head, tls
		end else set_into_list (w :: head) tls in
	  let rec set_into_rev_list head tail =
	    match head with
	      [] -> begin match z with
		None -> raise Not_found
	      | Some zz -> head, (i,zz) :: tail
	      end
	    | (k,a) as v :: hds ->
		if k < i then begin
		  match z with
		    None -> raise Not_found
		  | Some zz -> head, (i,zz) :: tail
		end else if k = i then begin
		  match z with
		    None -> hds, tail
		  | Some zz -> hds, (i,zz) :: tail
		end else set_into_rev_list hds (v :: tail) in
	  let sit head tail =
	    match head, tail with
	      [], [] ->
		begin match z with
		  None -> raise Not_found
		| Some zz -> (i, zz) :: head, tail
		end
	    | [], _ -> set_into_list head tail
	    | _, [] -> set_into_rev_list head tail
	    | ((j,a) as w) :: hds, ((k,b) as v) :: tls ->
		assert (j < k);
		if j < i && i < k then begin
		  match z with
		    None -> raise Not_found
		  | Some zz -> (i, zz) :: head, tail
		end else if k < i then set_into_list (v :: head) tls
		else if i < j then set_into_rev_list hds (w :: tail)
		else if k = i then begin
		  match z with
		    None -> head, tls
		  | Some zz -> (i,zz) :: head, tls
		end else begin
		  assert (j = i);
		  match z with
		    None -> hds, tail
		  | Some zz -> hds, (i,zz) :: tail
		end
	  in
	  try
	    let new_head, new_tail = sit c.head c.tail in
	    c.head <- new_head;
	    c.tail <- new_tail
	  with
	    Not_found ->
	      ()

	let choose c =
	  match c.head with
	    [] ->
	      begin match c.tail with
		[] -> None
	      | d :: _ -> Some d
	      end
	  | d :: _ -> Some d

	let first c =
	  let tl = List.rev_append c.head c.tail in
	  match tl with
	    [] ->
	      c.head <- []; c.tail <- [];
	      None
	  | a :: tls ->
	      c.head <- [a]; c.tail <- tls;
	      Some a

	let next c =
	  match c.tail with
	    [] ->
	      None
	  | a :: tls ->
	      c.head <- a :: c.head; c.tail <- tls;
	      Some a

	let to_list c =
	  List.rev_append c.head c.tail

	let iter fn c =
	  List.iter fn (to_list c)

	let of_list lst =
	  let c = empty () in
	  List.iter (fun (i,zz) -> set c i (Some zz)) lst;
	  c

      end

    type node = { nz_row_entries : matrix Context.t;
		  nz_col_entries : matrix Context.t }
    and sparse = Sparse of int * int * node array


    let no_of_row_partitions (Sparse (m, _, _)) = m
    and no_of_col_partitions (Sparse (_, n, _)) = n

    let row_partitions (Sparse (_, _, nodes)) =
      Array.map (fun node ->
	match Context.choose node.nz_row_entries with
	  None -> 0
	| Some (_, d) -> noOfRows d) nodes

    let col_partitions (Sparse (_, _, nodes)) =
      Array.map (fun node ->
	match Context.choose node.nz_col_entries with
	  None -> 0
	| Some (_, d) -> noOfCols d) nodes

    let no_of_rows sp =
      Array.fold_left ( + ) 0 (row_partitions sp)

    let no_of_cols sp =
      Array.fold_left ( + ) 0 (col_partitions sp)

    let get (Sparse (m, n, nodes)) i j =
      assert (0 <= i & i < m & 0 <= j & j < n);
      Context.get nodes.(i).nz_row_entries j

	
    let check_sparse2 (Sparse (m, n, nodes) as sp) =
      for i = 0 to m-1 do
	for j = 0 to n-1 do
	  ignore (get sp i j)
	done
      done


    let set (Sparse (m, n, nodes)) i j b =
      assert (0 <= i & i < m & 0 <= j & j < n);
      Context.set nodes.(i).nz_row_entries j b;
      Context.set nodes.(j).nz_col_entries i b


    let check_sparse (Sparse (m, n, nodes)) =
      assert (max m n = Array.length nodes);
      let pi = ref 0 in
      let rec chk fn context =
	let a = Context.first context in
	match a with
	  None -> true
	| Some (i, a) ->
	    let m = fn a in
	    pi := i;
	    let rec chk_fn () =
	      let b = Context.next context in
	      match b with
		None -> true
	      | Some (ni, b) ->
		  m = fn b &&
		  (let t = !pi < ni in pi := ni; t) &&
		  chk_fn () in
	    chk_fn () in
      Array.fold_left (fun prev node ->
	prev && 
	chk noOfRows node.nz_row_entries && chk noOfCols node.nz_col_entries)
	true nodes


    let empty m n =
      Sparse (m, n,
	      Array.init (max m n)
		(fun _ -> {nz_row_entries = Context.empty ();
			   nz_col_entries = Context.empty ()}))


    let shallow_copy (Sparse (m, n, nodes)) =
      let sp_c = empty m n in
      for i = 0 to m-1 do
	let row_i = nodes.(i).nz_row_entries in
	let rec set_row_i sp_ij =
	  match sp_ij with
	    None -> ()
	  | Some (j, sp_ij) -> set sp_c i j (Some sp_ij);
	      set_row_i (Context.next row_i) in
	let sp_ij = Context.first row_i in
	set_row_i sp_ij
      done;
      sp_c
      

    let copy ?(ooc=false) (Sparse (m, n, nodes)) =
      let sp_c = empty m n in
      for i = 0 to m-1 do
	let row_i = nodes.(i).nz_row_entries in
	let rec set_row_i sp_ij =
	  match sp_ij with
	    None -> ()
	  | Some (j, sp_ij) -> set sp_c i j (Some (Blop.copy ~ooc sp_ij));
	      set_row_i (Context.next row_i) in
	let sp_ij = Context.first row_i in
	set_row_i sp_ij
      done;
      sp_c
      

    let mul ?(transp=false) (Sparse (_, _, sp) as a) b ?(op=`AssignTo) c =
      if not transp then begin
	let bs = partitionNx1 (col_partitions a) b
	and cs = partitionNx1 (row_partitions a) c in
	for i = 0 to Array.length cs - 1 do
	  begin match op with
	    `AssignTo -> const (fromInt 0) cs.(i)
	  | _ -> ()
	  end;
	  begin match op with
	    `AssignTo | `Incr -> Context.iter
		(fun (j, sp_ij) -> cs.(i) <-| (sp_ij |*++| bs.(j)))
		sp.(i).nz_row_entries
	  | `Decr -> Context.iter
		(fun (j, sp_ij) -> cs.(i) <-| (sp_ij |*--| bs.(j)))
		sp.(i).nz_row_entries
	  end
	done
      end else begin
	(* Transpose sp *)
	let bs = partitionNx1 (row_partitions a) b
	and cs = partitionNx1 (col_partitions a) c in
	for i = 0 to Array.length cs - 1 do
	  begin match op with
	    `AssignTo -> const (fromInt 0) cs.(i)
	  | _ -> ()
	  end;
	  begin match op with
	    `AssignTo | `Incr -> Context.iter
		(fun (j, sp_ij) -> cs.(i) <-| (sp_ij |~*++| bs.(j)))
		sp.(i).nz_col_entries
	  | `Decr -> Context.iter
		(fun (j, sp_ij) -> cs.(i) <-| (sp_ij |~*--| bs.(j)))
		sp.(i).nz_col_entries
	  end
	done
      end


    (* ql factorization subroutines *)
    (* output is reversed *)
    let rec merge_indices acc lst1 lst2 =
      match lst1, lst2 with
	(i, _) :: lst1_rem , (j, _) :: lst2_rem ->
	  if i = j then
	    merge_indices (i :: acc) lst1_rem lst2_rem
	  else if i < j then
	    merge_indices (i :: acc) lst1_rem lst2
	  else
	    merge_indices (j :: acc) lst1 lst2_rem
      | _ :: _, [] ->
	  let lst1', _ = List.split lst1 in List.rev_append lst1' acc
      | [], _ :: _ ->
	  let lst2', _ = List.split lst2 in List.rev_append lst2' acc
      | [], [] -> acc


    let merge_sparse_rows ns (row1 : (int * matrix) list) row2 =
      let col_indices = List.rev (merge_indices [] row1 row2) in
      let col_partitions = List.map (fun i -> ns.(i)) col_indices in
      let n = List.fold_left ( + ) 0 col_partitions
      and m1 = match row1 with
	(_, a) :: _ -> noOfRows a
      | [] -> 0
      and m2 = match row2 with
	(_, a) :: _ -> noOfRows a
      | [] -> 0 in
      let merged_rows = zeros (m1 + m2) n in
      let r1, r2 = partition2x1 m1 m2 merged_rows in
      let r1s = partition1xInf col_partitions r1
      and r2s = partition1xInf col_partitions r2 in
      let rec write_partitions r1s r2s (row1 : (int * matrix) list) row2 =
	match row1, row2 with
	  (i, a) :: row1_rem, (j, b) :: row2_rem ->
	    if i = j then begin
	      List.hd r1s <-$ a;
	      List.hd r2s <-$ b;
	      write_partitions (List.tl r1s) (List.tl r2s) row1_rem row2_rem
	    end else if i < j then begin
	      List.hd r1s <-$ a;
	      write_partitions (List.tl r1s) (List.tl r2s) row1_rem row2
	    end else begin
	      List.hd r2s <-$ b;
	      write_partitions (List.tl r1s) (List.tl r2s) row1 row2_rem
	    end
	| _ :: _, [] -> List.iter2 (fun ra (_, a) -> ra <-$ a) r1s row1
	| [], _ :: _ -> List.iter2 (fun rb (_, b) -> rb <-$ b) r2s row2
	| [], [] -> () in
      write_partitions r1s r2s row1 row2;
      (merged_rows, List.combine col_indices r1s, List.combine col_indices r2s)
	  

    (* apply op to the merger of the two sparse rows. *)
    let apply_sparse_rows ?(ooc=false) op ns m1 row1 m2 row2 =
      let doOp i a1 a2 =
	let a = rawMatrix (m1+m2) ns.(i) in
	matrix2x1 m1 a1
	          m2 a2 a;
	op a;
	let (a1n, a2n) = partition2x1 m1 m2 a in
	(i, Blop.copy ~ooc a1n, Blop.copy ~ooc a2n)
      in
      let rec appsr row1 row2 acc =
	match (row1, row2) with
	  ([], []) -> acc
	| ((i1, a1) :: row1s, []) ->
	    appsr row1s row2 (doOp i1 (iD a1) (const zero) :: acc)
	| ([], (i2, a2) :: row2s) ->
	    appsr row1 row2s (doOp i2 (const zero) (iD a2) :: acc)
	| ((i1, a1) :: row1s, (i2, a2) :: row2s) ->
	    if i1 = i2 then
	      appsr row1s row2s (doOp i1 (iD a1) (iD a2) :: acc)
	    else if i1 < i2 then
	      appsr row1s row2 (doOp i1 (iD a1) (const zero) :: acc)
	    else
	      appsr row1 row2s (doOp i2 (const zero) (iD a2) :: acc)
      in
      let rec splitNrev ls acc1 acc2 =
	match ls with
	  [] -> (acc1, acc2)
	| (i, a, b) :: rs -> splitNrev rs ((i, a) :: acc1) ((i, b) :: acc2)
      in
      splitNrev (appsr row1 row2 []) [] []


    (* apply op to the merger of the two sparse columns. *)
    let apply_sparse_cols ?(ooc=false) op ms col1 n1 col2 n2 =
      let doOp i a1 a2 =
	let a = rawMatrix ms.(i) (n1+n2) in
	matrix1x2 n1 n2
	          a1 a2 a;
	op a;
	let (a1n, a2n) = partition1x2 n1 n2 a in
	(i, Blop.copy ~ooc a1n, Blop.copy ~ooc a2n)
      in
      let rec appsc col1 col2 acc =
	match (col1, col2) with
	  ([], []) -> acc
	| ((i1, a1) :: col1s, []) ->
	    appsc col1s col2 (doOp i1 (iD a1) (const zero) :: acc)
	| ([], (i2, a2) :: col2s) ->
	    appsc col1 col2s (doOp i2 (const zero) (iD a2) :: acc)
	| ((i1, a1) :: col1s, (i2, a2) :: col2s) ->
	    if i1 = i2 then
	      appsc col1s col2s (doOp i1 (iD a1) (iD a2) :: acc)
	    else if i1 < i2 then
	      appsc col1s col2 (doOp i1 (iD a1) (const zero) :: acc)
	    else
	      appsc col1 col2s (doOp i2 (const zero) (iD a2) :: acc)
      in
      let rec splitNrev ls acc1 acc2 =
	match ls with
	  [] -> (acc1, acc2)
	| (i, a, b) :: rs -> splitNrev rs ((i, a) :: acc1) ((i, b) :: acc2)
      in
      splitNrev (appsc col1 col2 []) [] []


    let insert_row ?(ooc=false) (Sparse (m, n, _) as sp) i row_i =
      assert (0 <= i && i < m);
      let hooc a = if ooc then to_ooc a else a in
      List.iter (fun (k, b) -> set sp i k (Some (hooc b))) row_i
      

    let insert_col ?(ooc=false) (Sparse (m, n, _) as sp) j col_j =
      assert (0 <= j && j < n);
      let hooc a = if ooc then to_ooc a else a in
      List.iter (fun (k, b) -> set sp k j (Some (hooc b))) col_j
      

    type sparse_transform =
	SparseQ of int array * (transform * int * int) list


    (* Spine is destroyed. *)
    let lq_factorize' ?(ooc=false) (Sparse (m, n, nodes) as sp) =
      let ms = row_partitions sp
      and ns = col_partitions sp
      and l = empty m (min m n)
      and q_ref = ref [] in
      let hooc z = if ooc && not (is_ooc z) then to_ooc z else z in
      for i = 0 to (min m n) - 1 do
        (* Eliminate i-th row after i-th column *)
	let eliminate_ij (qs, col_i, sp_ii) (j, sp_ij) =
	  assert (i < j);
	  (* Compute q *)
	  let piv = rawMatrix ~ooc ms.(i) (ns.(i) + ns.(j)) in
	  matrix1x2 ns.(i) ns.(j)
	        (iD sp_ii) (iD sp_ij) piv;
	  let (lii, q) = lq' piv in
	  set sp i j None;
	  (* Apply q*)
	  let col_i', col_j' = apply_sparse_cols ~ooc (fun d -> d *~@ q) ms
	      col_i ns.(i)
	      (Context.to_list nodes.(j).nz_col_entries) ns.(j) in
	  insert_col ~ooc sp j col_j';
	  ((q,i,j) :: qs, col_i', copyL lii)
	in
	let sp_ii = match get sp i i with
	  None -> zeros ms.(i) ns.(i)
	| Some b -> b
	in
	set sp i i None; (* knock off pivot *)
	let row_i = Context.to_list nodes.(i).nz_row_entries
	and col_i = Context.to_list nodes.(i).nz_col_entries in
	if List.length row_i = 0 then begin
	  (* only pivot in row *)
	  let (l_ii, q) = lq' (Blop.copy ~ooc sp_ii) in
	  let l_ii = copyL ~ooc l_ii in
	  List.iter (fun (k, d) -> assert (k > i); d *~@ q) col_i;
	  q_ref := (q, i, i) :: !q_ref;
	  insert_col ~ooc l i col_i;
	  set l i i (Some l_ii)
	end else begin
 	  let (qs, new_col_i, l_ii) = List.fold_left eliminate_ij
	      (!q_ref, col_i, sp_ii) row_i
	  in
	  q_ref := qs;
	  insert_col ~ooc l i new_col_i;
	  set l i i (Some (hooc l_ii))
	end;
	(* Knock off column i in original matrix. *)
	List.iter (fun (k, _) -> assert (k > i); set sp k i None) col_i
      done;
      (l, SparseQ (ns, !q_ref))


    (* Spine is destroyed. *)
    let ql_factorize' ?(ooc=false) (Sparse (m, n, nodes) as sp) =
      let ms = row_partitions sp
      and ns = col_partitions sp
      and l = empty (min m n) n
      and q_ref = ref [] in
      let hooc z = if ooc && not (is_ooc z) then to_ooc z else z in
      for j = n - 1 downto max (n - m) 0 do
	(* Eliminate j-th column above m - n + j -th row *)
	let eliminate_ij (qs, row_j, sp_jj) (i, sp_ij) =
	  assert (i < m - n + j);
	  (* Compute q *)
	  let piv = rawMatrix ~ooc (ms.(i) + ms.(m-n+j)) ns.(j) in
	  matrix2x1 ms.(i)     (iD sp_ij)
	            ms.(m-n+j) (iD sp_jj) piv;
	  let (q, ljj) = ql' piv in
	  set sp i j None;
	  (* Apply q*)
	  let (row_i', row_j') = apply_sparse_rows ~ooc ((@~*) q) ns
              ms.(i) (Context.to_list nodes.(i).nz_row_entries)
              ms.(j) row_j in
	  insert_row ~ooc sp i row_i';
	  ((q,i,m-n+j) :: qs, row_j', copyL ljj)
	in
	let sp_jj = match get sp (m-n+j) j with
	  None -> zeros ms.(m-n+j) ns.(j)
	| Some b -> b
	in
	set sp (m-n+j) j None; (* knock off pivot *)
	let row_j = Context.to_list nodes.(m-n+j).nz_row_entries
	and col_j = List.rev (Context.to_list nodes.(j).nz_col_entries) in
	if List.length col_j = 0 then begin
	  (* only pivot in column *)
	  let (q, l_jj) = ql' (Blop.copy ~ooc sp_jj) in
	  let l_jj = copyL ~ooc l_jj in
	  List.iter (fun (k, d) -> assert (k < j); q @~* d) row_j;
	  q_ref := (q, m-n+j, m-n+j) :: !q_ref;
	  if m >= n then begin
	    insert_row ~ooc l j row_j;
	    set l j j (Some l_jj)
	  end else begin
	    insert_row ~ooc l (m-n+j) row_j;
	    set l (m-n+j) j (Some l_jj)
	  end
	end else begin
 	  let (qs, new_row_j, l_jj) = List.fold_left eliminate_ij
	      (!q_ref, row_j, sp_jj) col_j
	  in
	  q_ref := qs;
	  if m >= n then begin
	    insert_row ~ooc l j new_row_j;
	    set l j j (Some (hooc l_jj))
	  end else begin
	    insert_row ~ooc l (m-n+j) new_row_j;
	    set l (m-n+j) j (Some (hooc l_jj))
	  end
	end;
	(* Knock off row j in original matrix. *)
	List.iter (fun (k, _) -> assert (k < j); set sp (m-n+j) k None) row_j
      done;
      (SparseQ (ms, List.rev !q_ref), l)


    let mul_transform ?(transp=false) ?(right=false)
	(SparseQ (ms, qs)) b =
      if not right then begin
	let n = noOfCols b
	and bs = partitionNx1 ms b in
	let apply_q op (qi, r1, r2) =
	  if r1 = r2 then
	    op qi bs.(r1)
	  else begin
	    let temp = rawMatrix (ms.(r1) + ms.(r2)) n in
	    matrix2x1 ms.(r1) (iD bs.(r1))
 	              ms.(r2) (iD bs.(r2)) temp;
	    op qi temp;
	    let temp1, temp2 = partition2x1 ms.(r1) ms.(r2) temp in
	    bs.(r1) <-$ temp1;
	    bs.(r2) <-$ temp2
	  end
	in
	if not transp then
	  List.iter (apply_q ( @* )) (List.rev qs)
	else
	  List.iter (apply_q ( @~* )) qs
      end else begin
	let m = noOfRows b
	and bs = partition1xN ms b in
	let apply_q op (qi, r1, r2) =
	  if r1 = r2 then
	    op bs.(r1) qi
	  else begin
	    let temp = rawMatrix m (ms.(r1) + ms.(r2)) in
	    matrix1x2 ms.(r1) ms.(r2)
 	         (iD bs.(r1)) (iD bs.(r2)) temp;
	    op temp qi;
	    let temp1, temp2 = partition1x2 ms.(r1) ms.(r2) temp in
	    bs.(r1) <-$ temp1;
	    bs.(r2) <-$ temp2
	  end
	in
	if not transp then
	  List.iter (apply_q ( *@ )) qs
	else
	  List.iter (apply_q ( *~@ )) (List.rev qs)
      end


    let fSubs' (Sparse (m, n, nodes) as l) b =
      assert (m = n);
      let ms = row_partitions l in
      let bs = partitionNx1 ms b in
      for i = 0 to m-1 do
	let row_i = nodes.(i).nz_row_entries in
	let rec sub_row jlji =
	  match jlji with
	    Some (j, l_ji) ->
	      if j < i then begin
		bs.(i) <-| (l_ji |*--| bs.(j));
		  sub_row (Context.next row_i)
	      end else ()
	  | None -> failwith "Huh"
	in
	let l_ji = Context.first row_i in
	sub_row l_ji;
	match get l i i with
	  Some l_ii ->
	    subs' (isLowerTriangular l_ii) bs.(i)
	| None -> failwith "Huh"
      done


    let bSubs' (Sparse (m, n, nodes) as r) b =
      assert (m = n);
      let ms = row_partitions r in
      let bs = partitionNx1 ms b in
      for i = m-1 downto 0 do
	let row_i = Context.to_list (nodes.(i).nz_row_entries) in
	let sub_row (j, r_ij) =
	  assert (i <= j);
	  if i < j then
	    bs.(i) <-| (r_ij |*--| bs.(j))
	  else ()
	in
	List.iter sub_row row_i;
	let (k, r_ii) = List.hd row_i in
	assert (k = i);
	subs' (isUpperTriangular r_ii) bs.(i)
      done


    (* Destroys the spine of the input matrix. *)
    let qr_factorize' ?(ooc=false) (Sparse (m, n, nodes) as sp) =
      let ms = row_partitions sp
      and ns = col_partitions sp
      and r = empty (min m n) n
      and q_ref = ref [] in
      let hooc z = if ooc && not (is_ooc z) then to_ooc z else z in
      for j = 0 to (min m n) - 1 do
	(* Eliminate j-th column below j-th row *)
	let eliminate_ij (qs, row_j_j, sp_jj) (i, sp_ij) =
	  assert (i > j);
	  (* Compute q *)
	  let piv = rawMatrix ~ooc (ms.(j) + ms.(i)) ns.(j) in
	  matrix2x1 ms.(j) (iD sp_jj)
	            ms.(i) (iD sp_ij) piv;
	  let (q, rjj) = qr' piv in
	  set sp i j None;
	  (* Apply q*)
	  let (row_j', row_i') = apply_sparse_rows ~ooc ((@~*) q) ns
              ms.(j) row_j_j
              ms.(i) (Context.to_list nodes.(i).nz_row_entries) in
	  insert_row ~ooc sp i row_i';
	  ((q,j,i) :: qs, row_j', copyU rjj)
	in
	let sp_jj = match get sp j j with
	| None -> zeros ms.(j) ns.(j)
	| Some b -> b in
	set sp j j None;
	let row_j_j = Context.to_list nodes.(j).nz_row_entries
	and col_j_j = Context.to_list nodes.(j).nz_col_entries in
	if List.length col_j_j = 0 then	begin
	  let (q, r_jj) = qr' (Blop.copy ~ooc sp_jj) in
	  let r_jj = copyU ~ooc r_jj in
	  List.iter (fun (k, d) -> assert (j < k); q @~* d) row_j_j;
	  q_ref := (q, j, j) :: !q_ref;
	  insert_row ~ooc r j row_j_j;
	  set r j j (Some r_jj)
	end else begin
	  let (qs, new_row_j, r_jj) = List.fold_left eliminate_ij
	      (!q_ref, row_j_j, sp_jj) col_j_j
	  in
	  q_ref := qs;
	  insert_row ~ooc r j new_row_j;
	  set r j j (Some (hooc r_jj))
	end;
	List.iter (fun (k, _) -> assert (j < k); set sp j k None) row_j_j
      done;
      (SparseQ (ms, List.rev !q_ref), r)


    (* Destroys the spine of the input matrix. *)
    let rq_factorize' ?(ooc=false) (Sparse (m, n, nodes) as sp) =
      let ms = row_partitions sp
      and ns = col_partitions sp
      and r = empty m (min m n)
      and q_ref = ref [] in
      let hooc z = if ooc && not (is_ooc z) then to_ooc z else z in
      for i = m-1 downto max (m - n) 0 do
	(* Eliminate i-th row before i-m+n-th column *)
	let eliminate_ij (qs, col_i, sp_ii) (j, sp_ij) =
	  assert (i-m+n > j);
	  (* Compute q *)
	  let piv = rawMatrix ~ooc ms.(i) (ns.(j) + ns.(i)) in
	  matrix1x2 ns.(j) ns.(i)
	        (iD sp_ij) (iD sp_ii) piv;
	  let (rii, q) = rq' piv in
	  set sp i j None;
	  (* Apply q*)
	  let (col_j', col_i') = apply_sparse_cols ~ooc ((@~*) q) ms
              (Context.to_list nodes.(j).nz_col_entries) ns.(j)
              col_i ns.(i) in
	  insert_col ~ooc sp j col_j';
	  ((q,j,i) :: qs, col_i', copyU rii)
	in
	let sp_ii = match get sp i (i-m+n) with
	| None -> zeros ms.(i) ns.(i-m+n)
	| Some b -> b in
	set sp i (i-m+n) None;
	let row_i = Context.to_list nodes.(i).nz_row_entries
	and col_i = Context.to_list nodes.(i-m+n).nz_col_entries in
	if List.length row_i = 0 then	begin
	  let (r_ii, q) = rq' (Blop.copy ~ooc sp_ii) in
	  let r_ii = copyU ~ooc r_ii in
	  List.iter (fun (k, d) -> assert (i > k); d *~@ q) col_i;
	  q_ref := (q, i-m+n, i-m+n) :: !q_ref;
	  if m >= n then begin
	    insert_col ~ooc r (i-m+n) col_i;
	    set r i (i-m+n) (Some r_ii)
	  end else begin
	    insert_col ~ooc r i col_i;
	    set r i i (Some r_ii)
	  end
	end else begin
	  let (qs, new_col_i, r_ii) = List.fold_left eliminate_ij
	      (!q_ref, col_i, sp_ii) row_i
	  in
	  q_ref := qs;
	  if m >= n then begin
	    insert_col ~ooc r (i-m+n) new_col_i;
	    set r i (i-m+n) (Some (hooc r_ii))
	  end else begin
	    insert_col ~ooc r i col_i;
	    set r i i (Some (hooc r_ii))
	  end
	end;
	(* Knock off column i-m+n in original matrix *)
	List.iter (fun (k, _) -> assert (i > k); set sp k (i-m+n) None) col_i
      done;
      (r, SparseQ (ns, !q_ref))


    (* b <- sp \ b. The spine of sp is destroyed. b is destroyed a portion of
       which is returned as the solution. *)
    let linear_solve' ?(use_ql=true) sp b =
      let m = no_of_rows sp
      and n = no_of_cols sp in
      assert (m >= n);
      if use_ql then begin
	let sp_q, sp_l = ql_factorize' sp in
	mul_transform ~transp:true sp_q b;
	let _, x = partition2x1 (m - n) n b in
	fSubs' sp_l x;
	x
      end else begin
	let sp_q, sp_r = qr_factorize' sp in
	mul_transform ~transp:true sp_q b;
	let x, _ = partition2x1 n (m - n) b in
	bSubs' sp_r x;
	x
      end


    (* Multiplies two conformal sparse matrices to produce a third one. *)
    let fastMul ?(ooc=false) (Sparse (ma, ka, nodesa) as spa)
	(Sparse (kb, nb, nodesb) as spb) =
      assert (ka = kb && no_of_cols spa = no_of_rows spb);
      let ms = row_partitions spa
      and ns = col_partitions spb in
      let rec intersects ras cbs =
	match ras, cbs with
	  [], _ -> false
	| _, [] -> false
	| (ia, _) :: rass, (ib, _) :: cbss ->
	    if ia = ib then true
	    else if ia < ib then
	      intersects rass cbs
	    else
	      intersects ras cbss in
      let rec ip ras cbs acc =
	match (ras, cbs) with
	  ([], _) -> Some acc
	| (_, []) -> Some acc
	| ((ia, ra) :: rass, (ib, cb) :: cbss) ->
	    if ia = ib then begin
	      (ra |*++| cb) acc;
	      ip rass cbss acc
	    end else if ia < ib then
	      ip rass cbs acc
	    else
	      ip ras cbss acc in
      let spc = empty ma nb in
      for i = 0 to ma - 1 do
	let row_i = Context.to_list nodesa.(i).nz_row_entries in
	for j = 0 to nb - 1 do
	  let col_j = Context.to_list nodesb.(j).nz_col_entries in
	  if intersects row_i col_j then
	    set spc i j (ip row_i col_j (zeros ~ooc ms.(i) ns.(j)))
	done
      done;
      spc


    (* Adds two sparse matrices. *)
    let fastAdd ?(ooc=false) (Sparse (ma, na, nodesa) as spa)
	(Sparse (mb, nb, nodesb) as spb) =
      assert (ma = mb && na = nb && no_of_rows spa = no_of_rows spb &&
	      no_of_cols spa = no_of_cols spb);
      let spc = empty ma na in
      let hooc z =
	if ooc && not (is_ooc z) then
	  to_ooc z
	else
	  z
      in
      let to_hooc (j, cb) = (j, hooc cb) in
      let rec addRows rowa rowb acc =
	match rowa, rowb with
	  [], _ -> acc @ (List.map to_hooc rowb)
	| _, [] -> acc @ (List.map to_hooc rowa)
	| (ja, ca) :: rowas, (jb, cb) :: rowbs ->
	    if ja = jb then
	      addRows rowas rowbs ((ja, hooc (ca +$ cb)) :: acc)
	    else if ja < jb then
	      addRows rowas rowb ((ja, hooc ca) :: acc)
	    else
	      addRows rowa rowbs ((jb, hooc cb) :: acc)
      in
      for i = 0 to ma - 1 do
	let rowi = addRows (Context.to_list nodesa.(i).nz_row_entries)
	    (Context.to_list nodesb.(i).nz_row_entries) [] in
	List.iter (fun (j, cj) -> set spc i j (Some cj)) rowi
      done;
      spc
      

    (* Subtracts two sparse matrices. *)
    let fastSub ?(ooc=false) (Sparse (ma, na, nodesa) as spa)
	(Sparse (mb, nb, nodesb) as spb) =
      assert (ma = mb && na = nb && no_of_rows spa = no_of_rows spb &&
	      no_of_cols spa = no_of_cols spb);
      let spc = empty ma na in
      let hooc z =
	if ooc && not (is_ooc z) then
	  to_ooc z
	else
	  z
      in
      let to_neg (j, cb) =
	let c = Blop.copy cb in
	scale_real' (-1.0) c;
	(j, hooc c)
      in
      let to_hooc (j, cb) = (j, hooc cb) in
      let rec subRows rowa rowb acc =
	match rowa, rowb with
	  [], _ -> acc @ (List.map to_neg rowb)
	| _, [] -> acc @ (List.map to_hooc rowa)
	| (ja, ca) :: rowas, (jb, cb) :: rowbs ->
	    if ja = jb then
	      subRows rowas rowbs ((ja, hooc (ca -$ cb)) :: acc)
	    else if ja < jb then
	      subRows rowas rowb ((ja, hooc ca) :: acc)
	    else
	      subRows rowa rowbs (to_neg (jb, cb) :: acc)
      in
      for i = 0 to ma - 1 do
	let rowi = subRows (Context.to_list nodesa.(i).nz_row_entries)
	    (Context.to_list nodesb.(i).nz_row_entries) [] in
	List.iter (fun (j, cj) -> set spc i j (Some cj)) rowi
      done;
      spc
  

    (* Applies a sparse orthogonal transform to another sparse matrix. *)
    let fastMul_transform ?(ooc=false) ?(transp=false) ?(right=false)
	(SparseQ (ms, qs))	(Sparse (mb, nb, nodesb) as b) =
      if not right then begin
	let ns = col_partitions b in
	let apply_q op (qi, r1, r2) =
	  let apq1 b1 b2 j =
	    let temp = rawMatrix (ms.(r1)+ms.(r2)) ns.(j) in
	    matrix2x1 ms.(r1) b1
	              ms.(r2) b2 temp;
	    op qi temp;
	    let t1, t2 = partition2x1 ms.(r1) ms.(r2) temp in	
	    (j, Blop.copy ~ooc t1, Blop.copy ~ooc t2) in (* memory leak *)
	  let rec apq row1 row2 acc =
	    match row1, row2 with
	      [], [] -> acc
	    | [], ((j, cb2) :: row2s) ->
		let nb12 = apq1 (const zero) (iD cb2) j in
		apq row1 row2s (nb12 :: acc)
	    | (j, cb1) :: row1s, [] ->
		let nb12 = apq1 (iD cb1) (const zero) j in
		apq row1s row2 (nb12 :: acc)
	    | (j1, cb1) :: row1s, (j2, cb2) :: row2s ->
		if j1 = j2 then
		  let nb12 = apq1 (iD cb1) (iD cb2) j1 in
		  apq row1s row2s (nb12 :: acc)
		else if j1 < j2 then
		  let nb12 = apq1 (iD cb1) (const zero) j1 in
		  apq row1s row2 (nb12 :: acc)
		else
		  let nb12 = apq1 (const zero) (iD cb2) j2 in
		  apq row1 row2s (nb12 :: acc)
	  in
	  if r1 = r2 then
	    List.iter (fun (_, bi) -> op qi bi)
	      (Context.to_list nodesb.(r1).nz_row_entries)
	  else
	    let nr = apq (Context.to_list nodesb.(r1).nz_row_entries)
		(Context.to_list nodesb.(r2).nz_row_entries) [] in
	    List.iter
	      (fun (j, b1, b2) ->
		set b r1 j (Some b1);
		set b r2 j (Some b2))
	      nr
	in
	if not transp then
	  List.iter (apply_q ( @* )) (List.rev qs)
	else
	  List.iter (apply_q ( @~* )) qs
      end else begin
	let mbs = row_partitions b in
	let apply_q op (qj, c1, c2) =
	  let apq1 b1 b2 i =
	    let temp = rawMatrix mbs.(i) (ms.(c1)+ms.(c2)) in
	    matrix1x2 ms.(c1) ms.(c2)
	              b1      b2       temp;
	    op temp qj;
	    let t1, t2 = partition1x2 ms.(c1) ms.(c2) temp in	
	    (i, Blop.copy ~ooc t1, Blop.copy ~ooc t2) in (* memory leak *)
	  let rec apq col1 col2 acc =
	    match col1, col2 with
	      [], [] -> acc
	    | [], ((i, rb2) :: col2s) ->
		let nb12 = apq1 (const zero) (iD rb2) i in
		apq col1 col2s (nb12 :: acc)
	    | (i, rb1) :: col1s, [] ->
		let nb12 = apq1 (iD rb1) (const zero) i in
		apq col1s col2 (nb12 :: acc)
	    | (i1, rb1) :: col1s, (i2, rb2) :: col2s ->
		if i1 = i2 then
		  let nb12 = apq1 (iD rb1) (iD rb2) i1 in
		  apq col1s col2s (nb12 :: acc)
		else if i1 < i2 then
		  let nb12 = apq1 (iD rb1) (const zero) i1 in
		  apq col1s col2 (nb12 :: acc)
		else
		  let nb12 = apq1 (const zero) (iD rb2) i2 in
		  apq col1 col2s (nb12 :: acc)
	  in
	  if c1 = c2 then
	    List.iter (fun (_, bj) -> op bj qj)
	      (Context.to_list nodesb.(c1).nz_col_entries)
	  else
	    let nc = apq (Context.to_list nodesb.(c1).nz_col_entries)
		(Context.to_list nodesb.(c2).nz_col_entries) [] in
	    List.iter
	      (fun (i, b1, b2) ->
		set b i c1 (Some b1);
		set b i c2 (Some b2))
	      nc
	in
	if not transp then
	  List.iter (apply_q ( *@ )) qs
	else
	  List.iter (apply_q ( *~@ )) (List.rev qs)
      end


    (* Sparse L^{-1} times sparse *)
    let fastFsubs' ?(ooc=false) (Sparse (m, n, nodes) as l)
	(Sparse (mb, nb, nodesb) as b) =
      assert (m = n);
      let hooc z =
	if ooc && not (is_ooc z) then
	  to_ooc z
	else
	  z
      in
      for i = 0 to m-1 do
	let row_i = nodes.(i).nz_row_entries in
	let rec sub_row jlij =
	  match jlij with
	    Some (j, l_ij) ->
	      if j < i then begin
		List.iter (fun (k, bjk) ->
		  match get b i k with
		    None ->
		      let lb = l_ij *$ bjk in
		      scale_real' (-1.0) lb;
		      set b i k (Some (hooc lb))
		  | Some bik -> bik <-| (l_ij |*--| bjk))
		  (Context.to_list nodesb.(j).nz_row_entries);
		sub_row (Context.next row_i)
	      end else ()
	  | None -> failwith "Huh"
	in
	let l_ji = Context.first row_i in
	sub_row l_ji;
	match get l i i with
	  Some l_ii ->
	    let ll_ii = isLowerTriangular l_ii in
	    List.iter (fun (_, bi) -> subs' ll_ii bi)
	      (Context.to_list nodesb.(i).nz_row_entries)
	| None -> failwith "Huh"
      done


    (* Sparse R^{-1} times sparse. *)
    let fastBsubs' ?(ooc=false) (Sparse (m, n, nodes))
	(Sparse (mb, nb, nodesb) as b) =
      assert (m = n);
      let hooc z =
	if ooc && not (is_ooc z) then
	  to_ooc z
	else
	  z
      in
      for i = m-1 downto 0 do
	let row_i = Context.to_list (nodes.(i).nz_row_entries) in
	let sub_row (j, r_ij) =
	  assert (i <= j);
	  if i < j then
	    List.iter (fun (k, bjk) ->
	      match get b i k with
		Some bik -> bik <-| (r_ij |*--| bjk)
	      | None ->
		  let rb = r_ij *$ bjk in
		  scale_real' (-1.0) rb;
		  set b i k (Some (hooc rb)))
	      (Context.to_list nodesb.(j).nz_row_entries)
	  else ()
	in
	List.iter sub_row row_i;
	let (k, r_ii) = List.hd row_i in
	assert (k = i);
	let rr_ii = isUpperTriangular r_ii in
	List.iter (fun (_, bi) -> subs' rr_ii bi)
	  (Context.to_list nodesb.(i).nz_row_entries)
      done


    let blkpart2x1 m1 m2 (Sparse (ma, na, nodes) as a) =
      let m = no_of_rows a in
      assert (m = m1 + m2);
      let ms = row_partitions a in
      let rec split acc i =
	if acc = m1 then i
	else if acc < m1 then split (acc+ms.(i)) (i+1)
	else failwith "Improper split" in
      let l = split 0 0 in
      let a1 = empty l na
      and a2 = empty (ma-l) na in
      for i = 0 to l-1 do
	let rowi = Context.to_list nodes.(i).nz_row_entries in
	List.iter (fun (j, aj) -> set a1 i j (Some aj)) rowi
      done;
      for i = l to ma-1 do
	let rowi = Context.to_list nodes.(i).nz_row_entries in
	List.iter (fun (j, aj) -> set a2 (i-l) j (Some aj)) rowi
      done;
      (a1, a2)


    (* b <- sp \ spb. The spine of sp is destroyed. b is destroyed a portion of
       which is returned as the solution. *)
    let fastLinear_solve' ?(use_ql=true) ?(ooc=false) sp spb =
      let m = no_of_rows sp
      and n = no_of_cols sp in
      assert (m >= n);
      if use_ql then begin
	let sp_q, sp_l = ql_factorize' ~ooc sp in
	fastMul_transform ~ooc ~transp:true sp_q spb;
	let _, x = blkpart2x1 (m - n) n spb in
	fastFsubs' ~ooc sp_l x;
	x
      end else begin
	let sp_q, sp_r = qr_factorize' ~ooc sp in
	fastMul_transform ~ooc ~transp:true sp_q spb;
	let x, _ = blkpart2x1 n (m - n) spb in
	fastBsubs' ~ooc sp_r x;
	x
      end


    (* Block 1-norm. *)
    let blkNorm1 (Sparse (m, n, nodes)) =
      let nrm = ref 0.0 in
      for j = 0 to n-1 do
	let colj = Context.to_list nodes.(j).nz_col_entries in
	let nj = List.fold_left (fun acc (_, bi) -> acc +. norm1 bi) 0.0 colj in
	if nj > !nrm then nrm := nj;
      done;
      !nrm


    let spy sp =
      let m = no_of_row_partitions sp
      and n = no_of_col_partitions sp in
      for i = 0 to m-1 do
	for j = 0 to n-1 do
	  match get sp i j with
	    None -> Printf.printf "  "
	  | Some _ -> Printf.printf "* "
	done;
	Printf.printf "\n"
      done;
      flush stdout

  end




module DSparse = Sparser (DBlop)
module SSparse = Sparser (SBlop)
module ZSparse = Sparser (ZBlop)
module CSparse = Sparser (CBlop)
