(*
  File: sparse.mli
  Author: Shivkumar Chandrasekaran
  Date: 9.9.2003
  Purpose: Block sparse linear algebra routines

  $Id$
*)


(** {1 Sparse block matrix routines}
  
  This module provides a set of routines for dealing with sparse block
  matrices. These are block partitioned matrices, in which most of the
  blocks are zero-blocks. All the code is written in OCaml, on top of
  Lapack and Blas via the OCaml module Nla.

  This sparse matrix package is unusual in that it only lets you
  insert and delete blocks from the sparse matrix. However, as long as
  these operations are done either in row-major or column-major order,
  each such operation costs only O(1) ops (amortized).

   The module also supports out-of-core solves via the ooc flag of the
   Nla modules. In particular the individual blocks of a sparse matrix
   can be ooc. Also, the rhs can be an ooc sparse matrix.

  @author Shivkumar Chandrasekaran

*)
module type SPARSE =
  sig

    (** This is [float] in modules [DSparse] and [SSparse] and
      [Complex.t] in modules [ZSparse] and [CSparse]. *)
    type t

    (** This is the same as [Nla.DBlop.matrix] in [DSparse],
      [Nla.SBlop.matrix] in [SSparse], [Nla.ZBlop.matrix] in [ZSparse] and
      [Nla.CBlop.matrix] in [CSparse]. *)
    type matrix

    (** The basic data type for storing sparse block matrices. *)
    type sparse

    (** [get sp i j] returns the [(i,j)]-th block from the sparse
      matrix [sp]. The indexing is 0 based! If the block in the
      [(i,j)]-th position was a missing (and hence zero) block, then the
      call returns [None]. Otherwise it returns [Some a], where [a] is
      the block in the [(i,j)]-th position. Note, by modifying the
      entries of [a] you will change the corresponding entries of
      [sp]. *)
    val get : sparse -> int -> int -> matrix option

    (** [set sp i j (Some a)] will set the [(i,j)]-th block of the
      sparse matrix [sp] to the [matrix] [a]. On the other hand [set
      sp i j None] will delete the block entry (if any) in the
      [(i,j)]-th position of [sp]. *)
    val set : sparse -> int -> int -> matrix option -> unit

    (** [empty m n] returns a zero sparse matrix with [m] row
      partitions and [n] column partitions. *)
    val empty : int -> int -> sparse

    (** [no_of_row_partitions (empty m n)] returns m. *)
    val no_of_row_partitions : sparse -> int

    (** [no_of_col_partitions (empty m n)] returns n. *)
    val no_of_col_partitions : sparse -> int

    (** [(row_partitions sp).(i)] is the size of the [i]-th row
      partition of [sp]. Remember that partition indices are zero
      based. You see why now. *)
    val row_partitions : sparse -> int array

    (** [(col_partitions sp).(j)] is the size of the [j]-th colun
      partition of [sp]. Partition indices are zero based. *)
    val col_partitions : sparse -> int array

    (** [no_of_rows sp] returns the actual number of rows that [sp] has. *)
    val no_of_rows : sparse -> int

    (** [no_of_cols sp] returns the actual number of columns that [sp]
      has. *)
    val no_of_cols : sparse -> int

    (** [mul sp b c] replaces the contents of the [matrix] [c] with
      the product of the [sparse] matrix [sp] and the [matrix]
      [b]. You can suuply optional arguments which indicate whether
      [sp] is to be conjugate-transposed or whether the product should
      increment or decrement the entries in [c]. *)
    val mul : ?transp:bool -> sparse -> matrix ->
      ?op:[ `AssignTo | `Incr | `Decr ] -> matrix -> unit

    (** The type of the unitary factor computed by [ql_factorize']. *)
    type sparse_transform

    (** [ql_factorize' sp] computes the QL factorization of the
      [sparse] matrix [sp]. The spine of [sp] is destroyed but not the
      block matrices that make up [sp]. So, if you want to save a copy
      of [sp], the function [shallow_copy] is sufficient. *)
    val ql_factorize' : ?ooc:bool -> sparse -> sparse_transform * sparse

    (** [lq_factorize' sp] computes the LQ factorization of the
      [sparse] matrix [sp]. The spine of [sp] is destroyed but not the
      block matrices that make up [sp]. So, if you want to save a copy
      of [sp], the function [shallow_copy] is sufficient. *)
    val lq_factorize' : ?ooc:bool -> sparse -> sparse * sparse_transform

    (** [qr_factorize' sp] computes the QR factorization of the
      [sparse] matrix [sp]. The spine of [sp] is destroyed but not the
      block matrices that make up [sp]. So, if you want to save a copy
      of [sp], the function [shallow_copy] is sufficient. *)
    val qr_factorize' : ?ooc:bool -> sparse -> sparse_transform * sparse


    (** [rq_factorize' sp] computes the RQ factorization of the
      [sparse] matrix [sp]. The spine of [sp] is destroyed but not the
      block matrices that make up [sp]. So, if you want to save a copy
      of [sp], the function [shallow_copy] is sufficient.*) 
    val rq_factorize' : ?ooc:bool -> sparse -> sparse * sparse_transform

    (** [mul_transform q_sp b] will apply the sparse unitary transform
      [q_sp] computed by [ql_factorize'] to the [matrix] [b]. It takes
      an optional argument that specifies whether [q_sp] should be
      conjugate-transposed first. Note the function returns the answer
      in [b]. The optional parameter right, if set to true, will apply
      [q_sp] from the right. *)
    val mul_transform : ?transp:bool -> ?right:bool ->
      sparse_transform -> matrix -> unit

    (** [fSubs' l b] applies the inverse of the sparse block
      lower-triangular [l] (usually computed by [ql_factorize']) to
      the [matrix] [b]. *)
    val fSubs' : sparse -> matrix -> unit

    (** [bSubs' r b] applies the inverse of the sparse block
      upper-triangular [r] (usually computed by [qr_factorize']) to
      the [matrix] [b]. *)
    val bSubs' : sparse -> matrix -> unit

    (** [linear_solve' sp b] applies the inverse of [sp] to [b]. The
      spine of [sp] is destroyed. Use [shallow_copy] to make a copy if
      necessary. [b] is also destroyed and its storage is used for the
      computed solution that is returned. [sp] need not be square. *)
    val linear_solve' : ?use_ql:bool -> sparse -> matrix -> matrix

    (** [shallow_copy sp] returns a new sparse matrix that shares its
      blocks with [sp] but not its spine. *)
    val shallow_copy : sparse -> sparse

    (** [copy sp] returns a new sparse matrix that does not share
      its spine or its blocks with [sp]. *)
    val copy : ?ooc:bool -> sparse -> sparse

    (** [fastAdd spa spb] returns [spa] + [spb] in sparse form. *)
    val fastAdd : ?ooc:bool -> sparse -> sparse -> sparse

    (** [fastSub spa spb] returns [spa] - [spb] in sparse form. *)
    val fastSub : ?ooc:bool -> sparse -> sparse -> sparse

    (** [fastMul ~ooc spa spb] returns a new sparse matrix that is the
       product of spa and spc. If [ooc] is true then the individual sub-blocks
       of the product are out-of-core matrices. *)
    val fastMul : ?ooc:bool -> sparse -> sparse -> sparse

    (** [fastMul_transform spq sp] applies the sparse orthogonal transform
       [spq] to the sparse matrix [sp] in place. *)
    val fastMul_transform : ?ooc:bool -> ?transp:bool -> ?right:bool ->
      sparse_transform -> sparse -> unit
       
    (** sparse forward substitution on sparse rhs. *)  
    val fastFsubs' : ?ooc:bool -> sparse -> sparse -> unit

    (** sparse back substitution on sparse rhs. *)
    val fastBsubs' : ?ooc:bool -> sparse -> sparse -> unit

    (** same as linear_solve' except that the right-hand side is sparse, as
       is the solution. *)
    val fastLinear_solve' : ?use_ql:bool -> ?ooc:bool -> sparse ->
      sparse -> sparse
	
    (** Computes the block 1-norm. *)
    val blkNorm1 : sparse -> float

    (** Prints non-zero pattern *)
    val spy : sparse -> unit

    val check_sparse : sparse -> bool

  end


(** {1 Real and Complex Modules} *)

(** Real double precision routines. *)
module DSparse : (SPARSE with type t = float
			 and type matrix = Nla.DBlop.matrix)

(** Real single precision routines. *)
module SSparse : (SPARSE with type t = float
			 and type matrix = Nla.SBlop.matrix)

(** Complex double precision routines. *)
module ZSparse : (SPARSE with type t = Complex.t
			 and type matrix = Nla.ZBlop.matrix)

(** Complex single precision routines. *)
module CSparse : (SPARSE with type t = Complex.t
			 and type matrix = Nla.CBlop.matrix)
