(*
  File: testFmm.ml
  Author: Shivkumar Chandrasekaran
  Package: CamlFloat
  Date: 4.04.2006
  Email: shiv@ece.ucsb.edu
  Purpose: test fmm.ml

   $Id: testFmm.ml,v 1.2 2006/05/13 00:47:28 shiv Exp shiv $
*)

open Bigarray;;
open Utils;;
open Nla;;
open Sparse;;
open Fmm;;


module Tester (Blop : BLOP)
    (Sparse : SPARSE with type t = Blop.t and type matrix = Blop.matrix)
    (Fmm : FMM with type matrix = Blop.matrix and type t = Blop.t
    and type transform = Blop.transform and type sparse = Sparse.sparse) =
struct

  open Blop
  open Fmm

  let spyFmm fmmSpine =
    let w = Graphics.size_x ()
    and h = Graphics.size_y () in
    let rec drawB ob x y w h =
      match ob with
      | NoSplit -> Graphics.draw_rect x y w h
      | Split (oo, oi, io, ii) ->
	  let w1 = w/2
	  and h1 = h/2 in
	  let w2 = w - w1
	  and h2 = h - h1 in
	  let mx = x + w/2
	  and my = y + h/2 in
	  drawB oo x my w1 h2;
	  drawB ii mx y w2 h1;
	  drawB oi mx my w2 h2;
	  drawB io x  y  w1 h1
    in
    let rec drawFmm spine x y w h =
      match spine with
      | Leaf _ -> Graphics.draw_rect x y w h
      | Node (a, lSpine, rSpine) ->
	  let w1 = w/2
	  and h1 = h/2 in
	  let w2 = w - w1
	  and h2 = h - h1 in
	  let mx = x + w/2
	  and my = y + h/2 in
	  drawFmm lSpine x my w1 h2;
	  drawFmm rSpine mx y w2 h1;
	  drawB a.ou mx my w2 h2;
	  drawB a.ol x  y  w1 h1
    in
    drawFmm fmmSpine 0 0 w h  

  type node_split = { ku : int; lv : int; ou : odiag_split; ol : odiag_split }
  and leaf_split = { m : int; n : int; k : int; l : int }
  and fmm_spine = (leaf_split, node_split) btree

  let randomFmm ?(ooc=false) spine =
    let get_ku spine =
      match spine with
      | Leaf u -> u.k
      | Node (a, _, _) -> a.ku
    in
    let get_lv spine =
      match spine with
      | Leaf v -> v.l
      | Node (a, _, _) -> a.lv
    in
    let rec rec2 split lSpine rSpine =
      match split, lSpine, rSpine with
      | NoSplit, Leaf u, Leaf v ->
	  Reg (randomMatrix ~ooc u.k v.l)
      | Split (s00, s01, s10, s11), Node (al, llSpine, lrSpine),
	  Node (ar, rlSpine, rrSpine) ->
	    Sup {
	    oo = rec2 s00 llSpine rlSpine; oi = rec2 s01 llSpine rrSpine;
	    io = rec2 s10 lrSpine rlSpine; ii = rec2 s11 lrSpine rrSpine}
      | NoSplit, _, _ ->
	  Reg (randomMatrix ~ooc (get_ku lSpine) (get_lv rSpine))
      |  _, _, _ -> failwith "Bad fmm spine"
    in
    let rec rec1 spine =
      match spine with
      | Node (a, lSpine, rSpine) ->
	  let (lFmm, k0, l0, m0, n0) = rec1 lSpine in
	  let (rFmm, k1, l1, m1, n1) = rec1 rSpine in
	  (Node ({ml = m0; mr = m1; nl = n0; nr = n1;
		  bu = rec2 a.ou lSpine rSpine;
		  bl = rec2 a.ol rSpine lSpine;
		  rl = randomMatrix ~ooc k0 a.ku;
		  rr = randomMatrix ~ooc k1 a.ku;
		  wl = randomMatrix ~ooc l0 a.lv;
		  wr = randomMatrix ~ooc l1 a.lv;},
		 lFmm, rFmm),
	   a.ku, a.lv, m0+m1, n0+n1)
      | Leaf a ->
	  (Leaf {u = randomMatrix a.m a.k (* ones ~ooc a.m a.k *);
		 v = randomMatrix ~ooc a.n a.l;
		 d = randomMatrix ~ooc a.m a.n},
	   a.k, a.l, a.m, a.n)
    in
    let (fmm, _, _, _, _) = rec1 spine in
    fmm


  let spine1 = Node ({ku=0; lv=0; ou=NoSplit; ol=NoSplit},
    (Leaf {k=1; l=1; m=3; n=3}),
    (Leaf {k=1; l=1; m=4; n=4}))
  let spine2 = Node ({ku=0; lv=0;
		      ou = Split (NoSplit, NoSplit, NoSplit, NoSplit);
		      ol = Split (NoSplit, NoSplit, NoSplit, NoSplit)},
		     Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
			   (Leaf {k=1; l=1; m=3; n=3}),
			   (Leaf {k=1; l=1; m=4; n=4})),
		     Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
			   (Leaf {k=1; l=1; m=2; n=2}),
			   (Leaf {k=1; l=1; m=3; n=3})))
  let spine3 = Node ({ku=0; lv=0;
		      ou = Split (Split (NoSplit, NoSplit, NoSplit, NoSplit),
				  Split (NoSplit, NoSplit, NoSplit, NoSplit),
				  Split (NoSplit, NoSplit, NoSplit, NoSplit),
				  Split (NoSplit, NoSplit, NoSplit, NoSplit));
		      ol = Split (Split (NoSplit, NoSplit, NoSplit, NoSplit),
				  Split (NoSplit, NoSplit, NoSplit, NoSplit),
				  Split (NoSplit, NoSplit, NoSplit, NoSplit),
				  Split (NoSplit, NoSplit, NoSplit, NoSplit))},
		     Node ({ku=3; lv=3;
			    ou = Split (NoSplit, NoSplit, NoSplit, NoSplit);
			    ol = Split (NoSplit, NoSplit, NoSplit, NoSplit)},
			   Node ({ku=2; lv=2; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=2; n=2}),
				 (Leaf {k=1; l=1; m=3; n=3})),
			   Node ({ku=2; lv=2; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=3; n=3}),
				 (Leaf {k=1; l=1; m=4; n=4}))),
		     Node ({ku=3; lv=3;
			    ou = Split (NoSplit, NoSplit, NoSplit, NoSplit);
			    ol = Split (NoSplit, NoSplit, NoSplit, NoSplit)},
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=3; n=3}),
				 (Leaf {k=1; l=1; m=2; n=2})),
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=2; n=2}),
				 (Leaf {k=1; l=1; m=4; n=4}))
			  ))
  let spine4 = Node ({ku=0; lv=0;
		      ou = Split (NoSplit, NoSplit,
				  Split (NoSplit, NoSplit,
					 Split (NoSplit, NoSplit,
						NoSplit, NoSplit),
					 NoSplit),
				  NoSplit);
		      ol = Split (NoSplit,
				  Split (NoSplit,
					 Split (NoSplit, NoSplit,
						NoSplit, NoSplit),
					 NoSplit, NoSplit),
				  NoSplit, NoSplit)},
		     Node ({ku=1; lv=1;
			    ou = Split (NoSplit, NoSplit,
					Split (NoSplit, NoSplit,
					       NoSplit, NoSplit),
					NoSplit);
			    ol = Split (NoSplit,
					Split (NoSplit, NoSplit,
					       NoSplit, NoSplit),
					NoSplit, NoSplit)},
			   Node ({ku=1; lv=1;
				  ou = Split (NoSplit, NoSplit,
					      NoSplit, NoSplit);
				  ol = Split (NoSplit, NoSplit,
					      NoSplit, NoSplit)},
				 Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				       (Leaf {k=1; l=1; m=3; n=3}),
				       (Leaf {k=1; l=1; m=4; n=4})),
				 Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				       (Leaf {k=1; l=1; m=2; n=2}),
				       (Leaf {k=1; l=1; m=3; n=3}))),
			   Node ({ku=1; lv=1;
				  ou = Split (NoSplit, NoSplit,
					      NoSplit, NoSplit);
				  ol = Split (NoSplit, NoSplit,
					      NoSplit, NoSplit)},
				 Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				       (Leaf {k=1; l=1; m=3; n=3}),
				       (Leaf {k=1; l=1; m=4; n=4})),
				 Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				       (Leaf {k=1; l=1; m=2; n=2}),
				       (Leaf {k=1; l=1; m=3; n=3})))),
		     Node ({ku=1; lv=1;
			    ou = Split (NoSplit, NoSplit,
					Split (NoSplit, NoSplit,
					       NoSplit, NoSplit),
					NoSplit);
			    ol = Split (NoSplit,
					Split (NoSplit, NoSplit,
					       NoSplit, NoSplit),
					NoSplit, NoSplit)},
			   Node ({ku=1; lv=1;
				  ou = Split (NoSplit, NoSplit,
					      NoSplit, NoSplit);
				  ol = Split (NoSplit, NoSplit,
					      NoSplit, NoSplit)},
				 Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				       (Leaf {k=1; l=1; m=3; n=3}),
				       (Leaf {k=1; l=1; m=4; n=4})),
				 Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				       (Leaf {k=1; l=1; m=2; n=2}),
				       (Leaf {k=1; l=1; m=3; n=3}))),
			   Node ({ku=1; lv=1;
				  ou = Split (NoSplit, NoSplit,
					      NoSplit, NoSplit);
				  ol = Split (NoSplit, NoSplit,
					      NoSplit, NoSplit)},
				 Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				       (Leaf {k=1; l=1; m=3; n=3}),
				       (Leaf {k=1; l=1; m=4; n=4})),
				 Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				       (Leaf {k=1; l=1; m=2; n=2}),
				       (Leaf {k=1; l=1; m=3; n=3})))))
  let spine5 = Node ({ku=0; lv=0; ou = NoSplit; ol = NoSplit},
		     Node ({ku=1; lv=1; ou = NoSplit; ol = NoSplit},
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=3; n=3}),
				 (Leaf {k=1; l=1; m=4; n=4})),
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=2; n=2}),
				 (Leaf {k=1; l=1; m=3; n=3}))),
		     Node ({ku=1; lv=1; ou = NoSplit; ol = NoSplit},
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=3; n=3}),
				 (Leaf {k=1; l=1; m=4; n=4})),
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=2; n=2}),
				 (Leaf {k=1; l=1; m=3; n=3}))))
  let spine6 = Node ({ku=0; lv=0;
		      ou = Split (NoSplit, NoSplit,
				  Split (NoSplit, NoSplit, NoSplit, NoSplit),
				  NoSplit);
		      ol = Split (NoSplit,
				  Split (NoSplit, NoSplit, NoSplit, NoSplit),
				  NoSplit, NoSplit)},
		     Node ({ku=1; lv=1; ou = NoSplit; ol = NoSplit},
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=3; n=3}),
				 (Leaf {k=1; l=1; m=4; n=4})),
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=2; n=2}),
				 (Leaf {k=1; l=1; m=3; n=3}))),
		     Node ({ku=1; lv=1; ou = NoSplit; ol = NoSplit},
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=3; n=3}),
				 (Leaf {k=1; l=1; m=4; n=4})),
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=2; n=2}),
				 (Leaf {k=1; l=1; m=3; n=3}))))
  let spine7 = Node ({ku=0; lv=0; ou = NoSplit; ol = NoSplit},
		     Node ({ku=1; lv=1;
			    ou = Split (NoSplit, NoSplit, NoSplit, NoSplit);
			    ol = Split (NoSplit, NoSplit, NoSplit, NoSplit)},
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=3; n=3}),
				 (Leaf {k=1; l=1; m=4; n=4})),
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=2; n=2}),
				 (Leaf {k=1; l=1; m=3; n=3}))),
		     Node ({ku=1; lv=1;
			    ou = Split (NoSplit, NoSplit, NoSplit, NoSplit);
			    ol = Split (NoSplit, NoSplit, NoSplit, NoSplit)},
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=3; n=3}),
				 (Leaf {k=1; l=1; m=4; n=4})),
			   Node ({ku=1; lv=1; ou=NoSplit; ol=NoSplit},
				 (Leaf {k=1; l=1; m=2; n=2}),
				 (Leaf {k=1; l=1; m=3; n=3}))))

  let spines = [| spine1; spine2; spine3; spine4; spine5; spine6; spine7 |]

  let baseSpine = Node ({mL= -1; mR= -1; nL= -1; nR= -1;
			 Fmm.ou=NoSplit; Fmm.ol=NoSplit},
			Leaf {Fmm.m=3; Fmm.n=3}, Leaf {Fmm.m=4; Fmm.n=4})
  let hssSpine1 = Node ({mL= -1; mR= -1; nL= -1; nR= -1;
			 Fmm.ou=NoSplit; Fmm.ol=NoSplit},
			baseSpine, baseSpine)
  let hssSpine2 = Node ({mL= -1; mR= -1; nL= -1; nR= -1;
			 Fmm.ou=NoSplit; Fmm.ol=NoSplit},
			hssSpine1, hssSpine1)
  let hssSpine3 = Node ({mL= -1; mR= -1; nL= -1; nR= -1;
			 Fmm.ou=NoSplit; Fmm.ol=NoSplit},
			hssSpine2, hssSpine2)
  let baseSplit = Split (NoSplit, NoSplit, NoSplit, NoSplit)
  let fmmSpine1 = Node ({mL= -1; mR= -1; nL= -1; nR= -1;
			 Fmm.ou=baseSplit; Fmm.ol=baseSplit},
			baseSpine, baseSpine)
  let swSplit = Split (NoSplit, NoSplit, baseSplit, NoSplit)
  let neSplit = Split (NoSplit, baseSplit, NoSplit, NoSplit)
  let fmmSpine2 = Node ({mL= -1; mR= -1; nL= -1; nR= -1;
			 Fmm.ou=swSplit; Fmm.ol=neSplit},
			fmmSpine1, fmmSpine1)
  let swSplit2 = Split (NoSplit, NoSplit, swSplit, NoSplit)
  let neSplit2 = Split (NoSplit, neSplit, NoSplit, NoSplit)
  let fmmSpine3 = Node ({mL= -1; mR= -1; nL= -1; nR= -1;
			 Fmm.ou=swSplit2; Fmm.ol=neSplit2},
			fmmSpine2, fmmSpine2)

  let fmmSpines = [| baseSpine;
		     hssSpine1; hssSpine2; hssSpine3;
		     fmmSpine1; fmmSpine2; fmmSpine3 |]

  let kernel alpha omega xs sr m sc n a =
    assert (m = noOfRows a && n = noOfCols a);
    assert (alpha >= -1.0);
    let dim = Array.length xs.(0) in
    let kernelFn x y =
      fromFloat (
	let d =
	  let acc = ref 0.0 in
	  for i = 0 to dim-1 do
	    acc := !acc +. abs_float (x.(i) -. y.(i))
	  done;
	  !acc
	in
	if d > 0.0 then
	  (* sin *) cos (omega *. d) *. (d ** alpha)
	else if alpha = -1.0 then 1.0 +. omega
	else 1.0
      ) in
    let gFn i j = kernelFn xs.(i-1) xs.(j-1) in
    let lFn i j = gFn (sr + i - 1) (sc + j - 1) in
    fun2mat' lFn a

  let test1 () =
    for i = 0 to Array.length spines - 1 do
      let fmm = randomFmm spines.(i) in
      let m = no_of_rows fmm
      and n = no_of_cols fmm in
      (* let nrhs = 1 in *)
      let x = eye n n
      and b = rawMatrix m m in
      fastMul fmm x b
    done;
    true

  let test2 () =
    let ok = ref true
    and i = ref 0 in
    while !ok && !i <= Array.length fmmSpines - 1 do
      let (m, n) = fmmSize fmmSpines.(!i) in
      let a = randomMatrix m n in
      let tol = sqrt (sqrt machine_epsilon) in
      let aFmm = fromFn tol fmmSpines.(!i)
	  (fun sr m sc n dst -> dst <-$ (a $ (sr <-> sr+m-1, sc <-> sc+n-1)))
      in
      let aFmmMul = rawMatrix m n in
      fastMul aFmm (eye n n) aFmmMul;
      ok := !ok && (norm1 (a -$ aFmmMul) <= tol *. norm1 a);
      incr i;
    done;
    !ok

  let randomPoint n a b =
    Array.init n (fun i -> a.(i) +. Random.float (b.(i) -. a.(i)))

  let mkInt n1 n2 =
    let rec mi i acc =
      if i < n1 then acc else mi (i-1) (i :: acc)
    in
    mi n2 []

  let test3 () =
    let ok = ref true in
    for dim = 1 to 3 do
      let a = Array.make dim 0.0
      and b = Array.make dim 1.0
      and m = 250 in
      let p = 10 in
      let xs = List.map (fun _ -> randomPoint dim a b) (mkInt 1 m) in
      let ndTree = nestedDissection p xs in
      let fmmSpine = fmmSpineFromND ndTree in
      let (m, n) = fmmSize fmmSpine in
      let a = randomMatrix m n in
      let tol = sqrt (sqrt machine_epsilon) in
      let aFmm = fromFn tol fmmSpine
	  (fun sr m sc n dst -> dst <-$ (a $ (sr <-> sr+m-1, sc <-> sc+n-1)))
      in
      let aFmmMul = rawMatrix m n in
      fastMul aFmm (eye n n) aFmmMul;
      ok := !ok && norm1 (a -$ aFmmMul) <= tol *. norm1 a
    done;
    !ok

  let test4 () =
    let ok = ref true in
    for dim = 1 to 3 do
      let a = Array.make dim 0.0
      and b = Array.make dim 1.0
      and m = 1000
      and alpha = -0.5
      and omega = 1.5 *. acos 0.0 in
      let p = 100 in
      let xs = List.map (fun _ -> randomPoint dim a b) (mkInt 1 m) in
      let ndTree = nestedDissection p xs in
      let ys = Array.of_list (pointsFromNDtree ndTree) in
      let fmmSpine = fmmSpineFromND ndTree in
      let (m, n) = fmmSize fmmSpine in
      let a = rawMatrix m n in
      kernel alpha omega ys 1 m 1 n a;
      let tol = sqrt (sqrt machine_epsilon) in
      let aFmm = fromFn tol fmmSpine
	  (fun sr m sc n dst -> dst <-$ (a $ (sr <-> sr+m-1, sc <-> sc+n-1)))
      in
      let aFmmMul = rawMatrix m n in
      fastMul aFmm (eye n n) aFmmMul;
      let err = norm1 (a -$ aFmmMul) /. norm1 a in
      ok := !ok && err <= tol
    done;
    !ok

  let test5 () =
    let ok = ref true in
    for i = 0 to Array.length spines - 1 do
      let fmm = randomFmm spines.(i) in
      let m = no_of_rows fmm
      and n = no_of_cols fmm in
      let nrhs = 1 in
      let x = randomMatrix n nrhs
      and b = rawMatrix m nrhs
      and xc = rawMatrix n nrhs in
      fastMul fmm x b;
      fastSolve' fmm b xc;
      ok := !ok && norm1 (x -$ xc) <= norm1 x *. machine_epsilon *.
	  float m *. float n;
    done;
    !ok

  let test6 () =
    let ok = ref true in
    for dim = 1 to 3 do
      let a = Array.make dim 0.0
      and b = Array.make dim 1.0
      and ms = 1000
      and alpha = -0.5
      and omega = 1.5 *. acos 0.0
      and nrhs = 1 in
      let p = 100 in
      let xs = List.map (fun _ -> randomPoint dim a b) (mkInt 1 ms) in
      let ndTree = nestedDissection p xs in
      let ys = Array.of_list (pointsFromNDtree ndTree) in
      let fmmSpine = fmmSpineFromND ndTree in
      let (m, n) = fmmSize fmmSpine in
      let a = rawMatrix m n in
      kernel alpha omega ys 1 m 1 n a;
      let tol = sqrt (sqrt machine_epsilon) in
      let aFmm = fromFn tol fmmSpine
	  (fun sr m sc n dst -> dst <-$ (a $ (sr <-> sr+m-1, sc <-> sc+n-1)))
      in
      let x = randomMatrix n nrhs in
      let b = randomMatrix m nrhs in
      fastMul aFmm x b;
      let xc = rawMatrix n nrhs in
      fastSolve' aFmm b xc;
      let err = norm1 (x -$ xc) /. norm1 x in
      ok := !ok && err <= machine_epsilon *. (float (m * n))
    done;
    !ok

  let test7 () =
    Graphics.open_graph "";
    Graphics.resize_window (2 * Graphics.size_x ()) (2 * Graphics.size_y ());
    for dim = 1 to 3 do
      let a = Array.make dim 0.0
      and b = Array.make dim 1.0
      and m = 16000 in
      let p = 100 in
      let xs = List.map (fun _ -> randomPoint dim a b) (mkInt 1 m) in
      let ndTree = nestedDissection p xs in
      let fmmSpine = fmmSpineFromND ndTree in
      spyFmm fmmSpine;
      ignore (Graphics.read_key ());
      Graphics.clear_graph ();
    done;
    true

  let tests = [| test1; test2; test3; test4; test5; test6; test7 |]
      
  let run_tests () =
    let suc = ref 0
    and fal = ref 0
    and abo = ref 0
    and start = 7
    and fin = 20 in
    for i = max 0 (start - 1) to min (fin - 1) (Array.length tests - 1) do
      print_string "Test ";
      print_int (i + 1);
      begin
	try
	  if tests.(i) () then
	    (print_string " : Pass";
	     suc := !suc + 1)
	  else
	    (print_string " : Fail";
	     fal := !fal + 1)
	with
	  exn ->
	    (print_string " : Abort";
	     abo := !abo + 1; raise exn)
      end;
      print_newline ()
    done;
    print_string "Passed: ";
    print_int !suc;
    print_string "\nFailed: ";
    print_int !fal;
    print_string "\nAborted: ";
    print_int !abo;
    print_newline ()

end;; (* module Tester *)


module Dtests = Tester (DBlop) (DSparse) (DFmm);;
module Stests = Tester (SBlop) (SSparse) (SFmm);;
module Ztests = Tester (ZBlop) (ZSparse) (ZFmm);;
module Ctests = Tester (CBlop) (CSparse) (CFmm);;


let main () =
  begin
(*    print_string "Single Precision Tests:\n";
    Stests.run_tests ();
    print_newline ();
*)
    print_string "Double Precision Tests:\n";
    Dtests.run_tests ();
    print_newline ();
(*    
    print_string "Double Complex Tests:\n";
    Ztests.run_tests ();
    print_newline ();
    
    print_string "Single Complex Tests:\n";
    Ctests.run_tests ();
    print_newline ();
*)
  end;;


main ();;
