(*
  File: testSparse.ml
  Author: Shivkumar Chandrasekaran
  Package: CamlFloat
  Date: 5.1.2003
  Email: shiv@ece.ucsb.edu
  Purpose: test sparse.ml
*)

open Bigarray;;
open Nla;;
open Sparse;;


module Tester (Blop : BLOP)
    (Sparse : SPARSE with type t = Blop.t and type matrix = Blop.matrix) =
  struct
    
    open Blop
    open Sparse

    (* m = number of partitions, bw = block bandwidth, k = block size *)
    let random_banded m bw k =
      let sp = empty m m in
      for i = 0 to m - 1 do
	for l = i - bw to i + bw do
	  if 0 <= l && l < m then
	    set sp i l (Some (randomMatrix k k))
	done
      done;
      sp

    let random_arrow m n k =
      let sp = empty m n in
      for i = 0 to min m n - 1 do
	set sp i i (Some (randomMatrix k k))
      done;
      for j = 1 to n - 1 do
	set sp 0 j (Some (randomMatrix k k))
      done;
      for i = 1 to m - 1 do
	set sp i 0 (Some (randomMatrix k k))
      done;
      sp

    let ones_arrow m n k =
      let sp = empty m n in
      for i = 0 to min m n - 1 do
	set sp i i (Some (ones k k))
      done;
      for j = 1 to n - 1 do
	set sp 0 j (Some (ones k k))
      done;
      for i = 1 to m - 1 do
	set sp i 0 (Some (ones k k))
      done;
      sp

    let fivePt m n =
      let a = Sparse.empty ((m-2) * (n-2)) ((m-2) * (n-2))
      and grid = Array.make_matrix m n (-1) in
      for y = 1 to m-2 do
	for x = 1 to n-2 do
	  grid.(y).(x) <- (y-1) * (n-2) + (x-1)
	done
      done;
      for y = 1 to m-2 do
	for x = 1 to n-2 do
	  let i = (y - 1) * (n - 2)  + (x - 1) in
	  Sparse.set a i i (Some (consts (fromFloat (-4.0)) 1 1));
	  if x > 1 then
	    Sparse.set a i (grid.(y).(x-1)) (Some (consts (fromFloat 1.0) 1 1));
	  if x < n-2 then
	    Sparse.set a i (grid.(y).(x+1)) (Some (consts (fromFloat 1.0) 1 1));
	  if y > 1 then
	    Sparse.set a i (grid.(y-1).(x)) (Some (consts (fromFloat 1.0) 1 1));
	  if y < m-2 then
	    Sparse.set a i (grid.(y+1).(x)) (Some (consts (fromFloat 1.0) 1 1))
	done
      done;
      a


    let test1 () =
      let ms = [| 10; 11; 12; 13; 14; 15; |]
      and ks = [| 5; 6; 7; 8; 9; 10; 11; 12 |]
      and bws = [| 1; 2; 3 |]
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for bwi = 0 to Array.length bws - 1 do
	  for ki = 0 to Array.length ks - 1 do
	    let m = ms.(mi)
	    and k = ks.(ki)
	    and bw = bws.(bwi) in
	    let sp = random_banded m bw k in
	    let mp = no_of_rows sp in
	    let b = eye mp mp in
	    let sp_c = rawMatrix mp mp in
	    mul sp b sp_c;
	    let sp_q, sp_l = ql_factorize' sp
	    and l_c = rawMatrix mp mp in
	    mul sp_l b l_c;
	    mul_transform sp_q l_c;
	    ok := !ok && (norm1 (l_c -$ sp_c) <= float m *. 3.0 *.
	      machine_epsilon *. norm1 sp_c)
	  done
	done
      done;
      !ok

    let test2 () =
      let ms = [| 10; 11; 12; 13; 14; 15; |]
      and ks = [| 5; 6; 7; 8; 9; 10; 11; 12 |]
      and bws = [| 1; 2; 3 |]
      and nrhss = [| 1; 2; 3 |]
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for bwi = 0 to Array.length bws - 1 do
	  for ki = 0 to Array.length ks - 1 do
	    for nrhsi = 0 to Array.length nrhss - 1 do
	      for uql = 0 to 1 do
		let m = ms.(mi)
		and k = ks.(ki)
		and bw = bws.(bwi)
		and nrhs = nrhss.(nrhsi) in
		let sp = random_banded m bw k in
		let sp_c = Sparse.copy sp in
		let mp = no_of_rows sp in
		let b = randomMatrix mp nrhs in
		let b_c = Blop.copy b in
		let x = linear_solve' ~use_ql:(uql=0) sp b in
		mul sp_c x ~op:`Decr b_c;
		ok := !ok && (norm1 b_c <= machine_epsilon *.
				float (m * k * nrhs * bw) *. norm1 x)
	      done
	    done
	  done
	done
      done;
      !ok

    let test3 () =
      let ms = [| 10; 11; 12; 13; 14; 15; |]
      and ks = [| 5; 6; 7; 8; 9; 10; 11; 12 |]
      and bws = [| 1; 2; 3 |]
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for bwi = 0 to Array.length bws - 1 do
	  for ki = 0 to Array.length ks - 1 do
	    let m = ms.(mi)
	    and k = ks.(ki)
	    and bw = bws.(bwi) in
	    let sp = random_banded m bw k in
	    let mp = no_of_rows sp in
	    let b = eye mp mp in
	    let sp_c = rawMatrix mp mp in
	    mul sp b sp_c;
	    let sp_q, sp_r = qr_factorize' sp
	    and r_c = rawMatrix mp mp in
	    mul sp_r b r_c;
	    mul_transform sp_q r_c;
	    ok := !ok && (norm1 (r_c -$ sp_c) <= float m *. 3.0 *.
	      machine_epsilon *. norm1 sp_c)
	  done
	done
      done;
      !ok

    let test4 () =
      let ms = [| 10; 11; 12; 13; 14; 15; |]
      and ks = [| 5; 6; 7; 8; 9; 10; 11; 12 |]
      and bws = [| 1; 2; 3 |]
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for bwi = 0 to Array.length bws - 1 do
	  for ki = 0 to Array.length ks - 1 do
	    let m = ms.(mi)
	    and k = ks.(ki)
	    and bw = bws.(bwi) in
	    let spa = random_banded m bw k
	    and spb = random_arrow m m k in
	    let mp = no_of_rows spa
	    and kp = no_of_cols spa
	    and np = no_of_cols spb in
	    let spc = fastMul spa spb in
	    let sp_b = rawMatrix kp np in
	    mul spb (eye np np) sp_b;
	    let sp_ab = rawMatrix mp np in
	    mul spa sp_b sp_ab;
	    let sp_ab_c = rawMatrix mp np in
	    mul spc (eye np np) sp_ab_c;
	    ok := !ok && (norm1 (sp_ab -$ sp_ab_c) <= float m *.
			    machine_epsilon *. norm1 sp_ab)
	  done
	done
      done;
      !ok

    let test5 () =
      let ms = [| 10; 11; 12; 13; 14; 15; |]
      and ks = [| 5; 6; 7; 8; 9; 10; 11; 12 |]
      and bws = [| 1; 2; 3 |]
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for bwi = 0 to Array.length bws - 1 do
	  for ki = 0 to Array.length ks - 1 do
	    let m = ms.(mi)
	    and k = ks.(ki)
	    and bw = bws.(bwi) in
	    let spa = random_banded m bw k
	    and spb = random_arrow m m k in
	    let mp = no_of_rows spa
	    and np = no_of_cols spb in
	    let spc = fastAdd spa spb in
	    let sp_b = rawMatrix mp np in
	    mul spb (eye np np) sp_b;
	    let sp_a = rawMatrix mp np in
	    mul spa (eye np np) sp_a;
	    let sp_ab_c = rawMatrix mp np in
	    mul spc (eye np np) sp_ab_c;
	    ok := !ok && (norm1 (sp_a +$ sp_b -$ sp_ab_c) <= float m *.
			    machine_epsilon *. norm1 sp_ab_c)
	  done
	done
      done;
      !ok

    let test6 () =
      let ms = [| 10; 11; 12; 13; 14; 15; |]
      and ks = [| 5; 6; 7; 8; 9; 10; 11; 12 |]
      and bws = [| 1; 2; 3 |]
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for bwi = 0 to Array.length bws - 1 do
	  for ki = 0 to Array.length ks - 1 do
	    let m = ms.(mi)
	    and k = ks.(ki)
	    and bw = bws.(bwi) in
	    let spa = random_banded m bw k
	    and spb = random_arrow m m k in
	    let mp = no_of_rows spa
	    and np = no_of_cols spb in
	    let spc = fastSub spa spb in
	    let sp_b = rawMatrix mp np in
	    mul spb (eye np np) sp_b;
	    let sp_a = rawMatrix mp np in
	    mul spa (eye np np) sp_a;
	    let sp_ab_c = rawMatrix mp np in
	    mul spc (eye np np) sp_ab_c;
	    ok := !ok && (norm1 (sp_a -$ sp_b -$ sp_ab_c) <= float m *.
			    machine_epsilon *. norm1 sp_ab_c)
	  done
	done
      done;
      !ok

    let test7 () =
      let ms = [| 10; 11; 12; 13; 14; 15; |]
      and ks = [| 5; 6; 7; 8; 9; 10; 11; 12 |]
      and bws = [| 1; 2; 3 |]
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for bwi = 0 to Array.length bws - 1 do
	  for ki = 0 to Array.length ks - 1 do
	    let m = ms.(mi)
	    and k = ks.(ki)
	    and bw = bws.(bwi) in
	    let sp = random_banded m bw k in
	    let mp = no_of_rows sp in
	    let b = eye mp mp in
	    let sp_c = rawMatrix mp mp in
	    mul sp b sp_c;
	    let sp_q, sp_l = ql_factorize' sp
	    and sp_l_c = rawMatrix mp mp in
	    fastMul_transform sp_q sp_l;
	    mul sp_l b sp_l_c;
	    ok := !ok && (norm1 (sp_l_c -$ sp_c) <= float m *. 3.0 *.
	      machine_epsilon *. norm1 sp_c)
	  done
	done
      done;
      !ok

    let test8 () =
      let ms = [| 10; 11; 12; 13; 14; 15; |]
      and ks = [| 5; 6; 7; 8; 9; 10; 11; 12 |]
      and bws = [| 1; 2; 3 |]
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for bwi = 0 to Array.length bws - 1 do
	  for ki = 0 to Array.length ks - 1 do
	    for uql = 0 to 1 do
	      let m = ms.(mi)
	      and k = ks.(ki)
	      and bw = bws.(bwi) in
	      let sp = random_banded m bw k in
	      let sp_c = Sparse.copy sp in
	      let spb = random_arrow m m k in
	      let spb_c = Sparse.copy spb in
	      let x = fastLinear_solve' ~use_ql:(uql=0) sp spb in
	      ok := !ok && (blkNorm1 (fastSub spb_c (fastMul sp_c x))
			      <= machine_epsilon *.
			      (blkNorm1 sp_c *. blkNorm1 x +. blkNorm1 spb_c))
	    done
	  done
	done
      done;
      !ok

    let test9 () =
      let ms = [| 10; 11; 12; 13; 14; 15; |]
      and ks = [| 5; 6; 7; 8; 9; 10; 11; 12 |]
      and bws = [| 1; 2; 3 |]
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for bwi = 0 to Array.length bws - 1 do
	  for ki = 0 to Array.length ks - 1 do
	    let m = ms.(mi)
	    and k = ks.(ki)
	    and bw = bws.(bwi) in
	    let sp = random_banded m bw k in
	    let mp = no_of_rows sp in
	    let b = eye mp mp in
	    let sp_c = rawMatrix mp mp in
	    mul sp b sp_c;
	    let sp_l, sp_q = lq_factorize' sp
	    and lt_c = rawMatrix mp mp in
	    mul ~transp:true sp_l b lt_c;
	    mul_transform ~transp:true sp_q lt_c;
	    ok := !ok && (norm1 (transp lt_c -$ sp_c) <= float m *. 3.0 *.
	      machine_epsilon *. norm1 sp_c)
	  done
	done
      done;
      !ok

    let test10 () =
      let ms = [| 10; 11; 12; 13; 14; 15; |]
      and ks = [| 5; 6; 7; 8; 9; 10; 11; 12 |]
      and bws = [| 1; 2; 3 |]
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for bwi = 0 to Array.length bws - 1 do
	  for ki = 0 to Array.length ks - 1 do
	    let m = ms.(mi)
	    and k = ks.(ki)
	    and bw = bws.(bwi) in
	    let sp = random_banded m bw k in
	    let mp = no_of_rows sp in
	    let b = eye mp mp in
	    let sp_c = rawMatrix mp mp in
	    mul sp b sp_c;
	    let sp_l, sp_q = lq_factorize' sp
	    and l_c = rawMatrix mp mp in
	    mul sp_l b l_c;
	    mul_transform ~right:true sp_q l_c;
	    ok := !ok && (norm1 (l_c -$ sp_c) <= float m *. 3.0 *.
	      machine_epsilon *. norm1 sp_c)
	  done
	done
      done;
      !ok

    let test11 () =
      let ms = [| 10; 11; 12; 13; 14; 15; |]
      and ks = [| 5; 6; 7; 8; 9; 10; 11; 12 |]
      and bws = [| 1; 2; 3 |]
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for bwi = 0 to Array.length bws - 1 do
	  for ki = 0 to Array.length ks - 1 do
	    let m = ms.(mi)
	    and k = ks.(ki)
	    and bw = bws.(bwi) in
	    let sp = random_banded m bw k in
	    let sp_c = copy sp in
	    let sp_l, sp_q = lq_factorize' sp in
	    fastMul_transform ~right:true sp_q sp_l;
	    ok := !ok && (blkNorm1 (fastSub sp_l sp_c) <= float m *. 3.0 *.
	      machine_epsilon *. blkNorm1 sp_c)
	  done
	done
      done;
      !ok

    let test12 () =
      let ms = [| 10; 11 |]
      and ks = [| 250; 500 |]
      and ooc = true
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for ki = 0 to Array.length ks - 1 do
	  let m = ms.(mi)
	  and k = ks.(ki) in
	  let a = empty m m
	  and b = empty m m in
	  for i = 0 to m-1 do
	    for j = 0 to m-1 do
	      set a i j (Some (randomMatrix ~ooc k k));
	      set b i j (Some (randomMatrix ~ooc k k))
	    done
	  done;
	  let ac = copy ~ooc a in
	  let bc = copy ~ooc b in
	  let t1 = Sys.time () in
	  let x = fastLinear_solve' ~ooc ac bc in
	  let t2 = Sys.time () -. t1 in
	  Printf.printf "blks = %i blk_size = %i time = %g\n" m  k t2;
	  flush stdout;
	  ok := !ok && (blkNorm1 (fastSub ~ooc (fastMul ~ooc a x) b) <=
			float m *. 3.0 *. machine_epsilon *. blkNorm1 a)
	done
      done;
      !ok

    let test13 () =
      let ms = [| 4; 9 |] in
      let ns = ms
      and ok = ref true in
      for mi = 0 to Array.length ms - 1 do
	for ni = 0 to Array.length ns - 1 do
	  let m = ms.(mi)
	  and n = ns.(ni) in
	  let sp = fivePt m n in
	  let mp = no_of_rows sp in
	  let b = eye mp mp in
	  let sp_c = rawMatrix mp mp in
	  mul sp b sp_c;
	  let sp_q, sp_l = ql_factorize' sp
	  and l_c = rawMatrix mp mp in
	  mul sp_l b l_c;
	  mul_transform sp_q l_c;
	  ok := !ok && (norm1 (l_c -$ sp_c) <= float m *. 3.0 *.
			  machine_epsilon *. norm1 sp_c)
	done
      done;
      !ok

    let tests = [| test1; test2; test3; test4; test5; test6; test7; test8;
		   test9; test10; test11; test12; test13 |]
      
    let run_tests () =
      let suc = ref 0
      and fal = ref 0
      and abo = ref 0
      and start = 12
      and fin = 1000 in
      for i = max 0 (start - 1) to min (fin - 1) (Array.length tests - 1) do
	print_string "Test ";
	print_int (i + 1);
	begin
	  try
	    if tests.(i) () then
	      (print_string " : Pass";
	       suc := !suc + 1)
	    else
	      (print_string " : Fail";
	       fal := !fal + 1)
	  with
	    exn ->
	      (print_string " : Abort";
	       abo := !abo + 1; raise exn)
	end;
	print_newline ()
      done;
      print_string "Passed: ";
      print_int !suc;
      print_string "\nFailed: ";
      print_int !fal;
      print_string "\nAborted: ";
      print_int !abo;
      print_newline ()
	
  end;; (* module Tester *)

module Dtests = Tester (DBlop) (DSparse);;
module Stests = Tester (SBlop) (SSparse);;
module Ztests = Tester (ZBlop) (ZSparse);;
module Ctests = Tester (CBlop) (CSparse);;

let main () =
  begin
    print_string "Double Precision Tests:\n";
    Dtests.run_tests ();
    print_newline ();
    
    print_string "Double Complex Tests:\n";
    Ztests.run_tests ();
    print_newline ();
    
    print_string "Single Complex Tests:\n";
    Ctests.run_tests ();
    print_newline ();
    
    print_string "Single Precision Tests:\n";
    Stests.run_tests ();
    print_newline ();
  end;;


main ();;
